;#pragma StringComparison partial
include "hjconst.jsh"


;================================ Module compat ================================
; Compatibility shim to make modern scripts run on older JAWS versions.
; When this module supports a function, it supports it back through JAWS 11.0 unless otherwise indicated.
; Functions begin with comment lines indicating the function's classification:
;	Override:  An override (Unknown() should be the only one).
;	Support:  Support code for other functions.
;	NativeStart:  A JAWS function replacement for JAWS 11.0 through the version just before the one indicated on the comment line.
;
; To add a new function catcher:
;	- Add it in the If block in Unknown().
;	- Write compat_<funcName> after the fashion of the others in this module.
;
; Author:  Doug Lee


globals
	int Skype__compat___JAWSVer

int function Skype__compat___getJFWVersion()
; Support
; Used internally to speed recognition of the current JAWS version by caching the getJFWVersion() result.
if !Skype__compat___JAWSVer
	Skype__compat___JAWSVer = getJFWVersion()
endIf
return Skype__compat___JAWSVer
endFunction

string function Skype__compat___StringRef(variant v)
; Support
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

variant function unknown(string theName, int isScript,
	variant byRef v1, variant byRef v2, variant byRef v3, variant byRef v4, variant byRef v5,
	variant byRef v6, variant byRef v7, variant byRef v8, variant byRef v9, variant byRef v10
)
; Override
; Hander-offer for functions that are missing in the running JAWS.
if isScript
	return unknown(theName, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endIf
var int jver = Skype__compat___getJFWVersion()
if jver >= 1400000 && v1
	; Indicates recursion; don't mess with those.
	; This prevents, for instance, a recursive stringStartsWith() call
	; from getting sent incorrect arguments below in JAWS 14+.
	return unknown(theName, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endIf
if stringCompare(theName, "stringStartsWith", False) == 0
	return Skype__compat___stringStartsWith(v1, v2)
elif stringCompare(theName, "collectionItemCount", False) == 0
	return Skype__compat___collectionItemCount(v1)
elif stringCompare(theName, "collectionItemExists", False) == 0
	return Skype__compat___collectionItemExists(v1, v2)
elif stringCompare(theName, "collectionRemoveItem", False) == 0
	return Skype__compat___collectionRemoveItem(v1, v2)
elif stringCompare(theName, "collectionRemoveAll", False) == 0
	return Skype__compat___collectionRemoveAll(v1)
elif stringCompare(theName, "collectionCopy", False) == 0
	return Skype__compat___collectionCopy(v1)
endIf
return unknown(theName, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endFunction

int function Skype__compat___stringStartsWith(string s1, string s2)
; NativeStart: 13.00
var int len1 = stringLength(s1)
var int len2 = stringLength(s2)
if len1 < len2
	return False
endIf
s1 = stringLeft(s1, len2)
return (stringCompare(s1, s2, False) == 0)
endFunction

int function Skype__compat___collectionItemCount(collection c)
; NativeStart: 14.00
var string k
var int cnt = 0
forEach k in c
	cnt = cnt +1
endForEach
return cnt
endFunction

int function Skype__compat___collectionItemExists(collection c, string k0)
; NativeStart: 12.00
var variant null
if c[k0]
	; Catches non-empty values.
	return True
endIf
; Now to catch empty values.
var string k
forEach k in c
	if stringCompare(k, k0, False) == 0
	&& c[k] != null
		return True
	endIf
endForEach
return False
endFunction

void function Skype__compat___collectionRemoveAll(collection c)
; NativeStart: 12.00
var variant null
var string k
forEach k in c
	c[k] = null
endForEach
endFunction

int function Skype__compat___collectionRemoveItem(collection c, string k)
; NativeStart: 12.00
var variant null
if !Skype__compat___collectionItemExists(c, k)
	return False
endIf
c[k] = null
return True
endFunction

int function Skype__compat___collectionCopy(collection c)
; NativeStart: 12.00
; ToDo: This is not a deep copy and so will not always work as expected.
var collection nc nc = new collection
var string k
forEach k in c
	nc[k] = c[k]
endForEach
return nc
endFunction

;============================== End module compat ==============================

