; Scripts for the Skype internet telephony program
; Current Author: Doug Lee <dgl@dlee.org> (Skype username dlee42)
; Current URL: http://www.dlee.org/skype/
; Authorship info from 2008 or before:
; Original Author: Chris Nestrud <ccn@panix.com> (Skype username ChrisN)
; Original URL: http://www.panix.com/~ccn/projects/jfw/skype.php
; Contributing Developers:
; Noel Romey <ner@uark.edu> (Skype username dj_ner)
; Chad Foster
; Tyler Kavanaugh <twkav@kc.rr.com> (Skype Username Compguy2007)

use "skype_jcpDict.jsb"  ; JAWS Control Property manager
use "skype_compat.jsb"  ; compatibility with older JAWS versions

;#pragma StringComparison partial
include "hjconst.jsh"
include "hjGlobal.jsh"
include "common.jsm"
include "skype.jsh"
include "Skype.jsm"
include "SkypeMSAA.jsh"

globals
	int Skype___intAutoStartRun,
	string Skype___strTextBuffer,
	int Skype___intGotNewMessage,
	int Skype___lastChatIdx,
	handle Skype___lastChatHwnd,
	object Skype___lastChatList

void function Skype___checkForNewChats()
; Watches for new chat messages.
;Called by Schedule internally and from AutoStartEvent.  Also is called in Skype__sayIncomingChat. [TK, 2007-08-20]
var
	handle hwnd,
	object o,
	int cnt
scheduleFunction("Skype___checkForNewChats", 5)
if !Skype__opt__getUserOption(SV_RecognizeChat)
	return
endIf
hwnd = getFocus()
if Skype__util__isSpecialFocus() || !Skype__schat__isChatWindow(hwnd, False)
	return
endIf
hwnd = Skype__schat__getWindow(Skype__schat__WOutput, hwnd)
if !hwnd
	return
endIf
if hwnd != Skype___lastChatHwnd
	if Skype__conv5__applies()
		Skype___lastChatList = Skype__conv5__chatContent(hwnd)
	else
		Skype__sutil__sayUnsupportedSkype()
		return
	endIf
endIf
o = Skype___lastChatList
if !o
	return
elif o.accRole(0) != Role_System_List && o.accRole(0) != Role_System_Pane
	return
endIf
cnt = Skype__schat__getMessageCount(o)
if hwnd != Skype___lastChatHwnd
	; Don't read right after a conversation window switch.
	Skype___lastChatHwnd = hwnd
	Skype___lastChatIdx = cnt
	return
endIf
if Skype___lastChatIdx >= cnt
	; Nothing to read, but the list may have shrunk.
	; This happens when the user reduces displayed history in this window.
	if Skype___lastChatIdx > cnt
		sayMessage(OT_Message, formatString(msgChatsNoLongerShown, intToString(Skype___lastChatIdx -cnt)))
	endIf
	Skype___lastChatIdx = cnt
	return
endIf
if cnt -Skype___lastChatIdx < 5
	; Say the next chat message.
	; The scheduleFunction at the top of this function will catch anything after that.
	Skype___lastChatIdx = Skype___lastChatIdx +1
else
	; Don't say a whole flood.
	; This can happen when the user clicks for conversation history in this window.
	; The one line that is read is probably not new, so don't read it here.
	sayMessage(OT_Message, formatString(msgNewChatsNotRead, intToString(cnt -Skype___lastChatIdx)))
	Skype___lastChatIdx = cnt
	return
endIf
Skype__sayChatEntry(Skype___lastChatHwnd, Skype___lastChatIdx, "n")
endFunction

void Function AutoStartEvent ()
Skype__log__open("", False)
Skype__incsearch__setFlags("h")
if (Skype___intAutoStartRun != 1)
	SayFormattedMessage (OT_APP_START, msgSkypeStart_L, msgSkypeStart_S)
	Skype___intAutoStartRun = 1
EndIf
; if we re-enter a conversation window, the new message code is run. force this code not to run for the first haf-second.
Skype___intGotNewMessage=1
ScheduleFunction ("Skype___resetGotNewMessage", 5)
ScheduleFunction("Skype___checkForTypingStatusChange", 6)
scheduleFunction("Skype__appOpt__resetSkypeCommandKey", 10)
Skype___lastChatIdx = 0
Skype___lastChatHwnd = Skype__util__vcast(0)
scheduleFunction("Skype___checkForNewChats", 10)
Skype__mkcs__setHelpText(msgMKCSHelpText)
endFunction

globals
	handle Skype___typingLastWindow,
	object Skype___typingLastObj,
	string Skype___typingLastStatus
int function Skype___typingObjectValid(object o)
; Returns True if o seems to hold a valid object reference.
if o.accState(0) & State_System_Invisible
	; Invisible, must have gone out of focus.
	return False
elif o.accChildCount
	; It has descendants, so it should be valid.
	return True
elif o.accRole(0)
	; It has a role, so it should be valid.
	return True
endIf
return False
endFunction

int function Skype___setSKCTObject()
; Sets the object for the typing indicators or returns False if unsuccessful.
; Manager for Skype___typingLastWindow and Skype___typingLastObj
var
	handle hwnd
if Skype___typingLastWindow && isWindowVisible(Skype___typingLastWindow)
&& Skype___typingObjectValid(Skype___typingLastObj)
	hwnd = Skype___typingLastWindow
endIf
; The next condition catches top-level focus change and the cases where Skype___typingLastWindow is null.
; hwnd is only non-null now if Skype___typingLastWindow and Skype___typingLastObj still look valid.
if getTopLevelWindow(hwnd) != getTopLevelWindow(getFocus())
	hwnd = Skype__appwin__getSkypeWindow(SWT_ChatOutput)
	if !hwnd
		return False
	endIf
	if hwnd != Skype___typingLastWindow
		var int childID, object nullObject
		if Skype__conv5__applies()
			Skype___typingLastObj = Skype__conv5__chatContent(hwnd)
		else
			Skype__sutil__sayUnsupportedSkype()
			return False
		endIf
		if !Skype___typingLastObj
			return False
		elif !Skype___typingObjectValid(Skype___typingLastObj)
			return False
		endIf
		Skype___typingLastWindow = hwnd
	endIf
endIf  ; top-level focus change or Skype___typingLastWindow not set yet
return True
endFunction

void function Skype___checkForTypingStatusChange()
; Handles checking for and announcing changes in the other users' typing.
; Manager for Skype___typingLastStatus, which is a speakable string.
var
	string status,
	string soundLevel
if !Skype___setSKCTObject()
	scheduleFunction("Skype___checkForTypingStatusChange", 10)
	return
endIf
if !Skype__getTypingStatus(status, 10, False)
	scheduleFunction("Skype___checkForTypingStatusChange", 10)
	return
endIf
if stringCompare(status, Skype___typingLastStatus) != 0
	Skype___typingLastStatus = status
	if Skype__opt__getUserOption(SV_TypingAnnounce) && stringLength(status)
		sayUsingVoice(VCTX_Message, status, OT_Status)
	endIf
endIf
soundLevel = Skype__opt__getUserOption(SV_TypingSounder)
if soundLevel
	if status
		Skype__sutil__play(formatString(fnameSkypeTyping, soundLevel))
	endIf
endIf
scheduleFunction("Skype___checkForTypingStatusChange", 3)
endFunction

int function Skype__getTypingStatus(string byRef status, int maxSize, int returnList)
; Manages Skype___typingLastStatus and uses Skype___typingLastObj to do it.
; Called by getSkypeTypingStatusChange and by the sayTypingStatus script.
; If maxSize is above 0, windows with more than maxSize participants are not scanned.
var
	int n,
	int typerCount,
	string typerList,
	string typist, string typist0
status = ""
if !Skype___setSKCTObject()
	return False
endIf
n = Skype___typingLastObj.accChildCount
if !n
	; Valid object but no message entries.
	return True
endIf
; TODO:  maxSize not used.
typerCount = 0
typerList = ""
while n
	if Skype__schat__getTypist(Skype___typingLastObj, n, typist0)
		if typist0
			if !typerCount || returnList
				typist = typist0
			endIf
			if returnList
				typerList = typist +"\r\n" +typerList
			endIf
			typerCount = typerCount +1
		endIf  ; typist0
	else
		n = 1  ; exits loop
	endIf
	n = n -1
endWhile
if !typerCount
	return True
endIf
if returnList
	status = typerList
elif typerCount == 1
	status = formatString(msgUserTyping, typist)
else
	status = formatString(msgUsersTyping, intToString(typerCount))
endIf
return True
endFunction

Script HotkeyHelp ()
var
string version
let version=GetVersionInfoString (GetAppFilePath (), "FileVersion")
SayString ("Stai usando skype versione"+version)
;Added 10/24/2004 by ner, outputs hotkey help info and home page access link.
if UserBufferIsActive () UserBufferDeactivate() EndIf
Skype__sutil__sayMessageToUserBuffer(msgHotkeyHelp, True)
EndScript

void function keyboardHelpHook(string ScriptName, string FrameName)
; When scripts have numerical arguments, it mistakes it for frame names.
; This converts it to an integer, checks that it is a valid integer, then nulls the frame name argument.
var int iFrame =StringToInt(FrameName)
If (!iFrame  && GetCharacterValue(FrameName)==48)
|| iFrame
	FrameName=cscNull
EndIf
keyboardHelpHook(ScriptName, FrameName)
endFunction

Script SkypeScriptsHomePage ()
Skype__sutil__gotoSkypeScriptPage("", False)
EndScript

script ScreenSensitiveHelp()
if !isSameScript()
	performScript ScreenSensitiveHelp()
	return
endIf
Skype__sutil__gotoSkypeScriptPage("skypeman6.php", False)
endScript

globals int Skype___insideSayWindowTitle
script sayWindowTitle()
; Enhances JAWSKey+T.
var
	handle hFocus,
	handle hTop,
	string status
Skype___insideSayWindowTitle = True
performScript sayWindowTitle()
Skype___insideSayWindowTitle = False
hFocus = getFocus()
hTop = getTopLevelWindow(hFocus)
if !Skype__appwin__isMainSkypeWindow(hTop)
	if Skype__schat__isChatWindow(hTop, False)
		; SayWindowTitle will have said the conversation name by now,
		; so just say the status of the other person, if available.
		status = Skype___getChatStatus(hTop, True)
		if status
			sayMessage(OT_Document_Name, status)
		endIf
	endIf
	return
endIf
hTop = Skype__schat__getTopOfChat(hFocus)
if Skype__schat__isChatWindow(hTop, False)
	status = Skype___getChatStatus(hTop, False)
	if status
		sayMessage(OT_Document_Name, status)
	endIf
	return
endIf
while hFocus
	if getWindowClass(hFocus) == "TConversationForm"
		; Say the name of this conversation.
		sayMessage(OT_Document_Name, Skype__schat__getChatName(hFocus, True))
		return
	endIf
	hFocus = getParent(hFocus)
endWhile
endScript

int function focusWindowSupportsUIA()
if Skype___insideSayWindowTitle && getFocus() != getTopLevelWindow(getFocus())
	return False
endIf
return focusWindowSupportsUIA()
endFunction

string function getWindowName(handle hwnd)
if !Skype___insideSayWindowTitle
|| hwnd != getTopLevelWindow(getFocus())
	return getWindowName(hwnd)
endIf
return Skype__MSAA__windowName(hwnd)
endFunction

string function Skype___getChatStatus(handle hTop, int excludeName)
; Get a status description for the indicated chat.
var string status, int pos, string tmp
status = Skype__schat__getChatName(hTop, True)
if excludeName
	if stringRight(status, 1) == ")"
		pos = stringContainsFromRight(status, "(")
		status = stringChopLeft(status, pos)
		status = stringChopRight(status, 1)
	else
		status = ""
	endIf
endIf
tmp = Skype__schat__convDescription(hTop)
if tmp
	if status
		status = status +" "
	endIf
	status = status +tmp
endIf
return status
endFunction

string function Skype__getNameNear(string dir, string substr)
; Gets and returns the text (chunk) in the given direction from the current position as field name text.
; Dir can be left, right, up, or top.
; Top is special and uses the top line of the current window (for StringGrids like in Options > Hotkeys).
; If substr is given, the name is returned only if it contains substr.
var
	int x, int y,
	handle hwnd,
	int left, int right, int top, int bottom,
	int oldRestriction,
	int safetyCount,
	string name
hwnd = getCurrentWindow()
; X/Y used so we need not know which cursor is active.
x = getCursorCol()
y = getCursorRow()
top = getLineTop() -3
bottom = y+4
left = getWindowLeft(getTopLevelWindow(hwnd))
right = getWindowRight(getTopLevelWindow(hwnd))
saveCursor() invisibleCursor() saveCursor()
oldRestriction = getRestriction()
moveTo(x, y+2)
setRestrictionToRect(left, top, right, bottom)
if dir == "left"
	right = getCursorCol()
	; PriorChunk works for speech but not for Braille.  [DGL, 2007-03-01]
	priorWord()
	if getCurrentWindow() == hwnd
		moveTo(getWindowLeft(hwnd), getCursorRow())
		right = getCursorCol()
		priorWord()
	endIf
elif dir == "right"
	moveTo(getWindowRight(hwnd), getCursorRow())
	nextWord()
elif dir == "top"
	setRestriction(oldRestriction)
	moveToWindow(getCurrentWindow())
else  ; "up"
	setRestriction(oldRestriction)
	moveToWindow(getCurrentWindow())
	priorLine()
	if getWindowSubtypeCode(hwnd) == WT_RadioButton || getWindowSubtypeCode(hwnd) == WT_Checkbox
		safetyCount = 0
		while (stringIsBlank(getChunk()) || (getWindowSubtypeCode(getCurrentWindow()) == WT_RadioButton || getWindowSubtypeCode(getCurrentWindow()) == WT_CheckBox)) && safetyCount < 10
			priorLine()
			safetyCount = safetyCount +1
		endWhile
		if safetyCount >= 10
			Skype__log__write("Loop ran out.")
		endIf
	endIf
endIf
if getWindowSubtypeCode(hwnd) == WT_RadioButton || getWindowSubtypeCode(hwnd) == WT_Checkbox
	if dir != "top" && (getCurrentWindow() != getParent(hwnd) && getCurrentWindow() != getParent(getParent(hwnd)))
		; Avoid using another control as a label for this one.
		; Combos end up two levels below the containing window, hence the second condition above.
		setRestriction(oldRestriction)
		return ""
	endIf
endIf
var int right1
; In some non-English JAWS versions, hjconst.jsh is missing const IT_Chunk=1.
if !getItemRect(getCursorCol(), getCursorRow(), left, right1, top, bottom, 1)
	return ""
endIf
right = min(right, right1)
setRestrictionToRect(left, top, right, bottom)
;name = getTextInRect(left, top, right, bottom)
name = getChunk()
setRestriction(oldRestriction)
if !stringLength(substr)
	return name
elif stringContains(substr, "|")
	if stringContains("|"+stringLower(substr)+"|", "|"+stringLower(name)+"|")
		return name
	endIf
else
	if stringContains(stringLower(name), stringLower(substr))
	return name
	endIf
endIf
return ""
endFunction

int function Skype___fieldIsLink(handle hwnd)
if stringCompare(getWindowClass(hwnd), "TRichView", False)
	return False
endIf
if getCurrentWindow() != hwnd || getCurrentWindow() != getWindowAtPoint(getCursorCol(), getCursorRow())
	; The first condition covers calls with hwnd being a non-current window.
	; The second condition has been seen on Options > General's first link and Options > Call Forwarding & Voicemail > Basic/Advanced Settings link:
	; The PC cursor can lag behind JAWS' idea of the focused and current window by up to a couple of seconds.
	; [DGL, 2007-03-04, JAWS 8.0]
	saveCursor() invisibleCursor() saveCursor()
	moveToWindow(hwnd)
endIf
return (getCharacterAttributes() & Attrib_Underline)
endFunction

string function Skype__getActiveOptionsPageName()
var
	handle hwnd
hwnd = getFocus()
if getWindowClass(hwnd) == "TOptionsPanel"
	var object o, int childID
	o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
	if o
		var int i, string name
		i = o.accFocus +0
		name = o.accName(i)
		while i > 1
			i = i -1
			if o.accValue(i) == 0
				name = formatString("%1, %2", o.accName(i), name)
				return name
			endIf
		endWhile
		return name
	endIf
endIf
while hwnd
	if stringContains(stringLower(getWindowClass(hwnd)), "tabsheet")
		return getWindowName(hwnd)
	endIf
	hwnd = getParent(hwnd)
endWhile
return ""
endFunction

string function Skype___getOrdinalPosition(handle hwnd)
; Gets the ordinal position of hwnd in the set of windows at its level with the same subtype code.
; The degenerate case "1 of 1" returns null here.
; Counts from the end because windows seem to be created in reverse in this application.
var
	int stc,
	handle h1,
	int pos, int count
if getWindowTypeCode(getParent(hwnd)) == WT_ComboBox
	hwnd = getParent(hwnd)
endIf
stc = getWindowSubtypeCode(hwnd)
h1 = getLastWindow(hwnd)
count = 0
pos = 0
while h1
	if h1 == hwnd || (isWindowVisible(h1) && getWindowSubtypeCode(h1) == stc)
		count = count +1
	endIf
	if h1 == hwnd
		pos = count
	endIf
	h1 = getPriorWindow(h1)
endWhile
if !count || !pos || (count == 1 && pos == 1)
	return ""
endIf
return formatString(msgMOfN, intToString(pos), intToString(count))
endFunction

string function Skype___getTextWithAttributesFromWindow(handle hwnd, int attribs)
; Return the text of the given attribute or attribute combination from the given window.
var
	int left, int top, int right, int bottom
if !getWindowRect(hwnd, left, right, top, bottom)
	return ""
endIf
return getTextInRect(left, top, right, bottom, attribs)
endFunction

void function Skype__jcp__setCustomProps(collection props, string sOrigin)
; Obtains any properties of the indicated control that stray from what JAWS would normally discover and report.
; Sets props to be a structure for other jcp* functions.
; This is a callback function called by skype_jcpdict.jsb.
var
	int tc,
	handle hwnd, object o, int childID,
	int stc,
	int x, int y,
	string name
tc = getTickCount()

if Skype__util__isSpecialFocus()
	return
elif isVirtualPCCursor()
	return
endIf
Skype__jcp__setByOrigin(sOrigin, "hm", hwnd, o, childID)
if !(o.accState(childID) & State_System_Focused)
	childID = Skype___getFocusedID(o)
endIf
; Default names and sometimes values/tutor messages for things.
; The If avoids double title speaking on focus switch into Skype.
if sOrigin != "hwnd" || hwnd != getTopLevelWindow(getFocus())
	Skype__nameByMSAATree(props, hwnd, o, childID)
endIf
if o.accRole(childID) == Role_System_Pane
&& !Skype__jcp__get(props, "Name")
	props.Name = Skype__msgUnknownField
	props.type = " "
elif o.accRole(childID) == Role_System_ComboBox
&& o.accChild(childID).accChildCount == 0
	props.subtypeCode = WT_ButtonMenu
endIf

if Skype___nameSkypeProfileFields(hwnd, props)
	return
elif Skype___handleTechInfoList(hwnd, props)
	return
elif Skype___handleUserCheckList(hwnd, props)
	return
elif Skype___handleLive5Selector(props)
	return
elif Skype___fixEmptyListBox(props, hwnd)
	return
elif getWindowClass(hwnd) != "TChatRichEdit"
&& Skype___fixIncompleteFocus(props, hwnd)
	return
elif Skype___handleChangePasswordForm(props, hwnd)
	return
endIf

var string wClass = getWindowClass(hwnd)
if wClass == "TContactProfile"
	var object oProf = Skype__MSAA__objectFromWindow(hwnd)
	var object profID = oProf.accFocus
	if oProf.accRole(profID) == Role_System_StaticText
	&& stringIsBlank(oProf.accName(profID))
	&& oProf.accRole(profID+1) == Role_System_StaticText
	&& !stringIsBlank(oProf.accName(profID+1))
	&& !stringIsBlank(oProf.accDescription(profID+1))
		; For the mood message field for the highlighted contact.
		Skype__setProp(props, "Name", oProf.accName(profID+1), False)
		Skype__setProp(props, "Type", " ", False)
	endIf
endIf

stc = getWindowSubtypeCode(hwnd)
var string pwClass = getWindowClass(getParent(hwnd))
var string twClass = getWindowClass(getTopLevelWindow(hwnd))

if getWindowTypeCode(hwnd) == WT_Button && getWindowName(hwnd) == "btn"
	name = stringChopLeft(getWindowName(hwnd), 3)
	Skype__setProp(props, "Name", name, False)
elif wClass == "TConversationsControl"
	var object CLObj, int CLID, string CLName
	CLObj = Skype__sutil__getFocusObject(CLID)
	if CLObj.accRole(CLID) == Role_System_ListItem
		CLName = CLObj.accName(CLID)
		if !CLName
			CLName = Skype__MSAA__rectText(CLObj, CLID)
		endIf
		if !stringIsBlank(CLObj.accDescription(CLID))
			CLName = formatString("%1, %2", CLName, CLObj.accDescription(CLID))
		endIf
		if (CLObj.accState(CLID) & State_System_Selected)
		&& (!stringIsBlank(CLObj.accName(CLID)))
			CLName = msgSelected +" " +CLName
		endIf
		Skype__setProp(props, "Name", " ", True)
		if CLObj.accRole(CLID) == Role_System_ListItem
			Skype__setProp(props, "Type", " ", True)
			Skype__setProp(props, "Position", " ", True)
		endIf
		Skype__setProp(props, "Value", CLName, True)
		if clObj.accRole(0) == Role_System_List
			Skype__setProp(props, "ContainerName", clObj.accName(0), False)
		else
			Skype__setProp(props, "ContainerName", clObj.accParent.accName(0), False)
		endIf
	endIf  ; listItem
elif wClass == "TNavigableTntEdit" && pwClass == "TConversationsControl"
	Skype__setProp(props, "Name", Skype___msgContactFilter, False)
endIf

; This is applied generically but was aimed specifically at sound widget combo boxes in Skype 3.x active call tabs.
; PositionInGroup() is overridden in this file to use MSAA to get position info for these controls.
if wClass == "TNavigableTntComboBox"
	Skype__setProp(props, "Position", positionInGroup(), False)
endIf

if stringContains(stringLower(getWindowName(getTopLevelWindow(hwnd))), scEditSkypeOutContact)
&& stc == WT_Edit
	; SkypeOut Contact list entry > Context menu > Edit Contact
	Skype__setProp(props, "Name", Skype__getNameNear("up", ""), False)
elif stc == WT_Edit && pwClass == "TConversationsControl"
	; This distinguishes the contact rename edit from the contact filter edit.
	if hwnd == getFirstChild(getParent(hwnd))
		; Contact list entry > Context > Rename, or F2
		; This is to prevent JAWS from using the previous contact's name as a name for this edit field.
		Skype__setProp(props, "Name", " ", False)
	else
		; The (I think unnamed) box for filtering contacts by substring.
		Skype__setProp(props, "Name", msgSearchContacts, False)
	endIf
elif (wClass == "TChatRichEdit" || wClass == "TskRichEdit") && pwClass == "TChatEntryControl"
	; Name conversation input boxes after the other party.
	Skype__setProp(props, "Name", Skype__schat__getChatName(hwnd, True), True)
	Skype__setProp(props, "subtypeCode", WT_Multiline_Edit, True)
elif wClass == "TChatContentControl"
	var int otc = getObjectTypeCode()
	if (otc == WT_Button || otc == WT_Link)
		Skype__setProp(props, "ContainerName", " ", False)
		Skype__setProp(props, "ContainerType", " ", False)
	elif otc == WT_ListBoxItem
		var object oList = Skype__appwin__getMSAAObject(SWT_ChatOutput)
		var int focusedMessageID = Skype__schat__getMessageIndex(o)
		var int cnt = Skype__schat__getMessageCount(oList)
		if focusedMessageID
			Skype__jcp__setItemPos(props, focusedMessageID, cnt)
			Skype__setProp(props, "name", oList.accName(0), True)
			if Skype__opt__getUserOption(SV_ChatMessageTimes) == 0
				var string msgText = Skype__schat__getMessageText(oList, focusedMessageID, "")
				Skype__setProp(props, "value", msgText, True)
			endIf
		endIf
	endIf
elif wClass == "TMyselfControl"
&& getObjectSubtypeCode() == WT_Link
	; Skype 6.7 (and maybe older/newer 6.x) "Me" link at least.
	var object oMe, int idMe
	oMe = getFocusObject(idMe)
	if oMe
		var string meTxt = Skype__MSAA__rectText(oMe, idMe)
		if meTxt
			Skype__setProp(props, "Value", meTxt, True)
		endIf
	endIf
elif stringContains(wClass, "Edit") && pwClass == "tMyselfControl"
	Skype__setProp(props, "Name", msgMoodText, False)
elif Skype___fieldIsLink(hwnd)
	Skype__setProp(props, "SubtypeCode", WT_Link, False)
elif stringContains(stringLower(twClass), "options")
	; One of the Tools > Options pages is showing.
	if getObjectSubtypeCode() == WT_RadioButton
	&& getWindowClass(hwnd) == "TAccessibleRadioButtonEx"
	&& getWindowClass(getParent(hwnd)) == "TAccessibleRichRadioButton"
		; New in Skype 6.7.
		Skype__setProp(props, "Name", getWindowText(getParent(hwnd), False), True)
		Skype__setProp(props, "ContainerName", getWindowTextEx(getParent(getParent(hwnd)), False, False), True)
		Skype__setProp(props, "Position", Skype___getOrdinalPosition(getParent(hwnd)), False)
	endIf
	var
		string pageName
	pageName = Skype__getActiveOptionsPageName()
	if wClass == "TOptionsPanel"
		; Make SayLine work right after landing on the tree,
		; and add the page group name.
		Skype__setProp(props, "Value", pageName, True)
		Skype__setProp(props, "Name", " ", True)
		; Tree, not tree item.
		Skype__setProp(props, "SubtypeCode", WT_TreeView, True)
		; Position is wrong because this tree doesn't work normally, so leave it out.
		Skype__setProp(props, "Position", " ", True)
		; States like "opened" are superfluous here, so leave them out also.
		Skype__setProp(props, "State", " ", True)
		; PageUp/Down are not standard for a tree, so explain them to the user.
		Skype__setProp(props, "Tutor", msgOptionsTreeTutor, True)
	elif pageName == "tsRingtones" || pagename == "RingtonesTab"
		if stringContains(stringLower(wClass), "list")
		|| stringContains(stringLower(wClass), "combo")
			; Using Name can upset JAWS' normal handling of listBoxes.
			Skype__setProp(props, "beforeField", Skype__getNameNear("up", ""), False)
			if stringContains(stringLower(wClass), "list")
				var int checkStat
				checkStat = Skype__getCheckStat(12)
				if checkStat == Ctrl_Checked
					Skype__setProp(props, "attributes", Ctrl_Checked, False)
				elif checkStat == Ctrl_Unchecked
					Skype__setProp(props, "attributes", Ctrl_Unchecked, False)
				endIf
			endIf
		endIf
	elif pageName == "tsCallForwarding"
		if getWindowTypeCode(getParent(hwnd)) == WT_ComboBox
			Skype__setProp(props, "Name", formatString(msgEnterPhone, Skype___getOrdinalPosition(hwnd)), False)
		endIf
	elif pageName == "tsHotkeys" || pageName == "HotkeysTab"
		if wClass == "TTntListView" || wClass == "TAccessibleListView"
			var object HKObj, int HKID
			HKObj = getFocusObject(HKID)
			Skype__setProp(props, "Name", " ", True)
			Skype__setProp(props, "Value", hkObj.accName(hkID), False)
			Skype__setProp(props, "SubtypeCode", getObjectSubtypeCode(False, 1), False)
			Skype__setProp(props, "Attributes", getControlAttributes() & (~Ctrl_Selected), False)
		endIf
	endIf
elif stringContains(stringLower(twClass), "callquality")
&& (!getObjectName() || stringStartsWith(getObjectName(), "*"))
&& !stc && stringContains(stringLower(wClass), "button")
	; Call Quality Feedback buttons (set of five).
	var string cqName
	var object oBtn, int btnID
	oBtn = getObjectFromEvent(hwnd, -4, 0, btnID)
	cqName = oBtn.accDescription(1)
	if !cqName
		cqName = callQualityButtonNameByPosition(hwnd)
	endIf
	if !cqName
		saveCursor() invisibleCursor() saveCursor()
		routeInvisibleToPC()
		nextWord()
		cqName = getWindowText(getCurrentWindow(), Read_Everything)
		Skype__string__replace(cqName, " ", ": ", "a")
	endIf
	Skype__setProp(props, "Name", cqName, True)
elif twClass == "TskSendContacts" || twClass == "TskStartConference" || twClass == "TskSMSAdd"
	; Contacts > Send Contacts, Tools > Create a Conference Call, Add To SMS.
	if stringContains(stringLower(wClass), "userlist")
		; Setting Name changes how JAWS says this from "listview <line> <pos>" to "listbox item selected <pos>."  We don't want this.
		; See Skype__jcp__say() for the handling of beforeField.
		Skype__setProp(props, "BeforeField", Skype__getNameNear("up", ""), False)
	elif stc == WT_Edit && (twClass == "TskStartConference" || twClass == "TskSMSAdd")
		Skype__setProp(props, "Name", Skype__getNameNear("up", ""), False)
	elif stc == WT_EditCombo
		; JAWS speaks this control fine until something gets put in it.
		; This would fix that but also causes JAWS to call it an Edit instead of an Edit Combo.
		;Skype__setProp(props, "Name", msgSendContacts, False)
	endIf
elif twClass == "TskSendSMS"
	if !stringCompare(wClass, "TRichViewEdit", False)
		Skype__setProp(props, "Name", msgSMSSend, False)
	endIf
elif twClass == "TskBlockUsers"
	if stringContains(stringLower(wClass), "list")
		Skype__setProp(props, "BeforeField", Skype__getNameNear("up", ""), False)
	endIf
elif pwClass == "tDialPad"
|| pwClass == "TLoginControl"
|| getWindowClass(getParent(getParent(hwnd))) == "TLoginControl"
	if getWindowTypeCode(hwnd) == WT_Edit || getWindowTypeCode(hwnd) == WT_ComboBox || getWindowTypeCode(hwnd) == WT_EditCombo
		Skype__setProp(props, "Name", Skype__getNameNear("up", ""), False)
	endIf
endIf
endFunction

void function Skype__setProp(collection props, string whichProp, variant val, int force)
; Property setter() that allows earlier property values to override later ones.
if !force && collectionItemExists(props, whichProp)
	return
endIf
props[whichProp] = val
endFunction

int function Skype__nameByMSAATree(collection props, handle hwnd, object o, int childID)
; Name fields by MSAA information.

; First deal with an anomaly found in Skype 6 profile inserts:
; Role_System_StaticText controls without State_System_Focused set,
; causing getFocusObject and accFocus to return the parent Pane object.
var int iRole = o.accRole(childID)
if iRole == Role_System_Pane
	var object o1, int childID1
	if !isPCCursor()
		saveCursor() PCCursor()
	endIf
	o1 = getObjectAtPoint(childID1, getCursorCol(), getCursorRow())
	if o1
		if !(o1.accState(childID1) & State_System_Focusable) && (
			o1.accRole(childID1) == Role_System_StaticText
			|| o1.accRole(childID1) == Role_System_Link
		)
			o = o1
			childID = childID1
			Skype__setProp(props, "SubtypeCode", WT_Link, True)
		endIf
	endIf
endIf

if Skype___nameFromMSAA(props, hwnd, o, childID)
	return True
endIf

; Noticed first in Skype beta build 6.3.0.529:
; Many fields' MSAA info can best be found by chasing .accFocus from
; the ObjID_Client object drawn from the window parenting the one in focus.
o = Skype__MSAA__objectFromWindow(getParent(hwnd))
if !o
	return False
endIf
childID = 0
var int safety = 8
var int found = False
var variant v = o.accFocus
while !found && safety && v
	if intToString(v)
		childID = v
		found = True
	else
		o = v
		childID = 0
		if o.accChildCount
			v = o.accFocus
		else
			found = True
		endIf
	endIf
	safety = safety -1
endWhile
if !found
	return False
endIf
return Skype___nameFromMSAA(props, hwnd, o, childID)
endFunction

int function Skype___nameFromMSAA(collection props, handle hwnd, object o, int childID)
; Name the given item via MSAA if possible.
; Helper for Skype__nameByMSAATree().
var int iRole = o.accRole(childID)
var string val
if iRole == Role_System_ListItem
|| iRole == Role_System_OutlineItem
|| iRole == Role_System_MenuItem
	val = o.accName(childID)
	if val
		Skype__setProp(props, "Value", val, False)
	endIf
	childID = 0
elif iRole == Role_System_Pane && !o.accChild(childID).accChildCount
	Skype__setProp(props, "SubtypeCode", WT_Link, False)
elif iRole == Role_System_StaticText
	Skype__setProp(props, "Type", " ", False)
endIf
val = o.accName(childID)
if !val
	val = o.accHelp(childID)
endIf
if !val
	return False
endIf
Skype__setProp(props, "Name", val, False)
val = o.accDescription(childID)
if val && stringLeft(val, 6) != "skype:"
	Skype__setProp(props, "Tutor", val, False)
endIf
return True
endFunction

int function Skype___nameSkypeProfileFields(handle hwnd, collection props)
; Name the fields in a Skype profile screen.
if getWindowClass(hwnd) != "TMyselfProfileControl"
	return False
endIf
var object o, int childID
o = Skype__sutil__getFocusObject(childID)
if !o
	return False
endIf
var int iRole = o.accRole(childID) +0
var int i, string name
if iRole == Role_System_Link
	; Name by previous text item if any.
	if o.accRole(childID-1) == Role_System_StaticText
		Skype__setProp(props, "Name", o.accName(childID-1), True)
		; Move the original name to the value.
		Skype__setProp(props, "Value", o.accName(childID), True)
		Skype__setProp(props, "Tutor", Skype___msgProfileFieldTutor, False)
		return True
	endIf
elif iRole == Role_System_ComboBox
	; These are really irregular combo boxes:
	; No descendants, and must be clicked or Spaced on to open.
	; They do have names, but sometimes incomplete ones and sometimes
	; better ones than their preceding text items.
	; This algorithm nicely names the three Birthdate fields,
	; includes the current time in the Time field name,
	; and avoids repetition when text and combo names mean the same thing.
	; [DGL, 2011-01-23, Skype 5.1.0.104]
	name = o.accName(childID)
	i = childID -1
	while childID-i < 5 && o.accRole(i) == Role_System_ComboBox
		i = i -1
	endWhile
	if o.accRole(i) == Role_System_StaticText
		var string textName = o.accName(i)
		if stringChopRight(textName, 1) != name
			name = textName +" " +name
			Skype__setProp(props, "Name", name, True)
		endIf
		; Let the users know how to use these unconventional things.
		Skype__setProp(props, "SubtypeCode", WT_ButtonDropDown, True)
		Skype__setProp(props, "Tutor", Skype___msgProfileFieldTutor, False)
		return True
	endIf
elif iRole == Role_System_StaticText
	; Applies to mood text.
	Skype__setProp(props, "Tutor", Skype___msgProfileFieldTutor, True)
elif iRole == Role_System_ButtonDropdown
	if !childID
		; Skype 5.1 returns these as full-blown Accessible Objects.
		; For naming, we need this object's position among the others.
		o = o.accParent
		childID = Skype___getFocusedID(o)
	endIf
	i = childID -1
	var int textSeen = False
	while i > 0 && (!textSeen || childID -i < 5 && o.accRole(i) == Role_System_StaticText)
		if o.accRole(i) == Role_System_StaticText
			textSeen = True
		endIf
		i = i -1
	endWhile
	i = i +1
	name = ""
	while i < childID
		if o.accRole(i) == Role_System_StaticText
			name = formatString("%1 %2", name, o.accName(i))
		endIf
		i = i +1
	endWhile
	name = stringChopLeft(name +" " +Skype___msgVisibility +":", 1)
	name = formatString("%1 %2", name, o.accName(childID))
	Skype__setProp(props, "Name", name, True)
	return True
endIf
return False
endFunction

int function Skype___getFocusedID(object o)
; Get and return the childID with State_System_Focused set.
var int childID
childID = o.accFocus
if childID && (o.accState(childID) & State_System_Focused)
	return childID
endIf
var int n = o.accChildCount
childID = 1
while childID <= n
	if o.accState(childID) & State_System_Focused
		return childID
	endIf
	childID = childID +1
endWhile
return 0
endFunction

string function callQualityButtonNameByPosition(handle hwnd)
; Name the five Call Quality dialog buttons.
; Returns the applicable name if found, or the null string otherwise.
; Tested in Skype 6.16.0.105.
var int top = getWindowTop(hwnd)
var int bottom = getWindowBottom(hwnd)
hwnd = getFirstWindow(hwnd)
while hwnd
	var int cy = (getWindowTop(hwnd) +getWindowBottom(hwnd)) /2
	if getWindowClass(hwnd) == "TRichLabel"
	&& cy > top
	&& cy < bottom
		var object o, int childID
		o = getObjectFromEvent(hwnd, 0, ObjID_Client, childID)
		var int n = o.accChildCount
		childID = 1
		var string buf = ""
		while childID <= n
			if buf
				buf = buf +"\10"
			endIf
			buf = formatString("%1%2", buf, o.accName(childID))
			childID = childID +1
		endWhile
		return buf
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return ""
endFunction

int function Skype___handleTechInfoList(handle hwnd, collection props)
; Handle the announcement of items in the Skype 5.1 Technical Call Info screen.
if getWindowClass(hwnd) != "TWidgetControl"
|| getWindowClass(getParent(hwnd)) != "TCallTechInfoForm"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
var string txt = Skype__MSAA__rectText(o, childID)
if !txt
	return False
endIf
; Block out the MSAA "name" that is more like an item type.
Skype__setProp(props, "Name", " ", True)
; Assign as "name" the screen area's text content.
Skype__setProp(props, "Value", txt, True)
; Avoid saying "1 of 37" because the item count changes.
; (Text items appear and disappear as the user arrows.)
Skype__setProp(props, "Position", " ", True)
return True
endFunction

int function Skype___handleUserCheckList(handle hwnd, collection props)
; Handle the otherwise unreported checkbox states in the Send Contacts list.
; In Skhype 5.1: Conversations or Contacts entry, Context menu, Send Contacts;
; the list in that dialog.
if getWindowClass(hwnd) != "TCheckBoxUserListControl"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
if childID == 0
	; No item selected yet, at least according to JAWS.
	return False
endIf
; Skype__MSAA__rectText() should work here,
; but Gianluca discovered that it can fail to include the checkmark/corner for some reason.
var string txt = getTextBetween(getWindowLeft(hwnd), getWindowRight(hwnd))
Skype__setProp(props, "Value", txt, False)
Skype__setProp(props, "Name", " ", False)
return True
endFunction

int function Skype___handleLive5Selector(collection props)
; Handle the otherwise unreported checkbox states in the Live Call area Add People list.
var int stat = Skype__live5selector__selected()
if !stat
	return False
endIf
if stat == 1
	Skype__setProp(props, "Name", getObjectName(), False)
else
	Skype__setProp(props, "Name", " ", False)
	Skype__setProp(props, "Value", getObjectName(), False)
endIf
if stat == 1
	; Right list, not a contact entry.
	Skype__setProp(props, "Tutor", Skype___msgPressForMenu, False)
	return True
endIf
var int attribs = getControlAttributes()
if stat == 2
	attribs = attribs | Ctrl_Unchecked
else
	attribs = attribs | Ctrl_Checked
endIf
Skype__setProp(props, "Attributes", attribs, False)
return True
endFunction

int function Skype__getCheckStat(int offset)
var string sColor = ColorToRGBString(getColorAtPoint(getCursorCol()-offset, getCursorRow()))
; Black-on-white checkmarks (seen on Sarah's machine) (Skype 3.x though).
if sColor == "000000000"
	return Ctrl_Checked
elif sColor == "255255255"
	return Ctrl_Unchecked
endIf

; Blue checkmarks on gray background (my machine (DGL)).
var int r, int g, int b
r = stringToInt(stringLeft(sColor, 3))
g = stringToInt(substring(sColor, 4, 3))
b = stringToInt(substring(sColor, 7, 3))
if r-g == 0 && g-b == 0
	return Ctrl_Unchecked
elif b > r && b > g
	return Ctrl_Checked
endIf
return 0
endFunction

void function Skype__sayCheckState()
; Report checked/unchecked attributes where JAWS can't figure them out by itself.
var
	handle hwnd,
	string pageName,
	int attribs,
	int checkStat
hwnd = getFocus()
if !stringContains(stringLower(getWindowClass(getTopLevelWindow(hwnd))), "options")
	return
endIf
attribs = getControlAttributes()
if (attribs & Ctrl_Checked) || (attribs & Ctrl_Unchecked)
	; JAWS needs no help apparently.
	; Exception for Hotkeys page, where the wrong checkbox may be being queried.
	var object o, int childID
	o = getFocusObject(childID)
	if !o
		return
	endIf
	return
endIf
pageName = Skype__getActiveOptionsPageName()
if (pageName == "TSRingtones" || pagename == "RingtonesTab")
&& (getWindowClass(hwnd) == "TAccessibleCheckListBox" || getWindowClass(hwnd) == "TCheckListBox")
	checkStat = Skype__getCheckStat(12)
else
	return
endIf
indicateControlState(WT_CheckBox, checkStat)
endFunction

int function Skype___fixEmptyListBox(collection props, handle hwnd)
; Fix announcement of empty listboxes where getFocusObject returns an invalid childID of 1.
var object o = Skype__MSAA__objectFromWindow(hwnd)
if !o
	return False
elif o.accChildCount != 0
	return False
endIf
var int childID = o.accFocus
if childID != 1
	return False
endIf
; childID 1 and accChildCount 0.
Skype__setProp(props, "Origin", formatString("Point%1,%2",
	intToString(getCursorCol()),
	intToString(getCursorCol())
), False)
return True
endFunction

int function Skype___fixIncompleteFocus(collection props, handle hwnd)
; Fix announcement of controls where getFocusObject doesn't go far enough.
var object o, int childID, int childID1
o = getFocusObject(childID)
if !o
	return False
elif o.accChildCount == 0 || childID
	return False
endIf
childID1 = o.accFocus
if !childID1
	return False
endIf
o.accSelect(SelFlag_TakeFocus, 0)
o.accSelect(SelFlag_TakeFocus, childID1)
return True
endFunction

int function Skype___handleChangePasswordForm(collection props, handle hwnd)
; Handle fields in the Skype 5 Change Password form.
; Written against Skype 5.9.
if getWindowClass(getTopLevelWindow(hwnd)) != "TChangePasswordForm"
|| getObjectTypeCode() == WT_Button
	return False
endIf

; Name is to the left, hint is below.
var int left, int right, int top, int bottom
if !getWindowRect(hwnd, left, right, top, bottom)
	return False
endIf
saveCursor()  invisibleCursor()  saveCursor()
moveTo(left, (top+bottom)/2)
priorWord()
if getCursorRow() < top
	return False
endIf
var string name = getChunk()
moveTo(left+7, bottom+7)
var object o, int childID
o = getCurrentObject(childID)
if o.accRole(childID) == Role_System_StaticText
	name = formatString("%1 (%2)", name, o.accName(childID))
endIf
Skype__setProp(props, "Name", name, False)
return True
endFunction

globals object Skype___CLObj, int Skype___CLID
void function Skype___sayConversationControlPosition(object o, int childID)
; Say the position and count in the conversationsControl window accurately.
; This is done on a delay to avoid slowing navigation through lists.

if o
	Skype___CLObj = o
	Skype___CLID = childID
	scheduleFunction("Skype___sayConversationControlPosition", 10)
	return
endIf
o = Skype___CLObj
childID = Skype___CLID

; Count backward for speed.
var int i = o.accChildCount
; Assume all items count for n and just subtract those that don't, again for speed.
var int n = i
; but count up the position, which favors being near the top of the list a bit.
var int pos = 0
var int curItemFound = 0
while i && !isKeyWaiting()
	; Things that count are all listItems.
	if o.accRole(i) == Role_System_ListItem
	; This keeps out things like "Today" and "Older than one week" in the Conversations list.
	&& stringCompare(o.accName(i), o.accDescription(i), True) != 0
		if i == childID
			curItemFound = True
		endIf
		if curItemFound
			pos = pos +1
		endIf
	else
		n = n -1
	endIf
	i = i -1
endWhile
if !pos
	; If we don't have something accurate to say, say nothing.
	; This applies, for example, to "All other contacts" in the contact list.
	return
endIf
sayMessage(OT_Position, formatString(msgMOfN, intToString(pos), intToString(n)))
endFunction

string function Skype___getBannerText(handle hwnd)
; Get and return any text from the given (assumed banner or Help/About) window.
var int childID
var object o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	return ""
endIf
var int n = o.accChildCount
childID = 1
var string txt, string part
while childID <= n
	part = o.accName(childID)
	if part
		if txt
			txt = txt +"\n"
		endIf
		txt = txt +part
	endIf
	childID = childID +1
endWhile
return txt
endFunction

void function Skype___clearBanners()
; Wipe any visible banner off the screen.
; These things contain no text visible to JAWS, don't take focus, and block out text in other windows and apps.
; Also works against the Help/About box when not in focus, because it acts the same way.
var
 handle hwnd, handle hwndFocus,
	int safety
if userBufferIsActive()
	return
endIf
hwndFocus = getFocus()
safety = 3
while safety
	hwnd = findTopLevelWindow("TDCBanner", "")
	if !hwnd || !isWindowVisible(hwnd) || hwnd == hwndFocus
		hwnd = findTopLevelWindow("TAboutSplash", "")
		if !hwnd || !isWindowVisible(hwnd)
			return
		endIf
	endIf
	var string txt = Skype___getBannerText(hwnd)
	if txt
		;playSound(findJAWSSoundFile("ascend.wav"))
		pause()
		postMessage(hwnd, 0x10, 0, 0)  ; WM_Close
		pause()
		Skype__sutil__sayMessageToUserBuffer(txt, False)
		; Let the next focus change catch any other such windows.
		return
	else
		playSound(findJAWSSoundFile("descend.wav"))
	endIf
	postMessage(hwnd, 0x10, 0, 0)  ; WM_Close
	pause()
	safety = safety -1
endWhile
if !safety
	Skype__log__write("Loop ran out.")
endIf
endFunction

globals
; The getTickCount() value when the last FocusChangedEventEx was suppressed.
	int Skype___lastFCESuppressTick
void function focusChangedEventEx(
	handle hwndFocus, int nObject, int nChild,
	handle hwndPrevFocus, int nPrevObject, int nPrevChild,
	int nChangeDepth)
var string wclass = stringLower(getWindowClass(hwndFocus))

; Block chat input focus events on each typed character in Skype 5.1.
; Same for the box for entering phone numbers (tested in Skype 5.3).
; Same for the mood text edit box.
if hwndFocus == hwndPrevFocus
&& stringContains(wclass, "richedit")
	Skype___lastFCESuppressTick = getTickCount()
	return
endIf

; Adjust focus on the Facebook tab's list control.
if Skype___fixFBButtonFocus()
	return
endIf

; Move focus sensibly on exit from the emoticon selector.
if wclass == "TSmileyButtonPanel"
&& (!getWindowStyleBits(hwndPrevFocus) || getWindowClass(hwndPrevFocus) == "TSmileySelector")
	setFocus(findWindow(getParent(hwndFocus), "TChatRichEdit", ""))
	return
endIf

; Speak alerts like the one for old-style conferences on chat focus.
if wclass == "TChatEntryControl" && nObject == ObjID_Client
	Skype___sayTextItems(hwndFocus)
	; And continue as normal.
endIf

; Fix focus when the Add To Conference button (Skype 5.3) is pressed.
; This brings up a sort of menu whose root window takes focus
; but whose first item does not.
var handle hwndChild = getFirstChild(hwndFocus)
if wclass == "TCallPopup" && getWindowClass(hwndChild) == "TWidgetControl"
	setFocus(hwndChild)
	Skype__MSAA__objectFromWindow(hwndChild).accSelect(SelFlag_TakeFocus, 1)
	return
endIf

; Fix focus in the Skype 5.1 Technical Call Info dialog.
; The TCheckBoxUserListControl is from the Send Contacts list.
var handle hwnd
if wclass == "TCallTechInfoForm" && getWindowClass(hwndChild) == "TWidgetControl"
	hwnd =  hwndChild
elif wclass == "TCheckBoxUserListControl"
	hwnd = hwndFocus
endIf
if hwnd
	; Replace this focus event with a better one.
	var object o, int childID, int iRole
	o = getFocusObject(childID)
	iRole = o.accRole(childID)
	if iRole == Role_System_Pane || iRole == Role_System_Client
		o = Skype__MSAA__objectFromWindow(hwnd)
		if hwnd != hwndFocus
			o.accSelect(SelFlag_TakeFocus, 0)
		endIf
		; The actual list is represented by a single descendant List object.
		; Focus that.
		; Note that 1 here is for the List object, not the first ListItem.
		o.accSelect(SelFlag_TakeFocus, 1)
		; And swallow the current focus event.
		return
	endIf
endIf

; Fix a problem that applies in JAWS 7.0 and later.
if hwndFocus != hwndPrevFocus && nChangeDepth == -1
	nChangeDepth = getFocusChangeDepth()
endIf
focusChangedEventEx(hwndFocus, nObject, nChild, hwndPrevFocus, nPrevObject, nPrevChild, nChangeDepth)
endFunction

int function Skype___fixFBButtonFocus()
; Adjust focus on the Facebook entry in the Contacts list.
; Returns True if a fix attempt was made and False if not.
var object o, int childID
o = getFocusObject(childID)
if !o || !childID
	return False
elif stringCompare(o.accName(childID), "FaceBookEmptyItem", False) != 0
	return False
endIf
var int n = o.accChildCount
var int i = childID +1
childID = 0
while i <= n && !childID
	if o.accRole(i) == Role_System_PushButton
		childID = i
	elif o.accRole(i) != Role_System_StaticText
		return False
	endIf
	i = i +1
endWhile
if !childID
	return False
endIf
o.accSelect(SelFlag_TakeFocus, childID)
return True
endFunction

void function Skype___sayTextItems(handle hwnd)
; Say the text items in the given window's MSAA top level.
var object o = Skype__MSAA__objectFromWindow(hwnd)
if !o
	return
endIf
var string buf = ""
var string thisText
var int iRole
var int textSeen = False
var int addThis
var int n = o.accChildCount
var int i = 1
while i <= n
	addThis = False
	iRole = o.accRole(i)
	if iRole == Role_System_StaticText
		textSeen = True
		addThis = True
	elif textSeen && iRole == Role_System_Link
		addThis = True
	elif textSeen
		; Something non-text and non-link after text was seen.
		i = n  ; exit loop early.
	endIf
	if addThis
		thisText = stringTrimTrailingBlanks(stringTrimLeadingBlanks(o.accName(i)))
		if thisText
			buf = buf +" " +thisText
		endIf
	endIf
	i = i +1
endWhile
buf = stringTrimLeadingBlanks(buf)
if buf
	sayMessage(OT_Message, buf)
endIf
endFunction

void function tutorMessageEvent(handle hwndFocus, int nMenuMode)
; Blocks extra TutorMessageEvent speech when Skype 5.1 issues extra
; focusChange events while the user types into a chat window.
if Skype___lastFCESuppressTick && getTickCount() -Skype___lastFCESuppressTick < 500
	Skype___lastFCESuppressTick = 0
	return
endIf
tutorMessageEvent(hwndFocus, nMenuMode)
endFunction

Void Function FocusChangedEvent (handle FocusWindow, handle PrevWindow)
if Skype__util__isSpecialFocus()
	focusChangedEvent(focusWindow, prevWindow)
	return
endIf
scheduleFunction("Skype___clearBanners", 5)
if Skype__appwin__isMainSkypeWindow(focusWindow)
	; Handle focus here by forcing it to the Conversations list.
	scheduleFunction("fixFocus", 2)
	return
endIf
if getTopLevelWindow(focusWindow) != getTopLevelWindow(prevWindow)
	; If we've switched to a conversation window, silence recognizing for a haf-second so it won't inadvertently speak
	if Skype__appwin__isChatWindow(focusWindow)
		Skype___intGotNewMessage=1
		ScheduleFunction ("Skype___resetGotNewMessage", 5)
		scheduleFunction("Skype___checkCapsLock", 5)
	EndIf
	var handle thwnd = getTopLevelWindow(focusWindow)
	if thwnd != getTopLevelWindow(prevWindow)
	&& Skype__schat__isChatWindow(focusWindow, True)
		Skype___intGotNewMessage=1
		ScheduleFunction ("Skype___resetGotNewMessage", 5)
		if (focusWindow == thwnd || focusWindow == Skype__schat__getWindow(Skype__schat__WMain, thwnd))
			;Skype__sutil__play("DoDoop3.wav")
			Skype__util__setFocus(Skype__schat__getWindow(Skype__schat__WInput, getTopLevelWindow(focusWindow)), "d")
			return
		endIf
	endIf
endIf
focusChangedEvent(focusWindow, prevWindow)
EndFunction

void function fixFocus()
var
	handle focusWindow
focusWindow = getFocus()
if Skype__appwin__isMainSkypeWindow(focusWindow)
	stopSpeech()
	Skype__conv5__focusShowingList()
endIf
endFunction

void function Skype___checkCapsLock()
; Reports if CapsLock is on.
var
	int bIsDown, int bIsOn
if !getKeyState("CapsLock", bIsDown, bIsOn)
	return
endIf
if !bIsOn
	return
endIf
sayMessage(OT_Error, msgCapsLockIsOn)
endFunction

int function Skype__saySpecialHighlighting(string moveFunc, string speechFunc, string origScript)
; Speak highlighted text as it draws in windows requiring special highlight handling.
; if moveFunc is non-null, it is called first to move the cursor/highlight.
; If speechFunc is non-null, it is called in some cases to speak text.
; origScript should always be non-null and is the name of the script calling this function.
var
	handle hwnd,
	int left, int right, int top, int bottom
if Skype__util__isSpecialFocus() || !isPCCursor()
	formatStringWithEmbeddedFunctions("<$" +origScript +">")  ; Calls the original script
	return False
elif !speechFunc && (
	handle42conversationNavigation(origScript)
	|| handleLiveCallVolume(origScript)
)
	return True
endIf
hwnd = getFocus()
if stringContains(stringLower(getWindowClass(getTopLevelWindow(hwnd))), "options")
&& (
	getWindowClass(hwnd) == "TAccessibleCheckListBox"
	&& !((getControlAttributes() & Ctrl_Checked) || (getControlAttributes() & Ctrl_Unchecked))
	&& stringContains("tsringtones ringtonestab", stringLower(Skype__getActiveOptionsPageName()))
)
	if moveFunc
		formatStringWithEmbeddedFunctions("<" +moveFunc +">")  ; moves as appropriate
		pause()
	endIf
	return True
endIf

if (stringContains(stringLower(getWindowClass(getFocus())), "history")
|| stringContains(stringLower(getWindowClass(getFocus())), "alertlist"))
&& moveFunc
	; Reduce double speech in the History page when the user arrows past the bottom and scrolls the list.
	; SayNextLine speaks when the list scrolls, but the ActiveItemChangedEvent is supposed to speak this.
	formatStringWithEmbeddedFunctions("<" +moveFunc +">")
else
	formatStringWithEmbeddedFunctions("<$" +origScript +">")  ; Calls the original script
endIf
pause()
return True
endFunction

int function handle42conversationNavigation(string origScript)
; For contact/conversation/history list area:
; This makes Up/Down cross between live and non-live, and double Home/End also do so.
var handle hwnd = getFocus()
if getWindowClass(hwnd) != "TConversationsControl"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
childID = o.accFocus
var int goTop = False
if !childID
	; Probably on the pane object.
	; I haven't found a way to focus the correct item here.  [DGL, 2010-04-25]
	if !stringContains(stringLower(origScript), "home")
		return False
	endIf
	; Make Home try to go to the top of the list.
	; This should make repeated Homes be a fast way to get to an incoming call.
	goTop = True
endIf
var string name = o.accName(childID)
formatStringWithEmbeddedFunctions("<$" +origScript +">")  ; Calls the original script
pause()
var int childID1
o = getFocusObject(childID1)
childID1 = o.accFocus
if childID1 != childID || stringCompare(o.accName(childID1), name, True) != 0
	; We moved, so don't do anything.
	Skype___sayConversationControlPosition(o, childID1)
	return True
endIf
var int delta = -1
if stringContains(stringLower(origScript), "next")
|| stringContains(stringLower(origScript), "end")
	delta = 1
endIf
if goTop
	childID1 = 1
	delta = 1
else
	childID1 = childID1 +delta
endIf
var int safety = 50
var int n = o.accChildCount
while childID1 > 0 && childID1 <= n && safety
	; This detects a "real" contact/conversation entry.
	if o.accRole(childID1) == Role_System_ListItem
	&& !stringIsBlank(o.accValue(childID1))
		o.accSelect(SelFlag_TakeFocus, childID1)
		Skype__sutil__play("qw_click1.wav")
		Skype___sayConversationControlPosition(o, childID1)
		return True
	elif o.accRole(childID1) == Role_System_PushButton
	&& o.accName(childID1) == scHideHistory
	&& delta == 1
		; A way to detect History is showing.
		; We only need this when going down.
		; The above MSAA name/value trick fails to find history items because they have no MSAA value.
		; This moves focus from Live to History.
		typeKey("Alt+3")
		Skype__sutil__play("qw_click1.wav")
		Skype___sayConversationControlPosition(o, childID1)
		return True
	endIf
	childID1 = childID1 +delta
	safety = safety -1
endWhile
if !safety
	Skype__log__write("Loop ran out.")
endIf
Skype___sayConversationControlPosition(o, childID1)
return True
endFunction

int function handleLiveCallVolume(string origScript)
; Handle the Skype 5.1/5.2 Volume button in a Live Call area.
; This makes Up/Down/Home/End set the volume and read the result.
; The slider will briefly appear during this process.
var handle hwnd = getFocus()
if getWindowClass(hwnd) != "TLiveConversation"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
; This may not be necessary but shouldn't hurt.
childID = o.accFocus
var string name = o.accName(childID)
if stringCompare(name, Skype__schat__scVolume, True) != 0
	return False
endIf

; Activate the slider.
speechOff()
typeKey("Space")
pause()
; Call the original script to move the slider.
formatStringWithEmbeddedFunctions("<$" +origScript +">")
pause()
; Refocus the button and thus close the slider.
speechOn()
o.accSelect(SelFlag_TakeFocus, childID)
return True
endFunction

script sayNextLine()
if !cycleThroughMSAAControls("next")
	Skype__saySpecialHighlighting("nextLine", "", "sayNextLine")
endIf
endScript

script sayPriorLine()
if !cycleThroughMSAAControls("prev")
	Skype__saySpecialHighlighting("priorLine", "", "sayPriorLine")
endIf
endScript

script JAWSHome()
if !cycleThroughMSAAControls("first")
	Skype__saySpecialHighlighting("JAWSHome", "", "JAWSHome")
endIf
endScript

script JAWSEnd()
if !cycleThroughMSAAControls("last")
	Skype__saySpecialHighlighting("JAWSEnd", "", "JAWSEnd")
endIf
endScript

script sayWord()
Skype__saySpecialHighlighting("", "sayWord", "sayWord")
endScript

script sayCharacter()
Skype__saySpecialHighlighting("", "sayCharacter", "sayCharacter")
endScript

void function sayLineUnit(int unitMovement, /*?*/ int bMoved)
; Avoid double speech when a list, like contacts or conversations, scrolls on arrowing.
var object o, int childID
if Skype__util__isSpecialFocus() || !isPCCursor()
	sayLineUnit(unitMovement, bMoved)
	return
endIf
o = getFocusObject(childID)
if !o
	sayLineUnit(unitMovement, bMoved)
	return
endIf
if o.accRole(childID) == Role_System_ListItem
	; Let SayObjectActiveItem handle speech for this navigation.
	return
endIf
sayLineUnit(unitMovement, bMoved)
endFunction

void function sayWordUnit(int unitMovement, /*?*/ int bMoved)
if Skype__util__isSpecialFocus() || !isPCCursor()
	; Works around JAWS 14's failure to speak here
	; when focus is in a contact or conversation list.
	sayWord()
	return
endIf
sayWordUnit(unitMovement, bMoved)
endFunction

int function isHotkeysList(object o, int childID)
; Returns True if the Options/Hotkeys list is in focus.
; o and childID are assumed to have come from a getFocusObject() call.
if o.accRole(0) == Role_System_List
&& Skype__getActiveOptionsPageName() == "HotkeysTab"
	return True
endIf
return False
endFunction

int function cycleThroughMSAAControls(string cmd)
; In MSAA windows, cycle among the subcontrols available.
; cmd: next, prev, first, or last, to indicate which way to move.
var
	string wclass,
	handle hwnd,
	object o, int childID, int childID1, int n
if !isPCCursor() || isVirtualPCCursor() || Skype__util__isSpecialFocus()
	return False
endIf
hwnd = getFocus()
wclass = getWindowClass(hwnd)
if stringCompare(wclass, "TAccessibleRadioButtonEx", False) == 0
	return handleRadioButtonFocus(hwnd, cmd)
endIf
if !(
wclass == "TMyselfControl"
|| wclass == "TPanelButtons"
|| wclass == "TZapCommunicator"
|| wclass == "TCallQualityControlHost"
|| wclass == "TConversationToolbar"
|| wclass == "TContactProfile"
|| wclass == "TConversationProfile"
|| wclass == "TLiveConversation"
|| wclass == "TConversationVideo"
|| wclass == "TChatContentControl"
|| wclass == "TConversationsControl"
)
	return False
endIf
o = getFocusObject(childID)
if childID != 0 && (
wclass == "TChatContentControl"
|| wclass == "TConversationsControl"
)
	return False
endIf
if stringCompare(o.accName(childID), Skype__schat__scVolume, True) == 0
	; See saySpecialHighlighting().
	return False
endIf
typeCurrentScriptKey()
pause()
if getFocus() != hwnd
	return True
endIf
if Skype__conv5__applies()
	; Avoids focusing Contact Request button in Skype 5.0 Gold.
	return True
endIf
o = getFocusObject(childID1)
if childID1 != childID
	return True
endIf
n = o.accChildCount
if cmd == "first"
	childID = 1
elif cmd == "last"
	childID = n
elif cmd == "next"
	childID = childID +1
	if childID > n
		childID = 1
	endIf
elif cmd == "prev"
	childID = childID -1
	if childID < 1
		childID = n
	endIf
else
	beep()
	return True
endIf
if childID == childID1
	sayObjectActiveItem()
else
	o.accSelect(SelFlag_TakeFocus, childID)
endIf
return True
endFunction

int function handleRadioButtonFocus(handle hwnd, string cmd)
typeCurrentScriptKey()
pause()
if getFocus() != hwnd
	return True
endIf
hwnd = getParent(hwnd)
var handle hwnd1 = hwnd
if cmd == "next"
	if getNextWindow(hwnd1)
		hwnd1 = getNextWindow(hwnd1)
	else
		hwnd1 = getFirstWindow(hwnd1)
	endIf
elif cmd == "prev"
	if getPriorWindow(hwnd1)
		hwnd1 = getPriorWindow(hwnd1)
	else
		hwnd1 = getLastWindow(hwnd1)
	endIf
elif cmd == "first"
	hwnd1 = getFirstWindow(hwnd1)
else  ; last
	hwnd1 = getLastWindow(hwnd1)
endIf
hwnd1 = getFirstChild(hwnd1)
if hwnd1 == hwnd
	return True
endIf
setFocus(hwnd1)
return True
endFunction

Void Function ActiveItemChangedEvent (handle curHwnd, int curObjectId, int curChildId, handle prevHwnd, int prevObjectId, int prevChildId)

var
	string wclass
wclass = stringLower(getWindowClass(curHwnd))
ActiveItemChangedEvent(curHwnd, curObjectId, curChildId, prevHwnd, prevObjectId, prevChildId)
if wclass == "TInboxList" || stringContains(wclass, "conversation")
	scheduleFunction("Skype__sayContactLine", 10)
endIf
endFunction

void function sayTreeViewItem()
; Says treeView items on arrows, PgUp/Dn, Home/End, etc.
; Does not work before JAWS 7.1.
var
	handle hwnd,
	string name
hwnd = getFocus()
name = ""
if getWindowClass(hwnd) == "TOptionsPanel"
	name = Skype__getActiveOptionsPageName()
	if name
		sayMessage(OT_Line, name)
		return
	endIf
endIf
sayTreeViewItem()
endFunction

void function Skype__sayChatEntry(handle hwnd, int which, string flags)
; Says one chat message.
; Which: positive = 1-based index, -1 = last line, -2 = line above that, -3 = line two above the last, etc.
; Positive indices only supported starting in Skype 3.5.
; Asking for line 0 returns the last line.
; flags:
;	v for virtual buffer.
;	n for new (incoming) message / no special info like "not delivered."
var
	object o, int childID,
	int iRole,
	int nmsgs
flags = stringLower(flags)
if !hwnd
	hwnd = Skype__schat__getWindow(Skype__schat__WOutput, getTopLevelWindow(getFocus()))
endIf
if !which
	which = -1  ; latest line sent
endIf
if stringContains(flags, "v")
	Skype__schat__virtualizeChatMessages(hwnd, which, 100, 0)
	return
endIf

if Skype__conv5__applies()
	o = Skype__conv5__chatContent(hwnd)
else
	Skype__sutil__sayUnsupportedSkype()
	return
endIf
iRole = o.accRole(0)
if !o || (iRole != Role_System_List && iRole != Role_System_Pane)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
nmsgs = Skype__schat__getMessageCount(o)
if which < 0
	which = which +nmsgs +1
endIf
if which < 1 || which > nmsgs
	if nmsgs < 1
		sayMessage(OT_Error, msgNoChatLines)
	else
		sayMessage(OT_Error, msgNotThatManyChatLines)
	endIf
else
	var string msg
	msg = Skype__schat__getMessageText(o, which, flags)
	if stringContains(flags, "v")
		Skype__sutil__sayMessageToUserBuffer(msg, False)
	else
		Skype__sutil__sayURMessage(msg, "", False)
	endIf
endIf  ; valid message number requested
endFunction

script sayAChatLine()
; Attached to alt+1 through alt+0, which read the last message on up respectively.
var
	handle hwnd,
	string keyname,
	int idx,
	string flags
if !Skype__schat__isChatWindow(getFocus(), False)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
hwnd = Skype__schat__getWindow(Skype__schat__WOutput, getTopLevelWindow(getFocus()))
if getFocus() == hwnd && isSameScript()
	; Allows virtualization of any message from the message list window.
	var object o, int msgno
	if Skype__conv5__applies()
		o = Skype__conv5__chatContent(hwnd)
	else
		Skype__sutil__sayUnsupportedSkype()
		return
	endIf
	msgno = Skype__schat__selectedMessage(o)
	Skype__sayChatEntry(hwnd, msgno, "v")
	return
endIf
keyname = getCurrentScriptKeyName()
idx = stringContains("1234567890", stringRight(keyname, 1))
flags = ""
if isSameScript()
	flags = flags +"v"
endIf
Skype__sayChatEntry(hwnd, 0-idx, flags)
endScript

int function Skype___countSelectedChatMessages(object o)
; Return how many messages are selected in the given MSAA list.
var
	int i, int n, int cnt
cnt = 0
n = o.accChildCount
i = 1
while i <= n
	if (o.accState(i) & State_System_Selected)
	&& o.accRole(i) == Role_System_ListItem
		cnt = cnt +1
	endIf
	i = i +1
endWhile
return cnt
endFunction

script sayRelativeChatLine(int which)
; Say the previous, current, next, first, or last chat message.
; If IsSameScript is True and the current message is being requested, the current message and context is virtualized.
; If IsSameScript is True and the first message is being requested, the first unread message is read.
var
	handle hwnd,
	object o, int msgno,
	int msgcount,
	string keyName,
	string flags
if !Skype__schat__isChatWindow(getFocus(), False)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
hwnd = Skype__schat__getWindow(Skype__schat__WOutput, getTopLevelWindow(getFocus()))
if !hwnd
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
flags = ""
if Skype__conv5__applies()
	o = Skype__conv5__chatContent(hwnd)
else
	Skype__sutil__sayUnsupportedSkype()
	return
endIf
if !o
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
msgcount = Skype__schat__getMessageCount(o)
if which == -2 || which == 2
	if which == -2
		msgno = 1
		if isSameScript()
			msgno = Skype___firstUnreadMessage(o)
		endIf
	elif which == 2
		msgno = -1
	endIf
	if msgno > msgcount
		beep()
		return
	endIf
	Skype__sayChatEntry(hwnd, msgno, flags)
	Skype__schat__selectMessage(o, msgno)
	return
endIf
msgno = Skype__schat__selectedMessage(o)
if !msgno
	beep()
	msgno = 1
	Skype__schat__selectMessage(o, msgno)
	pause()
	msgno = Skype__schat__selectedMessage(o)
endIf
if !msgno
	beep()
	return
endIf
msgno = msgno +which
if msgno < 1 || msgno > msgcount
	beep()
	msgno = msgno -which
else
	Skype__schat__selectMessage(o, msgno)
endIf
if which == 0 && !isSameScript()
	var int msgno1, string buf
	msgno1 = Skype__schat__selectedMessage(o)
	if msgno1 > 0
		buf = formatString(msgSelectedChatMessage, intToString(msgno1), intToString(msgcount))
	else
		msgno1 = Skype___countSelectedChatMessages(o)
		buf = formatString(msgSelectedChatMessages, intToString(msgno1), intToString(msgcount))
	endIf
	Skype__sutil__sayURMessage(buf, "", False)
endIf
if isSameScript() && !which
	flags = flags +"v"
endIf
Skype__sayChatEntry(hwnd, msgno, flags)
endScript

int function Skype___firstUnreadMessage(object o)
; Return the index of the first unread message in the indicated message list.
var int i = 1
var int n = o.accChildCount
var string name
var string sNew = " " +msgNew
var int len = stringLength(sNew)
for i = 1 to n
	name = o.accName(i)
	if stringRight(name, len) == sNew
		return i
	endIf
endFor
beep()
return n
endFunction

globals
	handle Skype___lastChatWindow
void function Skype__sayIncomingChat()
if Skype__sutil__getSkypeVersionCode(False) < "v3.5"
	; Work with the last window flagged by the NewTextEvent as having new message text.
	Skype__sayChatEntry(Skype___lastChatWindow, -1, "n")
else
	Skype___checkForNewChats()
endIf
endFunction

Void Function NewTextEvent (handle hwnd, string buffer, int nAttributes,
int nTextColor, int nBackgroundColor, int nEcho, string sFrameName)
var
	int doDefault,
	string strClass,
	int stc

doDefault = False
strClass = StringLower(GetWindowClass (hwnd))
if getScreenEcho() != 1
	; Handle everything in the default manner if screenEcho is None or All.
	doDefault = True
elif strClass == "tty" || strClass == "telnet" || strClass == "console"
	; Works around a bug in JAWS 8/9/10 or so that makes redrawn console text speak when a Skype dialog closes and uncovers it.
	return
elif !Skype__appwin__isSkypeWindow(hwnd) || Skype__appwin__isSkypeNotifyWindow(hwnd)
	; Handle non-Skype windows in the default manner.
	; Handle Skype notifications in the default manner so SkypeWatch can catch them.
	doDefault = True
endIf
if doDefault
	newTextEvent(hwnd, buffer, nAttributes, nTextColor, nBackgroundColor, nEcho, sFrameName)
	return
endIf

; Don't repeat text we've just seen
if !stringCompare(buffer, Skype___strTextBuffer, True)
Return
EndIf

; Speak various messages automatically.
if strClass == "TLoginControl"  ; during login
|| strClass == "TWhatsNewGuide"  ; Getting Started wizard pages
|| ((
	strClass == "TZapCommunicator"  ; Questions for active calls
	|| strClass == "TCallQuality"  ; pop-up caused by pressing the Check Settings button
	|| strClass == "TConversationVideo"  ; Allow/Do Not Allow prompts for video etc.
	) && !Skype___intGotNewMessage  ; new text not caused by a top-level focus change
)
;|| strClass == "TOptionTabCaptionControl"  ; Descriptions of Options pages
	sayMessage(OT_Message, buffer)
	focusPopupIfNecessary(hwnd)
	return
elif getWindowClass(getTopLevelWindow(hwnd)) == "TChangePasswordForm"
&& strClass == "TRichLabel"
	sayMessage(OT_Message, buffer)
	return
endIf

; Speak payment request windows (Skype Prime).
if stringContains(strClass, "premium")
	sayMessage(OT_Message, buffer)
	return
endIf

; Automate speech of banner controls.
if stringContains(strClass, "banner")
	sayMessage(OT_Message, buffer)
	return
endIf

Skype___strTextBuffer = buffer

; only speak conversation text if we're actually in the relevant conversation window, the user wants it, and the window update was not just a "Last message was posted x minutes ago" update.
if (Skype__appwin__isChatWindow(getFocus())
		&& (hwnd != getFocus())
		&& (buffer != "*** ")
		&& hwnd == Skype__schat__getWindow(Skype__schat__WOutput, getFocus())
		&& (Skype__opt__getUserOption(SV_RecognizeChat) == 1))
	if (Skype___intGotNewMessage == 0)
		Skype___lastChatWindow = hwnd
		scheduleFunction("Skype__sayIncomingChat", 5)
		Skype___intGotNewMessage=1
		ScheduleFunction ("Skype___resetGotNewMessage", 20)
	EndIf
	Return
EndIf
Skype___strTextBuffer = buffer
newTextEvent(hwnd, buffer, nAttributes, nTextColor, nBackgroundColor, nEcho, sFrameName)
EndFunction

globals
handle Skype___hwndPopup
void function focusPopupIfNecessary(handle hwnd)
; Focus popup call warning and feedback windows when they appear if apparently appropriate.
; Called from NewTextEvent with a window handle,
; and called internally on a schedule to make the actual focus change.
var
	string strClass
if !hwnd
	; Called from below after a delay.
	if Skype___intGotNewMessage
		; Prevent focus forcing when the focusChangedEvent fires after the newTextEvent on a switch into a conversation.
		return
	endIf
	if getFocus() != Skype___hwndPopup
		setFocus(Skype___hwndPopup)
	endIf
	return
endIf

; Called directly by NewTextEvent.
strClass = getWindowClass(hwnd)
if strClass != "TZapCommunicator"
; Call Quality omitted because it takes focus too often.
; TODO: Should be included but only when newly focusing a conversation doesn't trigger it.
;&& strClass != "TCallQuality"
		; We handle only those window classes just listed here.
		return
	elif Skype___intGotNewMessage
		; Focus just switched to this conversation,
		; which means the NewTextEvent was probably from that rather than from a fresh popup.
		; Don't focus this in that case.
		return
	endIf
	Skype___hwndPopup = hwnd
	scheduleFunction("focusPopupIfNecessary", 5)
endFunction

Void Function sayHighlightedText (handle hwnd, string buffer)
var
	int tc, int stc
; If we're in a special situation, don't play with it.
if Skype__util__isSpecialFocus() || inHJDialog()
	sayHighlightedText(hwnd, buffer)
	return
endIf
; Don't speak highlighting outside of the current top-level window.
; Example:  When someone signs in/out and you're in a conversation but the Contacts tab is also visible on screen,
; JAWS otherwise repeats the currently highlighted Contact List line.
; Exception: In at least the Look In combo box of Send File, JAWS 7.0 and older won't speak during arrowing unless the ComboLBox exception is included here.
if getTopLevelWindow(hwnd) != getTopLevelWindow(getFocus())
&& getWindowClass(hwnd) != "ComboLBox"
	return
endIf
tc = getWindowTypeCode(hwnd)
; Don't speak highlighting in Skype 2.0+ contact lists because lines are spoken via MSAA there.
; Exception for JAWS 7.0 and lower:  ActiveItemChangeEvent doesn't fire there during arrowing.
; WT_MultiLine_Edit added 2010-08-10 because this otherwise broke speaking of text selection, like Ctrl+A.
if tc != WT_Edit && tc != WT_ComboBox && tc != WT_MultiLine_Edit
&& Skype__sutil__isMSAASupport(hwnd)
	return
endIf
; Don't speak non-focused list highlight changes, such as the Hotkey list in Options when the user moves through the options page list to reach that page.
if getWindowTypeCode(hwnd) == WT_ListBox && hwnd != getFocus()
	return
endIf
sayHighlightedText(hwnd, buffer)
EndFunction

Script PickToolBarItem()
var
	string names, string data,
	string curcontact,
	int n, int i
n = Skype__tb__makeList(names, data)
if !n
	sayMessage(OT_Error, msgNoToolbarItems)
	return
endIf

i = dlgSelectItemInList(names, msgToolbarItemSelectionTitle, False)
speechOff()
stopSpeech()
pause()
speechOn()
if !i
	sayMessage(OT_Status, msgCanceled)
else
	Skype__tb__clickItem(stringSegment(data, "|", i))
endIf
pause()
scheduleFunction("sayObjectTypeAndText", 3)
EndScript

script sayTypingStatus()
var
	string status
if !Skype__getTypingStatus(status, 0, isSameScript())
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
if status
	if isSameScript() > 1
		spellString(status)
	else
		Skype__sutil__sayURMessage(status, "", False)
	endIf
	return
endIf
Skype__sutil__sayURMessage(msgNoTyping, "", False)
endScript

script toggleChatMessageTimes()
var
	handle hwnd,
	string msg
hwnd = getFocus()
if !Skype__schat__isChatWindow(hwnd, False)
|| Skype__conv5__focusedListName() == Skype__conv5__ListContacts
	sayCurrentScriptKeyLabel()
	typeCurrentScriptKey()
	return
endIf
msg = Skype__opt__userOptionCycler(SV_ChatMessageTimes, msgChatMessageTimeChoices, False)
msg = msgMessageTimes +" " +msg
sayUsingVoice(VCTX_Message, msg, OT_Status)
endScript

globals
	int Skype___keyCaseSch,
	string Skype___keyCaseID

void function Skype__keyCaseSet(string case)

Skype___keyCaseID = case
if Skype___keyCaseSch
	unscheduleFunction(Skype___keyCaseSch)
	Skype___keyCaseSch = 0
endIf
if !case
	return
endIf
var handle hwnd
hwnd = stringToHandle(case)
JAWSCursor()
moveToWindow(hwnd)
if getCurrentWindow() != hwnd && getParent(getCurrentWindow()) != hwnd
	beep()
endIf
if !Skype__util__setRestrictionToWindow(hwnd)
	beep()
endIf
if !Skype__isRepeatKey()
PCCursor()
endIf
sayMessage(OT_Message, getWindowText(hwnd, Read_Everything))
endFunction

string function Skype__keycaseGet()
return Skype___keyCaseID
endFunction

void function Skype__keyCaseCheck(int direct)
if direct
	Skype___keyCaseSch = scheduleFunction("Skype__keyCaseCheck", 5)
	return
endIf
if Skype__keycaseGet()
	if !isJAWSCursor()
		saveCursor() JAWSCursor()
		setRestriction(RestrictAppWindow)
		restoreCursor()
		Skype__keyCaseSet("")
	elif getCurrentWindow() != stringToHandle(Skype__keycaseGet())
	&& getParent(getCurrentWindow()) != stringToHandle(Skype__keycaseGet())
		beep()
		moveToWindow(stringToHandle(Skype__keycaseGet()))
	endIf
else
endIf
endFunction

Script ReadAccountPanel()
var
	handle hwnd,
	string buf
hwnd = Skype__appwin__getSkypeWindow(SWT_Communicator)
if !hwnd || !isWindowVisible(hwnd) || (getWindowBottom(hwnd) -getWindowTop(hwnd) <= 5)
	hwnd = Skype__appwin__getSkypeWindow(SWT_Myself)
endIf
if hwnd
	;if stringIsBlank(getWindowText(hwnd, Read_Everything))
		refreshWindow(hwnd)
		pause()
		stopSpeech()  ; the refresh otherwise causes a lot of double speech
	;endIf
	Skype__keyCaseSet(intToString(hwnd))
	return
endIf
sayMessage(OT_Error, msgNotMainWindow)
EndScript

Script ScriptFileName ()
ScriptAndAppNames (msgScriptName)
if !isSameScript()
	Skype__sutil__sayURMessage(customInfo(), "", False)
endIf
EndScript

void function addHotkeyLinks()
userBufferAddText(customInfo())
addHotkeyLinks()
endFunction

string function customInfo()
var
	int swStatus,
	string buf
buf = formatString(msgSkypeAppVersionString_L, Script_Version, Script_Revision)
buf = buf +"\13\10"
swStatus = Skype__sutil__isSkypewatchLoaded()
if swStatus == 2
	buf = buf +Skype__msgSWLoaded
elif swStatus == 1
	buf = buf +Skype__msgSWLoaded_old
else
	buf = buf +Skype__msgSWNotLoaded
endIf
return buf
endFunction

script sayAppVersion()
var
	string buf
; The funny length restriction makes versions like "3.0" get ignored so we can get a more specific answer.
; Skype, starting around v3.0.32.171 (and possibly 3.0.0.171), began returning versions like "3.0" for ProductVersion (used by default JAWS script)
; and the full version for FileVersion.  [DGL, 2006-12-03]
if stringLength(getVersionInfoString (getAppFilePath(), "ProductVersion")) < 5
	buf = getVersionInfoString(getAppFilePath(), "ProductName")
	if !buf
		buf = getAppFileName()
	endIf
	buf = formatString(msgSkypeAppVersionFmt, buf, stringChopLeft(Skype__sutil__getSkypeVersionCode(False), 1))
	if isSameScript()
		Skype__sutil__sayMessageToUserBuffer(buf, False)
	else
		Skype__sutil__sayURMessage(buf, "", False)
	endIf
	return
endIf
performScript sayAppVersion()
endScript

int function Skype__sayContactLine()
var handle hwnd = getFocus()
if !isPCCursor() || Skype__util__isSpecialFocus() || !Skype__sutil__isMSAASupport(hwnd)
	return False
endIf
var string wclass = stringLower(getWindowClass(hwnd))

if wclass == "TInboxList" || stringContains(wclass, "conversation")
	var object o
	if Skype__conv5__applies()
		o = Skype__conv5__chatContent(hwnd)
	else
		Skype__sutil__sayUnsupportedSkype()
		return False
	endIf
	if wclass == "TConversationList" || wclass == "TInboxList" || wclass == "TConversationsControl"
		var collection cContact = Skype__sutil__getCurrentContactInfo()
		var string flist
		if !stringCompare(cContact.Name, cContact.ID, True)
			flist = "Time,ExtStat"
		else
			flist = "ID,Time,ExtStat"
		endIf
		var string fld, string val, string line
		forEach fld in cContact
			val = cContact[fld]
			if val
				if !(fld == "Name" || fld == "ID" || fld == "Status")
					line = line +","
				endIf
				line = line +" " +val
			endIf
		endForEach
		sayMessage(OT_Line, line)
	else  ; not historyList or Contacts page
		line = Skype__sutil__removeHTML(o.accHelp(o.accFocus+0))
		if !line
			line = getObjectHelp()
		endIf
		sayMessage(OT_Line, line)
	endIf
	return True
endIf
Skype__log__write("Something got past the MSAA code.")
return False
endFunction

Script SayLine ()
; Makes SayLine read Options/Sounds checkbox list lines more effectively.
; The isSameScript part lets the normal SpellLine behavior work.
if Skype__util__isSpecialFocus() || isSameScript() >= 1 || !isPCCursor()
	performScript sayLine()
return
endIf
performScript sayLine()
Skype__sayCheckState()
EndScript

script switchToTab()
var
	variant null,
	handle hwnd, handle hwndWanted,
	handle hwndCurrent,
	string which,
	object o, int childID,
	int succeeded
Skype__util__waitForKeysToClear(15)
which = stringRight(getCurrentScriptKeyName(), 1)
if !Skype__conv5__applies()
	Skype__sutil__sayUnsupportedSkype()
	return
endIf
var string swt = stringSegment(formatString("%1|%2|%3|%4|%5|%6|%7|%8",
	SWT_ContactList, SWT_ConversationList, SWT_LiveConversations,
	SWT_ChatInput, SWT_ChatOutput, SWT_ChatContact,
	"", SWT_Home
), "|", stringToInt(which))
if !swt
	beep()
	return
endIf

if swt == SWT_ContactList
	typeKey("Alt+1")
	return
elif swt == SWT_ConversationList
	typeKey("Alt+2")
	return
elif swt == SWT_ChatInput
	hwnd = Skype__appwin__getSkypeWindow(swt)
	if hwnd
		if getFocus() != hwnd
			Skype__util__setFocus(hwnd, "d")
		else
			sayObjectTypeAndText()
		endIf
	else
		sayMessage(OT_Error, msgWindowNotFound)
	endIf
	return
elif swt == SWT_ChatContact
	; Profile area, just read except on isSameScript.
	hwnd = Skype__appwin__getSkypeWindow(swt)
	if !hwnd
		sayMessage(OT_Error, msgWindowNotFound)
		return
	endIf
	if !isSameScript()
		sayMessage(OT_User_Requested_Information, getWindowText(hwnd, Read_Everything))
		return
	endIf
	if getFocus() == hwnd
		sayObjectTypeAndText()
	else
		Skype__util__setFocus(hwnd, "d")
	endIf
	return
elif swt == SWT_Home
	speechOff()
	typeKey(ksViewMenu)
	pause()
	typeKey(ksSkypeHome)
	pause()
	scheduleFunction("speechOn", 2)
	scheduleFunction("sayLine", 5)
	return
endIf

; All we have left are Ctrl+3 (live conversations) and Ctrl+5 (chat message list).
o = Skype__appwin__getMSAAObject(swt)
if !o
	sayMessage(OT_Error, msgWindowNotFound)
	return
elif o.accState(childID) & State_System_Invisible
	sayMessage(OT_Error, msgWindowNotVisible)
	return
elif swt == SWT_LiveConversations
&& o.accChildCount == 0
	sayMessage(OT_Error, msgNoLiveConversations)
	return
endIf
if o.accParent.accState(0) & State_System_Focused
	sayObjectTypeAndText()
else
	o.accParent.accSelect(SelFlag_TakeFocus, 0)
endIf
o.accSelect(SelFlag_TakeFocus, 0)
endScript

string function positionInGroup()
; Override of standard JAWS function to make the current position in some combo boxes speak.
; Author:  Doug Lee
var
	handle hwnd,
	string result,
	object o, int childID,
	int i, int n
hwnd = getCurrentWindow()
result = positionInGroup()
; The "navigable" part is for TNavigableTntComboBox windows.
; These occurred in an active call tab as part of the input/output sound control widgets.
if stringLength(result) && !(result == "0 " && getWindowClass(hwnd) == "tNavigable")
	return result
endIf
if getWindowTypeCode(hwnd) == WT_ComboBox || stringContains(stringLower(getWindowClass(hwnd)), "combo")
	; Window.client.child3 (which is the list window) .client (which is the actual list object)
	o = getObjectFromEvent(hwnd, 0-4, 0, childID).accChild(3).accChild(-4)
	if !o
		return ""
	endIf
	i = o.accSelection +0
	n = o.accChildCount
	if i
		return formatString(msgMOfN, intToString(i), intToString(n))
	else
		return formatString(msgItems, intToString(n))
	endIf
endIf
return ""
endFunction

Void Function Skype___resetGotNewMessage ()
Skype___intGotNewMessage=0
EndFunction

int function dialogActive()
; Override to make the result more accurate for Skype.
; Author:  Doug Lee
var
	handle hwnd,
	string wclass
if dialogActive()
	return True
elif Skype__util__isSpecialFocus()
	return False
endIf
hwnd = getTopLevelWindow(getFocus())
wclass = getWindowClass(hwnd)
if stringRight(wclass, 13) == ".UnicodeClass"
	wclass = stringChopRight(wclass, 13)
endIf
if wclass == "Tsk" && stringRight(wclass, 4) == "Form" && wclass != "TskMainForm"
	return True
endIf
return False
endFunction

script Enter()
var
	handle hwnd, handle hwnd1
if Skype__util__isSpecialFocus() || isVirtualPCCursor()
	performScript enter()
	return
endIf
hwnd = getFocus()
typeCurrentScriptKey()
if !dialogActive()
	return
endIf
pause()
if hwnd != getFocus()
	return
endIf
hwnd = getTopLevelWindow(hwnd)
hwnd1 = findWindow(hwnd, "", Skype___scOK)
if !hwnd1
	hwnd1 = findWindow(hwnd, "", Skype___scSave)
endIf
if !hwnd1
	hwnd1 = findWindow(hwnd, "", Skype___scSearch)
endIf
if !hwnd1
	hwnd1 = findWindow(hwnd, "", Skype___scNext)
endIf
if !hwnd1
	hwnd1 = findWindow(hwnd, "", Skype___scFinish)
endIf
if hwnd1
	saveCursor() JAWSCursor() saveCursor()
	moveTo( (getWindowLeft(hwnd1)+getWindowRight(hwnd1))/2, (getWindowTop(hwnd1)+getWindowBottom(hwnd1))/2)
	leftMouseButton()
endIf
endScript

globals
	int Skype___KPELastKey,
	int Skype___KPERepeatCount

int function Skype__isRepeatKey()
; Like isSameScript without a time limit between presses of the same key/combination.
return Skype___KPERepeatCount
endFunction

void function keyPressedEvent(int nKey, string strKeyName, int nIsBrailleKey, int nIsScriptKey)
; Makes the Applications (Context Menu) key work where it should but doesn't already.
; Helps with setting the Skype command key.
; Helps JAWS keep up when Space or Ctrl+Space is pressed in a contact list.
; Makes JAWS speak volume sliders correctly on arrows.
; Fixes a Skype 6.9+ issue where JAWS can't see the content of a chat edit on first message.
; Adds incremental search to the Skype emoticon panel.
if userBufferIsActive() || menusActive()
	keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
	return
endIf
var handle hwnd = getFocus()
var string wclass = getWindowClass(hwnd)
if wclass == "TSmileySelector"
	keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
	Skype__incsearch__gotoNext(strKeyName)
	return
endIf
if handleSkypeCommandKey(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
	return
endIf
if nKey == Skype___KPELastKey
	Skype___KPERepeatCount = Skype___KPERepeatCount +1
else
	Skype___KPELastKey = nKey
	Skype___KPERepeatCount = 0
endIf
keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
Skype__keyCaseCheck(True)
if wclass == "TChatRichEdit"
	scheduleFunction("Skype__cre__manage", 4)
	return
endIf

var int isSpace = !stringCompare(strKeyName, "space", False)
if isSpace && (
	wclass == "TCheckBoxUserListControl"
	|| Skype__live5selector__selected() != 0
)
	saveCursor()  JAWSCursor()  saveCursor()
	routeJAWSToPC()
	leftMouseButton()
	restoreCursor()  restoreCursor()
	scheduleFunction("sayLine", 2)
elif isSpace || !stringCompare(strKeyName, "control+space", False)
	; Works in JAWS 8+ but causes no error messages in earlier versions.
	scheduleFunction("MSAARefresh", 5)
	scheduleFunction("Skype__sayCheckState", 4)
elif (isSpace || stringContains(stringLower(strKeyName), "arrow"))
&& wclass == "TVolumeControl"
	scheduleFunction("stopSpeech", 0)
	scheduleFunction("MSAARefresh", 2)
	scheduleFunction("sayWord", 3)
endIf
if nKey == 93
	scheduleFunction("checkForMenu", 3)
endIf
endFunction

int function handleSkypeCommandKey(int nKey, string strKeyName, int nIsBrailleKey, int nIsScriptKey)
; Handles the KeyPressedEvent part of the Skype command key system.
; Returns True if something was done and False if not.
if Skype__appOpt__giSettingSkypeCommandKey
	if nIsScriptKey || nIsBrailleKey
		; Scripts abort the change because otherwise things like Enter can make quite a mess!
		beep()
	elif stringContains(",esc,enter,tab,space,", ","+stringLower(strKeyName)+",")
		; Disallowed Skype Command keys.
		beep()
		if strKeyName == "Esc"
			; Abort the change.
		else
			; Stay in this mode.
			return
		endIf
	else
		; Write to the .ini from here, where we actually have the key name.
		Skype__opt__setSkypeVar(SV_SkypeCommandKey, strKeyName)
	endIf
	; The Space will cause another jvoSkypeCommandKey() call to be issued by DlgSelectFunctionToRun().
	; That will update the screen to show the new key, and jvoSkypeCommandKey() will wrap up this process.
	typeKey("Space")
	return True
endIf
return False
endFunction

void function checkForMenu()
; Helper for KeyPressedEvent.
if Skype__util__isSpecialFocus()
	return
endIf
if !stringContains(stringLower(getWindowClass(getFocus())), "profile")
	return
endIf
saveCursor() JAWSCursor() saveCursor()
routeJAWSToPC()
pause()
if getTopLevelWindow(getWindowAtPoint(getCursorCol(), getCursorRow())) == getTopLevelWindow(getFocus())
	rightMouseButton()
else
	beep()
endIf
endFunction

script sayBottomLineOfWindow()
; Says the total Skype online user count if possible,
; or otherwise reads the bottom line of the current window as usual.
; A double press does the normal function regardless.
var string s = Skype__conv5__peopleOnline()
if s
	Skype__sutil__sayURMessage(s, "", False)
	return
endIf
var
	handle hwnd,
	int oldRestriction,
	string buf
hwnd = Skype__appwin__getSkypeWindow(SWT_ContactList)
if hwnd && isWindowVisible(hwnd) && !isWindowObscured(hwnd)
	hwnd = getParent(hwnd)
	saveCursor() invisibleCursor() saveCursor()
	moveToWindow(hwnd)
	oldRestriction = getRestriction()
	Skype__util__setRestrictionToWindow(hwnd)
	JAWSPageDown()
	JAWSEnd() JAWSHome()
	buf = getLine()
	Skype__sutil__sayURMessage(buf, "", False)
	setRestriction(oldRestriction)
	return
endIf
performScript sayBottomLineOfWindow()
endScript

string function Skype___shortenFolderPath(object byRef fso, string fname)
; Convert long folder paths (names containing components not adhering to the old DOS 8.3 format) to short (8.3-compatible) paths.
; Used by CompileSkypeScripts to avoid numerous problems with passing space-containing file paths through batch files.
var
	object oFolder
if !fso
	fso = Skype__util__getFSObject()
	if !fso
		; Can't shorten, so return what we got.
		return fname
	endIf
endIf
oFolder = fso.getFolder(fname)
if !oFolder
	return fname
endIf
return oFolder.shortPath
endFunction

script compileSkypeScripts()
; Compile all scripts in this script set and, in JAWS 5.10 and later, reload them and make them take effect.
var
	object fso,
	string compiler, string JAWSDrive, string JAWSDir, string scriptDir,
	string cmd
JAWSDir = Skype___shortenFolderPath(fso, getJAWSDirectory())
scriptDir = Skype___shortenFolderPath(fso, getJAWSSettingsDirectory())
compiler = scriptDir +"\\SkypeComp.bat"
JAWSDrive = substring(compiler, 1, 2)
JAWSDir = stringChopLeft(JAWSDir, 2)
scriptDir = stringChopLeft(scriptDir, 2)
cmd = compiler +" " +JAWSDrive +" " +JAWSDir +" " +scriptDir
run(cmd)
scheduleFunction("reloadAllConfigs", 50)
endScript

Script ReadBoxInTabOrder ()
; Makes JAWS read text more effectively in Skype dialog boxes.
If !InHJDialog ()
	SayWindow(GetTopLevelWindow (GetFocus ()), Read_Everything)
	return
EndIf
PerformScript ReadBoxInTabOrder()
EndScript

script leftMouseButton()
if isPCCursor()
	saveCursor() JAWSCursor() saveCursor()
	routeJAWSToPC()
endIf
performScript leftMouseButton()
endScript

script rightMouseButton()
if isPCCursor()
	saveCursor() JAWSCursor() saveCursor()
	routeJAWSToPC()
endIf
performScript rightMouseButton()
endScript

script sayAChatName(int which)
; Say the name of a conversation, or list the names if which == 0.
Skype__schat__switchToChat(2, which)
endScript

script sayAllChatNames()
; Says the names of all open conversations.
Skype__schat__switchToChat(2, 0)
endScript

script switchToChat(int which)
; Switch to a specific conversation by index, or bring up a list for selection by passing 0.
Skype__schat__switchToChat(1, which)
endScript

void function Skype___refreshMSAAInfo()
var
	handle hwnd
if Skype__util__isSpecialFocus()
	return
endIf
hwnd = getFocus()
if !isMSAAWindow(hwnd)
	return
endIf
if getWindowClass(getParent(hwnd)) == "TConversationsControl"
&& stringContains(stringLower(getWindowClass(hwnd)), "list")
	if getWindowClass(hwnd) == "THistoryList"
		MSAARefresh()
	endIf
endIf
endFunction

void function sayObjectActiveItem(int bSayPosition)
Skype___refreshMSAAInfo()
sayObjectActiveItem(bSayPosition)
endFunction

void function ObjStateChangedEvent(handle hObj, int iObjType, int nChangedState, int nState, int nOldState)
; Block extra repeats of History List items during arrowing.
if !Skype__util__isSpecialFocus() && isPCCursor()
&& getWindowClass(getFocus()) == "THistoryList"
	return
endIf
ObjStateChangedEvent(hObj, iObjType, nChangedState, nState, nOldState)
endFunction

script toggleDialPad()
; Toggle the DialPad on/off and focus it when on.
var
	handle hwnd,
	object o, int childID,
	int left, int right, int top, int bottom,
	int x, int y
hwnd = findWindow(Skype__appwin__getSkypeWindow(SWT_ChatMain), "TLiveConversation", "")
if hwnd && getWindowClass(getFirstChild(hwnd)) == "TLiveConversation"
	hwnd = getFirstChild(hwnd)
endIf
if !hwnd || !isWindowVisible(hwnd) || !isWindowVisible(getParent(hwnd))
	sayMessage(OT_Error, msgCantFindCT)
	return
endIf
o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	sayMessage(OT_Error, msgCantGetCT)
	return
endIf
childID = Skype__MSAA__childByName(o, scCallToolBar)
if !childID
	sayMessage(OT_Error, msgNoDialPadButton)
	return
endIf
o = o.accChild(childID)
childID = Skype__MSAA__childByName(o, scDialPad)
if !childID
	sayMessage(OT_Error, msgNoDialPadButton)
	return
endIf
if !Skype__MSAA__rect(o, childID, left, right, top, bottom)
	sayMessage(OT_Error, msgCantFindDPS)
	return
endIf
x = (left+right)/2
y = (top+bottom)/2
speechOff()
Skype__winclick__send(hwnd, "left", x, y)
pause()
speechOn()
pause()
o = o.accParent
var int isOpen = False
if Skype__MSAA__childByName(o, scLiveDialPad)
	isOpen = True
endIf
if isOpen
	sayMessage(OT_Status, msgDPActive)
else
	sayMessage(OT_Status, msgDPClosed)
endIf
endScript

script JAWSDelete()
; Makes Delete work in lists where it otherwise deletes what's selected instead of what's in focus.
; At this writing, that's the Contact list and the Conversation list.
var
	string wclass,
	object o, int childID
wclass = getWindowClass(getFocus())
if Skype__util__isSpecialFocus() || (wclass != "TConversationList" && wclass != "TInboxList")
	performScript JAWSDelete()
	return
endIf
saveCursor()  PCCursor()
o = Skype__sutil__getFocusObject(childID)
if !(o.accState(childID) & State_System_Selected)
	typeKey("Space")
	pause()
endIf
performScript JAWSDelete()
endScript

script InsertEmoticon()
var handle hwnd = getFocus()
var string wclass = getWindowClass(hwnd)
if wclass != "TChatRichEdit"
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
hwnd = getPriorWindow(hwnd)
if getWindowClass(hwnd) != "TSmileyButtonPanel"
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
var object o = Skype__MSAA__objectFromWindow(hwnd)
if o.accChildCount != 1
|| o.accRole(1) != Role_System_PushButton
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
o.accSelect(SelFlag_TakeFocus, 1)
pause()
if o.accFocus() != 1
	beep()
	return
endIf
typeKey("space")
endScript

int function Skype__hasUnreadContent(object o, int childID)
; Return True if the given MSAA item is a "new" conversation.
; Assumes the item is a conversation list entry.
; "New" means "contains something unread."
; New conversations' entries are taller than read ones.
if Skype__sutil__getMSAAItemDim(o, childID, "h") > 30
	return True
endIf
return False
endFunction

void function sayCharacterUnit(int unit)
if getObjectSubtypeCode() == WT_LeftRightSlider
	MSAARefresh()
	pause()
endIf
sayCharacterUnit(unit)
endFunction

int function caretVisible()
; For some reason, the Skype 4 contact list reports caretVisible() True briefly after an up/down navigation that scrolls the list.
; Symptom:  JAWS saying "blank" or reading the just-left contact line before reading the newly selected one.
; This little intercept is to stop that.
if !Skype__util__isSpecialFocus() && isPCCursor()
&& stringContains(stringLower(getWindowClass(getFocus())), "list")
	return False
endIf
return caretVisible()
endFunction

script CloseSkype()
; Closes Skype without causing it to prompt for confirmation, for example, to confirm permission to close an active call.
; This is a convenience over having to close Skype from the system tray.
var
	handle hwnd
hwnd = Skype__appwin__getSkypeWindow(SWT_Main)
if !hwnd
	sayMessage(OT_Error, msgNoMainWindow)
	return
endIf
if !isSameScript()
	sayMessage(OT_User_Requested_Information, msgCloseSkypeConfirmation)
	return
endIf
; Heads up:  This will close Skype instantly, with all calls, file transfers, etc.
postMessage(hwnd, 0x10, 0, 0)  ; 0x10 is WM_Close
endScript

script upALevel()
; Close a Skype conversation without removing its entry from the Conversation list.
; Alt+F4 does this in Compact view, but no key does it natively in Default view:
; Ctrl+F4 removes from the Conversations list, Alt+F4 closes Skype's main window,
; and Esc does nothing.
var
	handle hwnd
hwnd = getFocus()
if Skype__util__isSpecialFocus() || !Skype__schat__isChatWindow(hwnd, False)
	performScript upALevel()
	return
endIf
while hwnd
	if getWindowClass(hwnd) == "TConversationForm" && getParent(hwnd)
		; Non top-level conversation; Default view.
		postMessage(hwnd, 0x10, 0, 0)  ; WM_Close
		return
	endIf
	hwnd = getParent(hwnd)
endWhile
; Not in a Default-view conversation, so just do what the key would have done.
performScript upALevel()
endScript

handle function appGetWindow(string which)
; Test harness compatibility; standard name for this function.
return Skype__appwin__getSkypeWindow(which)
endFunction

script sayWhoIsActive()
; Say who is in a multiperson voice conference.
Skype__sayConferenceParticipants(False)
endScript

script sayWhoIsSpeaking()
; Say who in a multi-person conference is currently speaking.
; Technically this really says who is currently making the most noise.
Skype__sayConferenceParticipants(True)
endScript

void function Skype__sayConferenceParticipants(int justSpeakers)
; Say who in a multi-person conference is currently participating or speaking.
var collection speakers = Skype__conf__getParticipants(justSpeakers)
var int err = speakers.err
if err
	if err == 1
		sayMessage(OT_Error, msgNotInChatWindow)
	elif err == 2
		sayMessage(OT_Error, msgNotInLiveChat)
	elif err == 3
		sayMessage(OT_Error, msgNotInConference)
	else
		sayMessage(OT_Error, formatString(msgUnknownError, intToString(err)))
	endIf
	return
endIf

var int i, collection speaker
var string buf = ""
for i = 1 to speakers.count
	speaker = speakers[intToString(i)]
	if buf
		buf = buf +" + "
	endIf
	buf = formatString("%1%2", buf, speaker.name)
endFor
if !buf
	buf =  msgNobodySpeaking
endIf
;buf = buf + " " +intToString(speakers.tc)
sayMessage(OT_User_Requested_Information, buf)
endFunction

; Interceptions of JAWS 15 scripts, to prevent "Unknown script call" messages.
script readHistoryItem()
sayCurrentScriptKeyLabel()
typeCurrentScriptKey()
endScript


;======================== Module schat (from SkypeChat) ========================
; Subsystem for handling Skype conversation windows and their contents.
; Interface:
;	Skype__schat__isChatWindow(hwnd, restricted)  --> lWellIsIt:  Indicates if the given handle is part of a valid conversation window structure.
;	Skype__schat__getChatName(hwnd[, includeStatus]) --> sChatName:  Gets a speakable/printable name for a conversation from its handle.
;	Skype__schat__getChatWindows() --> cChats:  Gets a collection of all open conversation window handles.
;	Skype__schat__switchToChat(int absolute, int delta) --> lSuccess:  Name, list, or move to a conversation.
;	Skype__schat__getWindow(hChatTop, sWhich) --> hwnd:  Gets a window by SKCW_* identifier from a given conversation.
;	Skype__schat__virtualizeChatMessages(hwnd, start, contextSize, contextPos) --> lSuccess: Virtualizes conversation messages with context.
;
; Author:  Doug Lee

const
; Link types for Skype__schat___addMessageToUserBuffer and supporting code.
	Skype__schat___LTMsg = 1,  ; The whole message is a single link ("Continue last conversation with ...").
	Skype__schat___LTHistory = 2,  ; "Show messages from: ..." history links.
	Skype__schat___LTURL = 3,  ; URLs in user-typed messages.
	Skype__schat___LTPlay = 4   ; A voicemail notification terminated by a Play link.

int function Skype__schat__isCompactView()
; Returns True if Skype is in Compact view and False if it is in Default view.
var
	handle hwnd
hwnd = Skype__appwin__getSkypeWindow(SWT_Main)
if hwnd
	hwnd = findWindow(hwnd, "TSkypeSplitter", "")
	if hwnd
		return !isWindowVisible(hwnd)
	endIf
endIf
; TODO: We can't determine the view, so Default is assumed here.
beep()
Skype__log__write("Unable to determine the active view.")
return False
endFunction

handle function Skype__schat__getTopOfChat(handle hwnd0)
; Get the top-level window of the conversation containing the given window.
; In Compact view, this will actually be a top-level window,
; but in Default view, it will be the TConversationForm containing the conversation,
; which will still be a descendant of the main Skype window.
var
	handle hwnd,
	handle hTop,
	variant null
hTop = getTopLevelWindow(hwnd0)
hwnd = hwnd0
while hwnd
	if getWindowClass(hwnd) == "TConversationForm"
		return hwnd
	endIf
	if hwnd == hTop
		hwnd = null
	else
		hwnd = getParent(hwnd)
	endIf
endWhile
; For Skype 4 Default view, in which exactly 0 or 1 conversations can be active.
; This means any window in the Skype default view is considered part of a conversation,
; as long as a conversation is actually showing.
hwnd = Skype__appwin__getActiveChatWindow(hTop)
if hwnd
	return hwnd
endIf
return 0
endFunction

int function Skype__schat__isChatWindow(handle hwnd, int restricted)
; Returns True if the given window is part of a Skype conversation window structure.
; Restricted: True if only windows inside a conversation should be allowed to return True.  Useful in Default view.
var
	handle hTop
hTop = Skype__schat__getTopOfChat(hwnd)
if !hTop
	return False
elif !restricted
	return True
endIf
while hwnd
	if hwnd == hTop
		return True
	endIf
	hwnd = getParent(hwnd)
endWhile
return False
endFunction

globals
; Cache to minimize Conversation List scans below.
	object Skype__schat___dvobj,
	int Skype__schat___dvid
int function Skype__schat___findMatchingCLEntry(string whichList, string wantedName)
var
	object o, int childID, int n
o = Skype__appwin__getMSAAObject(whichList)
if !o
	return False
endIf
n = o.accChildCount
childID = 1
while childID <= n
	if !stringCompare(wantedName, o.accName(childID), True)
		Skype__schat___dvobj = o
		Skype__schat___dvid = childID
		return True
	endIf
	childID = childID +1
endWhile
return False
endFunction

object function Skype__schat___getConversationObjectAndID(handle hwnd, int byRef childID)
; Get the Conversation List entry (MSAA object and childID) corresponding to the given conversation window.
var
	variant null,
	string wantedName
wantedName = getWindowName(hwnd)
if !stringCompare(wantedName, Skype__schat___dvobj.accName(Skype__schat___dvid), True)
	childID = Skype__schat___dvid
	return Skype__schat___dvobj
endIf
Skype__schat___dvobj = null
Skype__schat___dvid = 0
childID = 0
if Skype__schat___findMatchingCLEntry(SWT_ConversationList, wantedName)
|| Skype__schat___findMatchingCLEntry(SWT_LiveConversations, wantedName)
	childID = Skype__schat___dvid
	return Skype__schat___dvobj
endIf
childID = 0
return null
endFunction

int function Skype__schat__isLiveConversation(handle hwnd)
; Returns True if the given call represents a live conversation.
var
	object o, int childID
hwnd = Skype__schat__getTopOfChat(hwnd)
if !hwnd
	return False
endIf
o = Skype__schat___getConversationObjectAndID(hwnd, childID)
if !o
	return False
endIf
return (o.accName(0) == Skype__schat__scLive)
endFunction

string function Skype__schat__getChatName(handle hwnd, int includeStatus)
; Returns the name of the conversation indicated by the given window handle.
; Multichats tend to get a message line as a name.
var
	string name,
	int pos,
	object o, int childID, string buf
hwnd = Skype__schat__getTopOfChat(hwnd)
if !hwnd
	return ""
endIf
name = getWindowName(hwnd)
pos = stringContainsFromRight(name, ")")
if pos
	name = stringLeft(name, pos)
endIf
if !includeStatus
	return name
endIf
o = Skype__schat___getConversationObjectAndID(hwnd, childID)
if !o
	return name
endIf
buf = stringTrimLeadingBlanks(stringSegment(o.accDescription(childID), ",", 1))
if buf
	name = formatString("%1 (%2)", name, buf)
endIf
return name
endFunction

collection function Skype__schat___getChatWindows(handle hwnd, int requireOneVisible)
; Helper for Skype__schat__getChatWindows.
var
	handle hwndVisible,
	int compactView,
	collection cChats cChats = new collection
compactView = !(getParent(hwnd))
while hwnd
	if (!compactView || isWindowVisible(hwnd)) && Skype__schat___isAvailableConversation(hwnd)
		if isWindowVisible(hwnd)
			hwndVisible = hwnd
		endIf
		cChats[intToString(hwnd)] = hwnd
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
if !hwndVisible && requireOneVisible
	; If none are visible (can happen in Default view), there really aren't any available conversations.
	return ""
endIf
return cChats
endFunction

collection function Skype__schat__getChatWindows()
; Returns a collection of all the conversation windows currently open.
; The list is guaranteed to be in a stable order from one call to the next; however, this order is not specifically defined.
var
	handle hwnd,
	handle hTop,
	collection cChats
; If the empty canvas is visible, Default view is active and there is no current conversation.
hTop = Skype__appwin__getSkypeWindow(SWT_Main)
if isWindowVisible(findWindow(hTop, "TEmptyCanvas", ""))
	hwnd = findWindow(hTop, "TConversationForm", "")
	if !hwnd
		hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
	endIf
	if hwnd
		cChats = Skype__schat___getChatWindows(hwnd, False)
		return cChats
	endIf
endIf
; Try Compact view next because a visible conversation window at top level definitively identifies that this view is active.
cChats = new collection
hwnd = findTopLevelWindow("TConversationForm", "")
if !hwnd
	hwnd = findTopLevelWindow("TConversationForm.UnicodeClass", "")
endIf
if hwnd
	cChats = Skype__schat___getChatWindows(hwnd, True)
	if collectionItemCount(cChats) > 0
		return cChats
	endIf
endIf
; Try Default view if Compact view is not active.
cChats = new collection
hwnd = findWindow(hTop, "TConversationForm", "")
if !hwnd
	hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
endIf
if hwnd
	cChats = Skype__schat___getChatWindows(hwnd, False)
endIf
return cChats
endFunction

variant function Skype__schat__getWindow(string which, handle hTop)
; Returns the requested window from under hTop, which should be the top-level window of a Skype conversation, though it can actually be any window in the conversation.
; If which is "?" the list of understood window names is returned.  This is for testing.
var
	handle hwnd
hwnd = Skype__schat__getTopOfChat(hTop)
if !hwnd
	hwnd = Skype__appwin__getActiveChatWindow(hTop)
endIf
if !hwnd
	return 0
endIf
hTop = hwnd
if which == "?"
	return Skype__schat__WMain +"|" +Skype__schat__WInput +"|" +Skype__schat__WOutput +"|" +Skype__schat__WTopic +"|" +Skype__schat__WUserList
	+"|" +Skype__schat__WConvToolBar +"|" +Skype__schat__wContactArea
elif which == Skype__schat__WMain
	return hTop
elif which == Skype__schat__WInput
	var handle hwnd1
	hwnd1 = findWindow(hTop, "TChatEntryControl", "")
	if hwnd1
		return findWindow(hwnd1, "TChatRichEdit", "")
		|| findWindow(hwnd1, "TChatRichEdit.UnicodeClass", "")
	endIf
elif which == Skype__schat__WOutput
	hwnd = findWindow(hTop, "TChatContentControl", "")
	if hwnd
		return hwnd
	endIf
	hwnd = Skype__winfinder__findWindowByClassSubstring(hTop, "mecontrol")
	if !hwnd || !Skype__winfinder__findWindowByClassSubstring(getFirstWindow(hwnd), "richview")
		return 0
	endIf
	hwnd = Skype__winfinder__findWindowByClassSubstring(getFirstWindow(hwnd), "richview")
	if hwnd
		return hwnd
	endIf
	hwnd = getFirstChild(hTop)
	while hwnd && (getWindowClass(hwnd) != "TPanel" || !findWindow(hwnd, "", "pnlSmileys"))
		hwnd = getNextWindow(hwnd)
	endWhile
	if hwnd
		hwnd = findWindow(hwnd, "TRichView", "")
		return hwnd
	endIf
elif which == Skype__schat__WTopic
	return findWindow(findWindow(hTop, "", "pnlTopic"), "TRichView", "")
elif which == Skype__schat__WUserList
	return findWindow(hTop, "TChatUserList", "")
elif which == Skype__schat__WConvToolBar
	return findWindow(hTop, "TConversationToolbar", "")
elif which == Skype__schat__WContactArea
		hwnd = findWindow(hTop, "TContactProfile", "")
		if !hwnd
			hwnd = findWindow(hTop, "TConversationProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hTop, "TParticipantList", "")
		endIf
		return hwnd
else
	Skype__log__write(formatString(msgUnknownWindow, which), True)
endIf
endFunction

int function Skype__schat___isAvailableConversation(handle hwnd)
; Returns True if hwnd is a valid and either active or activatable conversation window.
; "Available" means it has a corresponding Conversation List entry.
; Skype 4 can leave invisible conversation windows around that can't be activated.
var
	int compactView,
	object o, int childID
if getWindowClass(hwnd) != "TConversationForm"
	; Rule out windows with the wrong class.
	return False
endIf
compactView = !getParent(hwnd)
if compactView && !isWindowVisible(hwnd)
	; Rule out invisible windows unless in Default view.
	return False
endIf
if !Skype__schat__isChatWindow(hwnd, False)
	return False
endIf
if isWindowVisible(hwnd)
	; By now, it's valid if it's visible.
	; This allows default-view conversations without corresponding Conversation List entries to be valid if visible.
	return True
endIf
if compactView
	return False
endIf
o = Skype__schat___getConversationObjectAndID(hwnd, childID)
if o
	return True
endIf
return False
endFunction

int function Skype__schat___focusDefaultViewConversation(handle hwnd)
; Focus the given Default view conversation window by clicking its corresponding Conversation list entry.
; Intended for use only in Default view, where conversations all descend from the main Skype window.
var
	handle hwndInbox,
	object o, int childID
; First make sure the Conversation List is visible.
if Skype__conv5__applies()
	o = Skype__appwin__getMSAAObject(SWT_ConversationList)
	if o.accState(0) & State_System_Invisible || !o
		typeKey("Alt+2")
		pause()
		stopSpeech()
		o = Skype__appwin__getMSAAObject(SWT_ConversationList)
		if o.accState(0) & State_System_Invisible || !o
			beep()
			return False
		endIf
	endIf
else
	Skype__sutil__sayUnsupportedSkype()
	return False
endIf
o = Skype__schat___getConversationObjectAndID(hwnd, childID)
if !o
	return False
endIf
; Fire this next to make sure the entry is visible and its text present on screen.
; This will not actually switch focus to the Conversation List window.
o.accSelect(SelFlag_TakeFocus, childID)
pause()
if !Skype__MSAA__click(o, childID)
	beep()
	return False
endIf
pause()
hwnd = Skype__schat__getWindow(Skype__schat__WInput, hwnd)
if getFocus() != hwnd
	stopSpeech()
	setFocus(hwnd)
endIf
return True
endFunction

int function Skype__schat__switchToChat(int absolute, int delta)
; Switch forward or backward by delta conversations from the current one, say or show a list of conversations, or say the name of a conversation:
;	absolute == 0 or 1, delta == 0:  Show a list of conversations for selection.
;	absolute == 0, delta != 0:  Move delta (plus or minus) conversations from the current one in round-robin fashion.
;	absolute == 1, delta != 0:  Move to a specific conversation (negative counts from end).
;	absolute == 2, delta == 0:  Say a list of conversations.
;	absolute == 2, delta != 0:  Say the name of a specific conversation (negative counts from end).
; If moving relatively and focus is not currently in a conversation, a delta of 1 will go to the first conversation, -1 to the last, 0 to a list with the first as default, and others logically from this specification.
; Returns True on success and False on failure or user cancellation.
var
	collection cChats, int nchats, int i,
	handle hwnd,
	string chatNames, string name,  ; used when building a list of conversations to show the user
	string desc,
	int curPos

; Collect conversation window handles, a count of them, and the current position among them, and make sure delta is in bounds.
cChats = Skype__schat__getChatWindows()
nchats = collectionItemCount(cChats)
if !nchats
	sayMessage(OT_Error, msgNoChatsOpen)
	return False
endIf
var handleArray aChats aChats = new handleArray[nchats]
var string shwnd
i = 0
forEach shwnd in cChats
	i = i +1
	aChats[i] = stringToHandle(shwnd)
endForEach

if absolute == 1 && !delta
	absolute = 0
endIf
if absolute
	curpos = delta
	if !curpos
		Skype__sutil__sayURMessage(
			Skype__string__quantify(formatString(msgChatsOpen_L, intToString(nchats))),
			Skype__string__quantify(formatString(msgChatsOpen_S, intToString(nchats))),
		False)
		forEach shwnd in cChats
			hwnd = stringToHandle(shwnd)
			name = Skype__schat__getChatName(hwnd, False)
			Skype__string__replace(name, "|", " ", "ect")
			desc = Skype__schat__convDescription(hwnd)
			if desc
				name = name +" (" +desc +")"
			endIf
			Skype__sutil__sayURMessage(name, "", False)
		endForEach
		return True
	elif curpos < 0
		curpos = nchats +curpos +1
	endIf
	if curpos < 1 || curpos > nchats
		sayMessage(OT_Error, msgNotThatManyChats_L, msgNotThatManyChats_S)
		return False
	endIf
	if absolute == 2
		name = Skype__schat__getChatName(aChats[curpos], False)
		Skype__string__replace(name, "|", " ", "ect")
		Skype__sutil__sayURMessage(name, "", False)
		  return True
	endIf
else  ; not absolute
	var int wantList
	wantList = !delta
	var handle curHwnd = Skype__schat__getTopOfChat(getFocus())
	curPos = nchats
	while curPos && aChats[curpos] != curHwnd
		curPos = curPos -1
	endWhile
	if !curPos && delta < 0
		curPos = 1
	endIf
	delta = delta %nchats  ; guarantees reasonable bounds in case somebody does something crazy like delta==125

	; Set curPos to be the conversation we want to go to next.
	if wantList
		; Show list of conversations, starting on current one, and set curPos from the user's choice.
		i = 1
		while i <= nchats
			hwnd = aChats[i]
			name = Skype__schat__getChatName(hwnd, False)
			Skype__string__replace(name, "|", " ", "ect")
			desc = Skype__schat__convDescription(hwnd)
			if desc
				name = name +" (" +desc +")"
			endIf
			chatNames = Skype__string__join(chatNames, "|", name)
			i = i +1
		endWhile
		curPos = DLGSelectItemInList(chatNames, msgSelectAChat, False, curPos)
		if !curPos
			sayMessage(OT_Message, msgChatSelectCanceled)
			return False
		endIf
	else
		; Round-robin ring of conversations...
		curPos = curPos +delta
		if curPos < 1
			curPos = curPos +nchats
		elif curPos > nchats
			curPos = curPos -nchats
		endIf
	endIf
endIf  ; absolute

; Switch to the indicated conversation.
hwnd = aChats[curpos]
if !hwnd
	Skype__log__write(msgChatSwitchFailed, True)
	Skype__log__write("Chat window not found", True)
	return False
endIf
var handle hwndInput
hwndInput = Skype__schat__getWindow(Skype__schat__WInput, hwnd)
if getFocus() == hwndInput
	if nchats == 1
		sayMessage(OT_Message, msgOnlyChat)
	endIf
	sayObjectTypeAndText()
elif Skype__sutil__isAncestorOf(getFocus(), hwnd)
	; SetFocus should work within the current conversation.
	setFocus(hwndInput)
else
	; This is a conversation switch, and setFocus may need a little help.
	; In default view, a lot of help.
	if hwnd == getTopLevelWindow(hwnd)
		; Compact view:  Conversations are separate Windows tasks.
		; We first have to focus and possibly unobscure the conversation's main window.
		if Skype__schat___windowIsFullyVisible(hwnd)
			setFocus(hwnd)
		else
			if !Skype__schat___focusViaTaskButton(hwnd)
				beep()
				return False
			endIf
		endIf
		pause()
		if !getFocus()
			; The switched-to window is minimized, so unminimize it.
			typeKey("Enter")
		endIf
		if getFocus() != hwndInput
			setFocus(hwndInput)
		endIf
	else
		; Default view:  Conversations are all under the Skype main window.
		; The new one will be invisible, and it can only be made visible by clicking its corresponding Conversation List entry.
		if !Skype__schat___focusDefaultViewConversation(hwnd)
			Skype__log__write(msgChatSwitchFailed, True)
			Skype__log__write("Default-view focus failed.", True)
			return False
		endIf
		if getFocus() != hwndInput
			setFocus(hwndInput)
		endIf
	endIf
endIf
return True
endFunction

int function Skype__schat___windowIsFullyVisible(handle hwnd)
; Return True if hwnd is really visible according to MSAA.
; Chats that pop up because of an incoming message tend to return True for isWindowVisible(hwnd)
; but still be invisible in MSAA.
; SetFocus will not work to focus these under JAWS 11.
; This function actually uses a shortcut that may be Skype specific:
; If the window is completely off screen, we assume it's invisible.
if getWindowRight(hwnd) < 0 || getWindowTop(hwnd) < 0
	return False
endIf
return True
endFunction

int function Skype__schat___focusViaTaskButton(handle hwnd)
; Attempt to set focus to hwnd by clicking its corresponding taskbar button.
var string name = getWindowName(getTopLevelWindow(hwnd))
hwnd = Skype__util__getTaskBarWindow()
if !hwnd
	return False
endIf
if getWindowTypeCode(getFirstChild(hwnd)) == WT_ToolBar
	hwnd = getFirstChild(hwnd)
endIf
var object o, int childID
o = getObjectFromEvent(hwnd, -4, 0, childID)  ; -4 == ObjID_Client
if !o
	return False
endIf
childID = Skype__MSAA__childByName(o, name, "c")
if childID
	o.accDoDefaultAction(childID)
	return True
endIf
return False
endFunction

int function Skype__schat__isSameText(string s1, string s2)
; Stripped-blank text comparison.
s1 = stringReplaceChars(s1, "\10\13\160", " ")
s2 = stringReplaceChars(s2, "\10\13\160", " ")
return s1 == s2
endFunction

int function Skype__schat__getTypist(object o, int i, string byRef typist)
; Returns by reference the typist indicated by the given MSAA element.
; The typist may be empty if the given element does not represent a typist.
; Returns False on encountering a "real" chat message and True until then.
; This function is meant to be called from a loop that travels through chat messages in reverse from the end.
var
	string name, string desc
typist = ""
; ChildID 1 is not allowed to be a typist because either:
;	(1) It's the "Show messages from" entry, or
;	(2) This is a brand-new incoming IM window, which had to be opened by an IM that should be here, or
;	(3) This is a brand-new outgoing IM, in which case you're typing and they're not yet.
; For case (1), the date/time stamp is wrong just as for typing indicators,
; and we use that funny date as a heuristic below (thanks Aaron Smith of GW Micro for this idea),
; which is why it's worth making this exclusion now.
; Non listItems such as buttons and links are excluded for more obvious reasons.
if i <= 1 || o.accRole(i) != Role_System_ListItem
	return False
endIf
var int nState = o.accState(i)
if !(nState & State_System_Unavailable)
	return False
endIf
desc = stringTrimLeadingBlanks(o.accDescription(i))
if desc
	return False
endIf
name = stringTrimLeadingBlanks(o.accName(i))
if !name
	return True
endIf
if stringRight(name, stringLength(Skype__schat___scTyping)) != Skype__schat___scTyping
	return False
endIf
if nState & State_System_OffScreen
	return True
endIf
name = stringChopRight(name, stringLength(Skype__schat___scTyping))
typist = name
return True
endFunction

int function Skype__schat___yearDifference(string desc)
; Returns the absolute value of the difference between the current year and the one in desc.
; If there is an error, returns -1.
; Used to help distinguish between typing indicators and regular messages.
; Idea from Aaron Smith of GW Micro, 2009-09-18.
var
	int curYear, int messageYear,
	int diff,
	int pos
curYear = stringToInt(sysGetDate("yyyy"))
if !curYear
	return -1
endIf
pos = stringContainsFromRight(desc, " ")
if !pos
	return -1
endIf
messageYear = stringToInt(stringChopLeft(desc, pos))
if messageYear < 2000
	return -1
endIf
diff = messageYear -curYear
if diff < 0
	diff = 0 -diff
endIf
return diff
endFunction

int function Skype__schat__selectMessage(object o, int msgno)
; Select the indicated message.
; msgno: 1-based (positive) message number, or negative number indicating a message relative to the last one.
; -1 is the last message, -2 the next to last, etc.
; 0 is not a valid value of msgno.
var int n = Skype__schat__getMessageCount(o)
if msgno < 0
	msgno = n +1 +msgno
endIf
o.accSelect(SelFlag_TakeSelection, msgno)
	return True
endFunction

int function Skype__schat___selectedMessageID(object o)
; Indicate the MSAA ChildID of the selected chat message.
var int childID = o.accSelection +0
if childID
	return childID
endIf
childID = 1
var int n = o.accChildCount
while childID <= n
	if (o.accState(childID) & State_System_Selected)
	&& o.accRole(childID) == Role_System_ListItem
		return childID
	endIf
	childID = childID +1
endWhile
; Default result if nothing is selected.
return 1
endFunction

int function Skype__schat__selectedMessage(object o)
; Indicate the 1-based message number of the selected chat message.
return Skype__schat___selectedMessageID(o)
endFunction

int function Skype__schat__focusedMessageID(object o)
; Indicate the number of the focused chat message.
; If MSAA doesn't indicate which message is focused, the last "real" message is returned.
; "Real" means not a typist.
var int childID = o.accFocus +0
if childID
	return childID
endIf
if o.accRole(2) != Role_System_ListItem
	return 1
endIf
return Skype__schat__getMessageCount(o)
endFunction

int function Skype__schat__getMessageCount(object o)
; Return the number of "real" messages in the given chat.
; "Real" means not typist items.
var
	int n,
	string typist
n = o.accChildCount
while n
	if !Skype__schat__getTypist(o, n, typist)
		return n
	endIf
	n = n -1
endWhile
return 0
endFunction

int function Skype__schat__getMessageIndex(object o)
; Given the MSAA object for a single message, return its 1-based index among messages.
; Returns 0 on error or failure.
var object oList = o.accParent
if !oList
	return 0
endIf
var int n = Skype__schat__getMessageCount(oList)
if !n
	return 0
endIf
; Bounding rectangle for the given message.
var int ml, int mr, int mt, int mb
if !Skype__MSAA__rect(o, 0, ml, mr, mt, mb)
	return 0
endIf

; This is a binary search based on bounding rectangle coordinates.
; Coordinates increase monotonically but can start negative.
; [DGL, 2014-09-14, Skype 6.18.59.106]
var int lbound = 1, int ubound = n, int i = 1
; Message to test against and its bounding rectangle.
var object oTest, int tl, int tr, int tt, int tb
var int safety = 100
while safety && ubound >= lbound
	oTest = oList.accChild(i)
	if !Skype__MSAA__rect(oTest, 0, tl, tr, tt, tb)
		return 0
	endIf
	if tt > mt
		; This message is below, hence after, the given one.
		ubound = i -1
		i = (lbound +i) /2
	elif tt < mt
		; This message is above, hence before, the given one.
		lbound = i +1
		i = (ubound +i) /2
	elif tl == ml && tr == mr && tb == mb
		; We have a winner!
		;sayInteger(100 -safety +1)
		return i
	endIf
	if i < lbound
		; Can happen due to integer arithmetic.
		i = lbound
	endIf
	safety = safety -1
endWhile
; The safety net ran out.
return 0
endFunction

int function Skype__schat___displayMessage(object oMsgs, int msgno)
; Force the indicated message to show on screen.
var
	handle hwnd,
	int childID
hwnd = getFocus()
if !(oMsgs.accState(msgno) & State_System_Selected)
	; This makes sure this item is on screen.
	speechOff()
	Skype__schat__selectMessage(oMsgs, msgno)
	pause()
	if getFocus() != hwnd
		setFocus(Skype__appwin__getSkypeWindow(SWT_ChatOutput))
		pause()
		setFocus(hwnd)
		pause()
	endIf
	speechOn()
endIf
return True
endFunction

collection function Skype__schat___parseMessage(object oMsgs, int childID)
; Get and return a collection breaking down the indicated message into component parts.
; This function is definitive on how to interpret Skype chat MSAA messages.
; Returned fields:
;	raw: The raw message text from MSAA.
;	sender, time, text: Parts of the message if the message is parsable.
;	message: The actual message (includes sender and text).
;		For unparseable messages, this is the same as raw.
;	meta: Meta info about the message, from its end, if parsable.
;		message + meta == raw.*
;	flags: edited, new, not delivered, etc.; drawn from meta.
;		Currently understood special messages: Edited, New, Not delivered.
;		For unparsable messages, this will be blank.
;		NOTE: SMS statuses sent and delivered are not handled specially here.
;	extra: Extra material drawn from elements under the message, if any.
;		This occurs for special "messages" like file transfers.
;	msgno, msgcount: 1-based index of message and message count.
; Assumes a message's accName property takes the approximate form
; from <sender>: <message>, sent at <date/time>[ edited by <editor> at <date/time>][ not delivered|new]
; This is current as of Skype 6.18 for most message types. [DGL, 2014-08-19]
; * info.message may have Skype__schat___scFrom removed for convenience in reading messages,
; and messages starting with sender name (/me lines and contact sharing)
; have the duplicate text removed; so message+meta may not always be equal to raw.
var
	collection info,
	string buf,
	string flags,
	int pos

; raw, extra, msgno and msgcount can be set right away.
; Set the rest to blank until filled in later.
buf = oMsgs.accName(childID)
info = new collection
info.raw = buf
info.sender = ""
info.time = ""
info.text = ""
info.message = ""
info.meta = ""
info.flags = ""
info.extra = Skype__schat___getExtraMessageMaterial(oMsgs, childID)
info.msgno = childID
info.msgcount = Skype__schat__getMessageCount(oMsgs)

; Identify "normal" message types.
if !buf || !stringStartsWith(buf, Skype__schat___scFrom)
	; We don't know how to parse these or there's nothing there.
	info.message = buf
	return info
endIf
; This just makes reading messages faster, and users have requested it.
buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(Skype__schat___scFrom)))

; Split meta from message.
var string sep = Skype__schat___scSent
pos = stringContainsFromRight(buf, sep)
if !pos
	; Special case for file transfers.
	sep = Skype__schat___scOn
	pos = stringContainsFromRight(buf, sep)
	if !pos
		info.message = buf
		return info
	endIf
endIf
info.message = stringLeft(buf, pos-1)
; This leaves sep in meta as well.
; This means info.message +info.meta == info.raw.
; Well almost; see above removal of Skype__schat___scFrom...
buf = stringChopLeft(buf, pos-1)
info.meta = buf

; Go back and get sender and text from message.
pos = stringContains(info.message, ", ")
if pos
	info.sender = stringLeft(info.message, pos-1)
	info.text = stringChopLeft(info.message, pos+1)
	; Extra simplification for contact detail sharing and /me lines:
	; If the sender name heads the message, the duplicate is cut.
	if stringStartsWith(info.text, formatString("%1 ", info.sender))
		info.sender = ""
	endIf
endIf

; Now to pull flags from the metadata, which remains in buf.
; Complex metadata example:
;	Wednesday, July 01, 2009 1:08:44 PM Edited by Doug Lee at Wednesday, July 01, 2009 1:09:17 PM Not delivered
; The initial time is from "sent on."
; I believe date and time values are formatted per locale here.

flags = ""
pos = stringContainsFromRight(buf, Skype__schat___scEditedBy)
if pos
	; Looks like this message has been edited.
	flags = flags +Skype__schat___msgEdited
	info.time = stringLeft(buf, pos-1)
endIf
if stringRight(buf, stringLength(Skype__schat___scNew)) == Skype__schat___scNew && childID != 1
	; childID != 1: Skype 4.1 (at least) likes to put "New" on the line with the links for showing history.
	flags = flags +Skype__schat___scNew
	buf = stringChopRight(buf, stringLength(Skype__schat___scNew))
endIf
if stringRight(buf, stringLength(Skype__schat___scNotDelivered)) == Skype__schat___scNotDelivered
	flags = flags +Skype__schat___scNotDelivered
	buf = stringChopRight(buf, stringLength(Skype__schat___scNotDelivered))
endIf
; Any unrecognized flags will end up in info.time.
info.time = buf
flags = stringTrimLeadingBlanks(flags)
info.flags = flags
return info
endFunction

string function Skype__schat___getExtraMessageMaterial(object o, int childID)
; Get and return any special content drawn from descendants of the indicated message.
; This applies to voicemails, file transfers, contact requests, etc.
; o/childID may be the message list and a childID or the message object and 0.
var int isFirst = (childID == 1)
if childID
	o = o.accChild(childID)
	childID = 0
endIf
var int n = o.accChildCount
if !n
	return ""
endIf

; ToDo: Slow but functional depth-first search without recursion.
var string buf = ""
var collection stack = Skype__schat__stl__stack(20)
Skype__schat__stl__push(stack, o)
Skype__schat__stl__push(stack, 0)
while Skype__schat__stl__size(stack)
	childID = Skype__schat__stl__pop(stack).value +0
	o = Skype__schat__stl__pop(stack).value
	childID = childID +1
	if childID <= o.accChildCount
		var object o1 = o.accChild(childID)
		if o1
			Skype__schat__stl__push(stack, o)
			Skype__schat__stl__push(stack, childID)
			o = o1
			childID = 0
		endIf
		if isFirst || o.accRole(childID) != Role_System_Link
			var string tmp = o.accName(childID)
			if !stringIsBlank(tmp)
				buf = buf +" " +tmp
			endIf
			tmp = o.accDescription(childID)
			if !stringIsBlank(tmp)
				buf = buf +" " +tmp
			endIf
			tmp = o.accHelp(childID)
			if !stringIsBlank(tmp)
				buf = buf +" " +tmp
			endIf
		endIf  ; not a link
		Skype__schat__stl__push(stack, o)
		Skype__schat__stl__push(stack, childID)
	endIf
endWhile
return stringTrimLeadingBlanks(buf)
endFunction

string function Skype__schat__getMessageText(object oMsgs, int msgno, string flags)
; Get the speakable/Braillable text of the indicated message.
; flags:
;	n for new (incoming) message / no special info like "not delivered."
;	Any other flags are silently ignored.
var
	int childID,
	string specialInfo, int noSpecialInfo,
	string msg
if !oMsgs
	return ""
elif msgno <= 0
	return ""
elif msgno > Skype__schat__getMessageCount(oMsgs)
	return ""
endIf
var collection info = Skype__schat___parseMessage(oMsgs, msgno)
msg = info.text
if info.sender
	msg = formatString("%1:\n%2", info.sender, msg)
endIf
if stringIsBlank(msg)
	msg = info.message
endIf
if info.extra
	msg = formatString("%1\n%2", msg, info.extra)
endIf
if Skype__opt__getUserOption(SV_ChatMessageTimes) == 1
	; Using message+meta allows extra to be put between them.
	var string meta = info.meta
	if stringContains(",:", stringLeft(meta, 1))
		meta = stringTrimLeadingBlanks(stringChopLeft(meta, 1))
	endIf
	msg = formatString("%1\n%2", msg, meta)
endIf
flags = stringLower(flags)
noSpecialInfo = (stringContains(flags, "n"))
if !noSpecialInfo
	specialInfo = info.flags
endIf
if specialInfo && !noSpecialInfo
	msg = specialInfo +", " +msg
endIf
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(msg))
endFunction

globals
; The set of messages being used by an active userBuffer of messages,
; the index of the current message among them,
; and the index of the wanted link within that.
	object Skype__schat___oMsgs, int Skype__schat___msgno, int Skype__schat___linkno

int function Skype__schat___clickItem(int msgno, int linkno)
; Clicks a link when the user activates it from a JAWS user buffer or its link list.
; Part of the support code for Skype__schat__virtualizeChatMessages().
if userBufferIsActive()
	userBufferDeactivate()
endIf
if msgno
	;sayString(formatString("Click message %1 item %2", intToString(msgno), intToString(linkno)))
	Skype__schat___msgno = msgno
	Skype__schat___linkno = linkno
	scheduleFunction("Skype__schat___clickItem", 0)
	return True
endIf

stopSpeech()
msgno = Skype__schat___msgno
linkno = Skype__schat___linkno
Skype__schat___msgno = 0
Skype__schat___linkno = 0

var object oMsg = Skype__schat___oMsgs.accChild(msgno)
if !oMsg
	beep()
	return False
endIf
return oMsg.accDoDefaultAction(linkno)
endFunction

int function Skype__schat___addMessageToUserBuffer(handle hwnd, object byRef oMsgs, int msgno)
; Adds the indicated message to a JAWS virtual view of conversation messages.
; Links in the message become links in the JAWS view as well.
; Part of the support code for Skype__schat__virtualizeChatMessages().
if !oMsgs
	if Skype__conv5__applies()
		oMsgs = Skype__conv5__chatContent(hwnd)
	else
		Skype__sutil__sayUnsupportedSkype()
		return False
	endIf
endIf
if oMsgs.accRole(0) != Role_System_List && oMsgs.accRole(0) != Role_System_Pane
return False
endIf
var string msg = Skype__schat__getMessageText(oMsgs, msgno, "")
var object oMsg = oMsgs.accChild(msgno)
if !oMsg
	oMsg = oMsgs
endIf
if !oMsg.accChildCount
	userBufferAddText(msg)
	return
endIf

; This message includes links and/or buttons and/or special text parts.
; To avoid duplication of text, we now have the slippery task of
; finding, in msg, the right occurrence of each link/button's text,
; so we can wrap userBuffer action support around it.
; The plan is to split msg into parts around each actionable item.
var int i, int iRole, string name, int pos, int wt
var int n = oMsg.accChildCount
var int checkForColon, string punct
for i = 1 to n
	iRole = oMsg.accRole(i)
	if iRole == Role_System_Link || iRole == Role_System_PushButton
		name = oMsg.accName(i)
		pos = 0
		checkForColon = False
		punct = ""
		if iRole == Role_System_Link
			pos = stringContains(msg, name)
			checkForColon = True
		elif iRole == Role_System_PushButton
		; Buttons don't split message text; they occur after it;
		; but we added extra stuff, including button names, to msg already.
		; We therefore scan from the right for these.
			pos = stringContainsFromRight(msg, name)
		endIf
		if pos
			; Add msg text up to but not including the actionable item.
			var string lhs = stringTrimTrailingBlanks(stringLeft(msg, pos-1))
			if lhs
				userBufferAddText(lhs)
			endIf
			; Then just keep whatever comes after the actionable item.
			msg = stringTrimLeadingBlanks(stringChopLeft(msg, pos +stringLength(name) -1))
			if checkForColon && stringStartsWith(msg, ":")
				msg = stringTrimLeadingBlanks(stringChopLeft(msg, 1))
				punct = ":"
			endIf
		elif msg
			; Add and clear all message text.
			userBufferAddText(msg)
			msg = ""
		endIf
		; Add the actionable item itself with its text.
		if iRole == Role_System_PushButton
			wt = WT_Button
		else
			wt = WT_Link
		endIf
		userBufferAddText(name +punct,
		"Skype__schat___clickItem(" +intToString(msgno) +", " +intToString(i) +")",
		; Include message number in links so, e.g., "Play" for five different
		; voicemails in one userBuffer will be unique in the Insert+F7 list.
		formatString(Skype__schat___msgMessageLinkInList, name, intToString(msgno)),
		"", 0, 0, IgnoreColor, IgnoreColor, True,
		wt, 0
		)
	endIf
endFor
if msg
	userBufferAddText(msg)
endIf
return True
endFunction

int function Skype__schat__virtualizeChatMessages(handle hwnd, int start, int contextSize, int contextPos)
; Virtualize contextSize+1 messages containing start.
; Negative start counts from the end (-1 is the last message).
; contextPos determines where start appears in the messages shown:
;	< 0:  Last message.
;	= 0:  Center message (contextSize/2 messages above and below).
;	> 0:  First message.
var
	object oMsgs, int offset,
	int n, int nmsgs,
	int idx, int idx1, int idx2,
	string title
hwnd = Skype__schat__getWindow(Skype__schat__WOutput, hwnd)
if !hwnd
	sayMessage(OT_Error, msgNoChatOutputWindow)
	return False
endIf
if Skype__conv5__applies()
	oMsgs = Skype__conv5__chatContent(hwnd)
else
	Skype__sutil__sayUnsupportedSkype()
	return False
endIf
if !oMsgs
	Skype__log__write(msgNoMSAA, True)
	return False
endIf

; Set start, idx1, and idx2.
; The shown message range is from idx1 through idx2 inclusive.
; Start is the message where the cursor begins and must be between idx1 and idx2 inclusive.
n = contextSize +1
nmsgs = Skype__schat__getMessageCount(oMsgs)
if start < 0
	start = nmsgs +1 +start
endIf
start = min(nmsgs, max(1, start))
if contextPos > 0
	idx1 = start
elif contextPos == 0
	idx1 = start - (contextSize /2)
else  ; contextPos < 0
	idx1 = start -contextSize
endIf
idx1 = min(nmsgs, max(1, idx1))
idx2 = idx1 +contextSize
;sayString(formatString("%1 %2 %3", intToString(idx1), intToString(start), intToString(idx2)))
offset = idx2 -nmsgs
if offset > 0
	idx1 = idx1 -offset
	idx2 = idx2 -offset
	idx1 = max(1, idx1)
endIf
Skype__schat___oMsgs = oMsgs
if userBufferIsActive()
	userBufferDeactivate()
endIf
userBufferClear()
var int pos
title = Skype__schat__getChatName(hwnd, False)
pos = stringContainsFromRight(title, "(")
if pos
	; Take off what is probably the conversation partner's online status.
	title = stringTrimTrailingBlanks(stringLeft(title, pos-1))
endIf
title = formatString(msgVChatTitle,
		intToString(idx1), intToString(idx2), intToString(nmsgs), title)
userBufferAddText(title)
offset = 0
idx = idx1
userBufferActivate()
while idx <= idx2
	if idx == start
		; +3: 1 to go from end of last message to start of this one,
		; 1 to skip the blank line atop this one,
		; 1 more to skip the origin and go to the message itself.
		offset = getLineCount() +3
	endIf
	userBufferAddText(" ")
	Skype__schat___addMessageToUserBuffer(hwnd, oMsgs, idx)
	idx = idx +1
endWhile
userBufferAddText(" ")
var string buf, int msgcount
if idx1 > 1
	msgcount = min(n-1, idx1-1)
	buf = formatString(msgVChatPrev, intToString(msgcount))
	userBufferAddText(buf,
			formatString("Skype__schat___deferredVirtualizeChatMessages(%1, %2, %3, -1)",
				intToString(hwnd), intToString(idx1), intToString(contextSize)),
			buf)
endIf
if idx2 < nmsgs
	msgcount = min(n-1, nmsgs-idx2+1)
	buf = formatString(msgVChatNext, intToString(msgcount))
	userBufferAddText(buf,
			formatString("Skype__schat___deferredVirtualizeChatMessages(%1, %2, %3, 1)",
				intToString(hwnd), intToString(idx2), intToString(contextSize)),
			buf)
endIf
userBufferAddText(cmsgBuffExit)
if offset
	goToLineNumber(offset)
endIf
if stringIsBlank(getLine())
	priorLine()
endIf
sayLine()
sayUsingVoice(VCTX_Message, title, OT_Help)
return True
endFunction

globals
	handle Skype__schat___vcHwnd,
	int Skype__schat___vcStart,
	int Skype__schat___vcContextSize,
	int Skype__schat___vcContextPos
int function Skype__schat___deferredVirtualizeChatMessages(handle hwnd, int start, int contextSize, int contextPos)
; Used by Skype__schat__virtualizeChatMessages when creating a new buffer from a Previous or Next link.
; This avoids the problem of secondary buffers just saying "F7" instead of bringing up a link list on JAWSKey+F7.
; [DGL, 2008-05-09]
if hwnd
	Skype__schat___vcHwnd = hwnd
	Skype__schat___vcStart = start
	Skype__schat___vcContextSize = contextSize
	Skype__schat___vcContextPos = contextPos
endIf
speechOn()
Skype__schat__virtualizeChatMessages(Skype__schat___vcHwnd, Skype__schat___vcStart, Skype__schat___vcContextSize, Skype__schat___vcContextPos)
endFunction

collection function Skype__schat__convInfo(handle hwnd)
; Return a collection whose properties describe the given conversation window.
; Properties returned (not all for every conversation):
;	window: The window handle to the conversation.
;	callActive: True if this conversation is (or contains) an active (connected) call.
;	multi: True if this conversation has multiple participants.
;	callHeld: Blank, "l" for local hold, or "r" for remote hold.
;	callMuted: True if muted (can be True on hold).
;	videoOut: y, n, or blank, for yes, no, or apparently not available.
;	videoIn: Same as for videoOut for incoming video signal.
;	callQuality: Call quality description if present.
;	callDuration: Call duration if valid, in [hh:]mm:ss format.
; All Skype__schat___convInfo* functions called by this one must know the above properties.
var collection null null = new collection
var collection oConv oConv = new collection
if Skype__conv5__applies()
	Skype__schat__chat5__getInfo(oConv, hwnd)
else
	Skype__sutil__sayUnsupportedSkype()
	return null
endIf
return oConv
endFunction

string function Skype__schat__convDescription(variant which)
; Return a brief description of the conversation indicated by the indicated conversation.
; which: Window handle of conversation, or collection from a Skype__schat__convInfo() call on that conversation.
var
	collection oConv,
	string sMulti, string sCall, string sMuted, string sHeld,
	string duration,
	string buf
if intToString(which) != 0
	oConv = Skype__schat__convInfo(which)
else
	oConv = Skype__util__vcast(which)
endIf
buf = ""
sMulti = ""
if oConv.multi
	sMulti = " " +Skype__schat__msgMulti
endIf

sCall = ""
if oConv.callActive
	if oConv.videoIn == "y"
		if oConv.videoOut == "y"
			sCall = " " +Skype__schat__msgVideoCall
		else  ; Both disabled and unsupported outgoing video cases.
			sCall = " " +Skype__schat__msgVideoInCall
		endIf
	else  ; No incoming video.
		if oConv.videoOut == "y"
			sCall = " " +Skype__schat__msgVideoOutCall
		elif oConv.videoOut == "n"
			sCall = " " +Skype__schat__msgVoiceCall
		else
			sCall = " " +Skype__schat__msgCall
		endIf
	endIf
	duration = oConv.callDuration
	if duration
		sCall = sCall +" " +duration
	endIf
else
	; Nothing else to say...
	return stringTrimLeadingBlanks(sMulti)
endIf

sMuted = ""
if oConv.callMuted
	sMuted = " " +Skype__schat__msgMuted
endIf

sHeld = ""
if oConv.callHeld == "l"
	sHeld = " " +Skype__schat__msgLocalHold
elif oConv.callHeld == "r"
	sHeld = " " +Skype__schat__msgRemoteHold
endIf

buf = formatString(Skype__schat__msgConvDescFormat, sMulti, sCall, sMuted, sHeld)
return stringTrimLeadingBlanks(buf)
endFunction


;------------------------ Module chat5 (from SkypeChat5) -----------------------
; Code for filling the conversation info structure from a Skype 5.x conversation window tree.
; Supports Skype 5.0 beta 2 and later.
; Does not support Skype 3.x, 4.x, or 5.0 beta 1.
;
; Author:  Doug Lee

void function Skype__schat__chat5__getInfo(collection byRef oConv, handle hwnd)
; Fill the given conversation info collection from the given Skype 5.0 (beta 2 or later) conversation window.
; oConv:  An empty conversation info collection; see Skype__schat__convInfo() for details..  Must be empty on call.
; hwnd: A conversation window or a member of one.
var
	handle hwndConvToolBar,
	handle hwndContactArea,
	object o, int childID,
	string buf

; First check for a Skype 5.1/5.2 live call area.
var collection oLiveInfo = Skype__live5__getInfo(hwnd)
if !Skype__util__collectionIsEmpty(oLiveInfo)
	oConv.window = hwnd
	oConv.callActive = True
	var collection oParticipants = oLiveInfo.participants
	oConv.multi = (collectionItemCount(oParticipants) > 1)
	oConv.callHeld = oLiveInfo.held
	oConv.callMuted = oLiveInfo.muted
	oConv.videoOut = oLiveInfo.videoOut
	oConv.videoIn = oLiveInfo.videoIn
	oConv.callQuality = oLiveInfo.callQuality
	oConv.callDuration = oLiveInfo.duration
	return
endIf
hwndConvToolBar = Skype__schat__getWindow(Skype__schat__wConvToolBar, hwnd)
if !hwndConvToolBar
	return
endIf
o = getObjectFromEvent(hwndConvToolBar, ObjID_Client, 0, childID)
if !o
	return
endIf
oConv.window = hwnd

; Does this conversation contain multiple paerticipants?
hwndContactArea = Skype__schat__getWindow(Skype__schat__WContactArea, hwnd)
if hwndContactArea
	if Skype__MSAA__windowProp(hwndContactArea, "Name") == Skype__schat__scParticipants
		oConv.multi = True
	else
		oConv.multi = False
	endIf
endIf

; Does this conversation contain or consist of a call?
if Skype__MSAA__childByName(o, Skype__schat__scEndCall)
	oConv.callActive = True
else
	oConv.callActive = False
	; The rest of the properties only apply to calls, so we're done.
	return
endIf

; Hold status/type.
if Skype__MSAA__childByName(o, Skype__schat__scResume)
	; This is a button.
	oConv.callHeld = "l"
elif Skype__MSAA__childByName(o, Skype__schat__scRemoteHold)
	; This has a role of Toolbar.
	oConv.callHeld = "r"
endIf

; Mute status.
if Skype__MSAA__childByName(o, Skype__schat__scUnmute)
	oConv.callMuted = True
else
	oConv.callMuted = False
endIf

; Outgoing video status.
if Skype__MSAA__childByName(o, Skype__schat__scStopVideo)
	oConv.videoOut = "y"
elif Skype__MSAA__childByName(o, Skype__schat__scStartVideo)
	oConv.videoOut = "n"
endIf
; TODO: Receiving part not done.

; Call quality if present.
; This is apparently blank unless the call quality is suboptimal.
childID = Skype__MSAA__childByName(o, Skype__schat__scCallQuality, "p")
if childID
	buf = o.accName(childID)
	buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(Skype__schat__scCallQuality)))
	if buf
		oConv.callQuality = buf
	endIf
endIf

; Call duration if valid.
if Skype__conv5__applies()
endIf
return
endFunction

;---------------------- End module chat5 (from SkypeChat5) ---------------------


;---------------------------------- Module stl ---------------------------------
; Standard data structures for JAWS.
;
; Author: Doug Lee of SSB BART Group

const
; List types supported.
	Skype__schat__stl___LstStack = 1,
	Skype__schat__stl___LstQueue = 2

collection function Skype__schat__stl__stack(optional int maxSize)
; Return a stack for use via Skype__schat__stl__push/Skype__schat__stl__pop etc.
return Skype__schat__stl___new(Skype__schat__stl___LstStack, maxSize)
endFunction

collection function Skype__schat__stl__queue(optional int maxSize)
; Return a queue for use via Skype__schat__stl__push/Skype__schat__stl__pop etc.
return Skype__schat__stl___new(Skype__schat__stl___LstQueue, maxSize)
endFunction

collection function Skype__schat__stl___new(int ltype, optional int maxSize)
; Return a structure of the indicated type.
var collection lst lst = new collection
lst.type = ltype
if ltype == Skype__schat__stl___LstQueue
	; ToDo: maxSize ignored.
	lst.size = 0
	return lst
endIf

lst.maxSize = 0
lst.top = 0
if maxSize
	; More efficient.
	lst.maxSize = maxSize
	var variantArray ar ar = new variantArray[lst.maxSize]
	lst.content = ar
else
	; A bit less efficient but more flexible.
	var collection nc nc = new collection
	lst.content = nc
endIf
return lst
endFunction

int function Skype__schat__stl__push(collection struct, variant item)
; Push an item onto whatever structure is passed.
; Returns True on success and False on failure.
if struct.ltype == Skype__schat__stl___LstQueue
	var collection node node = new collection
	node.value = item
	var collection node0 = struct.tail
	if node0
		node0.next = node
		struct.tail = node
	else
		struct.head = node
		struct.tail = node
	endIf
	struct.size = struct.size +1
	return True
endIf

struct.top = struct.top +1
if struct.maxSize
	if struct.top > struct.maxSize
		struct.top = struct.top -1
		return False
	endIf
	var variantArray ar = struct.content
	ar[struct.top] = item
	;sayString("Push " +intToString(struct.top))
else
	var collection col = struct.content
	col[intToString(struct.top)] = item
	;sayString("Dynamic push " +intToString(struct.top))
endIf
return True
endFunction

collection function Skype__schat__stl__top(collection struct, optional int alsoPop)
; Return the top element of whatever structure is passed.
; Also pops the element off if alsoPop is True.
; The item is returned as the .value property of a simple collection.
; On error, null is returned rather than a collection.
var collection item item = new collection
if struct.type == Skype__schat__stl___LstQueue
	var collection node = struct.head
	if !node
		return
	elif alsoPop
		struct.head = node.next
		struct.size = struct.size -1
	endIf
	return node
endIf

if struct.top < 1
	return
endIf
if struct.type == Skype__schat__stl___LstStack
	if struct.maxSize
		var variantArray ar = struct.content
		item.value = ar[struct.top]
	else
		var collection col = struct.content
		item.value = col[intToString(struct.top)]
	endIf
	if alsoPop
		;sayString("Pop " +intToString(struct.top))
		struct.top = struct.top -1
	endIf
	return item
endIf
return
endFunction

collection function Skype__schat__stl__pop(collection struct)
; Pop an item from whatever structure is passed.
; The item comes back as the .value property of a collection.
; On error, returns null instead of a collection.
return Skype__schat__stl__top(struct, True)
endFunction

int function Skype__schat__stl__size(collection struct)
; Return the size (item count) of whatever structure is passed.
if struct.type == Skype__schat__stl___LstQueue
	return struct.size
endIf
return struct.top
endFunction

;-------------------------------- End module stl -------------------------------

;====================== End module schat (from SkypeChat) ======================


;================================== Module cre =================================
; Code for fixing the following issue in Skype 6.9+ TChatRichEdit input boxes:
; On just the first message typed into a particular chat window,
; if enhanced edit support is used (default), no text is seen.
; In this case, using the OSM is better, until the message is sent.
;
; Author: Doug Lee

/*
Notes:
In Default view, where chats stay in the main Skype window, chat input
edit controls often disappear and reappear with new window handles. On
the first appearance of any window handle in  Default view, JAWS 15
(through at least 15.0.6025) seems unable to retrieve the
iTextDocument interface internally and/or use it effectively to
implement its enhanced edit mode. When this happens, users are unable
to see any text in the window, locate text to edit, hear anything but
"blank" on arrows and backspace, etc. Using the OSM, i.e., avoiding
enhanced edit mode, works better in this case. Alt+Tabbing out of and
back into Skype, forcing focus changes away from and back to the edit
window, and even reloading JAWS do not fix this problem. However, as
soon as one message is sent from the window, things begin to work
properly in enhanced edit mode (assuming it's still active).

The above problem can be detected programmatically by checking for a
non-blank MSAA accValue for the edit control and a blank
getObjectValue() result for it. These two conditions occur
simultaneously only when JAWS is unable to access the control
correctly.

This module deactivates enhanced edit mode when the problem is
detected. Window handles of chat input controls are remembered both to
minimize time required to grab IAccessible objects for them and to
allow tracking of which windows have fixed themselves. Enhanced mode
is enabled again when either the accValue is again empty or a
correctly-working chat window takes focus when enhanced mode is not in
effect.
*/

globals
; The current edit window, 0 if the focused window is not one.
	handle Skype__cre___curWindow,
; The per-window data for this module. Fields:
;	hwnd: The window handle.
;	obj: The MSAA Client object for the window.
;	fixed: True if enhanced edit mode has been seen to work in this window.
	collection Skype__cre___hwnds

int function Skype__cre__manage()
; Manage the Enhanced Edit Support usage for TChatRichEdit windows.
var handle hwnd = getFocus()
Skype__cre___updateCurrentWindow(hwnd)
var int opt = Opt_Edit_Use_OSM

; If the MSAA mode is above 1, the below system doesn't work.
if getJCFOption(OPT_MSAA_Mode) > 1
	return False
endIf

if getWindowClass(hwnd) != "TChatRichEdit"
	return False
endIf
var collection rec = Skype__cre__winrecs__record(Skype__cre___hwnds, hwnd)
if !rec.hwnd
	rec.obj = Skype__MSAA__objectFromWindow(hwnd)
	if !rec.obj
		return False
	endIf
	rec.fixed = False
	rec.hwnd = hwnd
endIf

; ToDo: Undocumented JAWS function.
var int enhanced = usingEnhancedEditSupport(hwnd)
if rec.fixed
	; Once fixed, always fixed until the window goes away.
	if !enhanced
		setJCFOption(opt, 0)
	endIf
	return False
endIf

; If the actual value is blank, we can't tell if we have a problem yet.
var object o = rec.obj
if !o.accValue(0)
	; But we can make sure to use enhanced mode in case it starts working.
	if !enhanced
		setJCFOption(opt, 0)
	endIf
	return False
endIf

if !enhanced
	; Apparently already compensating for this issue.
	return False
endIf

; If getObjectValue is not blank, things should be working normally.
; False means do not use MSAA here.
if getObjectValue(False)
	rec.fixed = True
	return False
endIf

; Problem: objectValue blank, MSAA value not blank.
; Meaning: JAWS can't read the edit control in enhanced mode.
; Solution: Quit using enhanced edit support.
setJCFOption(opt, 1)
return True
endFunction

void function Skype__cre___updateCurrentWindow(handle hwnd)
; Update data associated with the current edit window.
if hwnd == Skype__cre___curWindow
	return
endIf
; ToDo: Update any required info.
Skype__cre___curWindow = Skype__util__vcast(0)
endFunction


;-------------------------------- Module winrecs -------------------------------
; A manager for data structures (records) indexed by a window handle.
;
; Author:  Doug Lee of SSB BART Group

collection function Skype__cre__winrecs__record(collection byRef recs, handle hwnd)
; Return the record for the given window.
; Registers this window in the set if necessary.
; Also removes stale records for windows that no longer exist.
; The caller must handle record structure creation.
; Suggestion: Use an "hwnd" property so there's a uniform way to test for a new record.
Skype__cre__winrecs__clean(recs)
if !hwnd then
	return
endIf
var string shwnd let shwnd = intToString(hwnd)
var collection rec
if !recs then
	let recs = new collection
endIf
if !collectionItemExists(recs, shwnd) then
	; We have to make a new record for this one.
	let rec = new collection
	recs[shwnd] = rec
	; Caller must detect that this record is empty and build it.
else
	let rec = recs[shwnd]
endIf
return rec
endFunction

int function Skype__cre__winrecs__remove(collection recs, handle hwnd)
; Remove a window's record.
var string shwnd let shwnd = intToString(hwnd)
if !collectionItemExists(recs, shwnd) then
	return False
endIf
collectionRemoveItem(recs, shwnd)
return True
endFunction

int function Skype__cre__winrecs__clean(collection recs, /*?*/ int clearAll)
; Remove from recs any records for windows that no longer exist.
; If clearAll is True, removes all records.
; Returns the number of records removed.
; Note that this function is called by Skype__cre__winrecs__record(),
; so manual use of it is optional.
var string shwnd
var int nremoved = 0
if clearAll then
	var int n = 0
	forEach shwnd in recs
		n = n +1
	endForEach
	collectionRemoveAll(recs)
	return n
endIf
var handle hwnd
forEach shwnd in recs
	let hwnd = stringToHandle(shwnd)
	if !getWindowStyleBits(hwnd) then
		if Skype__cre__winrecs__remove(recs, hwnd) then
			let nremoved = nremoved +1
		endIf
	endIf
endForEach
return nremoved
endFunction

;------------------------------ End module winrecs -----------------------------

;================================ End module cre ===============================


;================================== Module opt =================================
; User option code.

const
; A constant used to detect missing, as opposed to set-to-empty, .ini keys.
	Skype__opt___NoVal = 12345678

variant function Skype__opt__getUserOption(string varName)
; Retrieves and returns the value of the indicated user-settable variable from skype.ini.
; This function makes sure to return integer 0 for the integer value 0, rather than returning the string "0" which shows up as True in a boolean test.
var
	string val
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
val = iniReadString(CnstIniSectName, varname, intToString(Skype__opt___NoVal), CnstIniFileName)
; Favor integer 0 (False) rather than string "0" (True).
if stringContains(stringLeft(stringTrimLeadingBlanks(val), 1), "0")
	return ""
endIf
if val == Skype__opt___NoVal
	; All default values for options go here.
	if varname == SV_RecognizeChat
		val = "1"
	elif varname == SV_ChatMessageTimes
		val = "1"
	elif varname == SV_SpeakOnlineStatus
		val = "1"
	elif varname == SV_TypingAnnounce
		val = "1"
	elif varname == SV_TypingSounder
		val = "1"
	elif varname == SV_SkypeCommandKey
		val = "["
	elif varname == SV_SkypeWatchMessages
		val = "1"
	else
		val = ""
	endIf
endIf
return val
endFunction

int function Skype__opt__setSkypeVar(string varName, variant val)
; Sets the indicated user-settable variable in skype.ini to the given value.
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
return iniWriteString(CnstIniSectName, varname, val, CnstIniFileName)
endFunction

variant function Skype__opt__userOptionCycler(string varname, string choices, int iRetCurVal)
; Used by DlgSelectFunctionToRun cycler/toggler functions to cycle/toggle variable values.
; Varname should be the option name being handled (SV_* constant).
; Choices should be a list of text values, separated by vertical bars (|).
; iRetCurVal should be the integer taken by the calling toggler/cycler function.
; The return value of this function behaves according to specifications for DlgSelectFunctionToRun functions and should be returned directly by the toggler/cycler function.
; The integer value saved in skype.ini will be the 0-based index of the chosen option.
var
	int i,
	int n
i = Skype__opt__getUserOption(varname)
n = Skype__string__segmentCount(choices, "|")
if !iRetCurVal
	; Update it.
	i = i +1
	if i >= n
		i = 0
	endIf
	Skype__opt__setSkypeVar(varname, i)
endIf
; i+1 rather than just i because i is 0-based and stringSegment is 1-based.
return stringSegment(choices, "|", i+1)
endFunction

;================================ End module opt ===============================


;================================ Module appOpt ================================
; User option code.

String Function jvoRecognizeChat (int RetCurValue)
return Skype__opt__userOptionCycler(SV_RecognizeChat, msgDisabledEnabled, RetCurValue)
EndFunction

String Function jvoSpeakOnlineStatus (int RetCurValue)
return Skype__opt__userOptionCycler(SV_SpeakOnlineStatus, msgDisabledEnabled, RetCurValue)
EndFunction

String Function jvoTypingAnnounce (int RetCurValue)
return Skype__opt__userOptionCycler(SV_TypingAnnounce, msgDisabledEnabled, RetCurValue)
EndFunction

String Function jvoTypingSounder (int RetCurValue)
var
	string result,
	string idx
result = Skype__opt__userOptionCycler(SV_TypingSounder, msgTypingSounderChoices, RetCurValue)
if !RetCurValue
	; A change, so play a sample
	idx = Skype__opt__getUserOption(SV_TypingSounder)
	if idx
		Skype__sutil__play(formatString(fnameSkypeTyping, idx))
	endIf
endIf
return result
EndFunction

String Function jvoSkypeWatchMessages (int RetCurValue)
return Skype__opt__userOptionCycler(SV_SkypeWatchMessages, msgSkypeWatchMessageChoices, RetCurValue)
EndFunction

String Function jvoChatMessageTimes (int RetCurValue)
return Skype__opt__userOptionCycler(SV_ChatMessageTimes, msgChatMessageTimeChoices, RetCurValue)
EndFunction

String Function jvoSkypeCommandKey (int iRetCurVal)
; A special verbosity/options function for managing the Skype command key.
var
	string varname,
	string sKey
varname = SV_SkypeCommandKey
sKey = Skype__opt__getUserOption(varname)
if !iRetCurVal
	if Skype__appOpt__giSettingSkypeCommandKey == 0
		; The user just pressed Space on this entry:
		; Ask the user for a key .
		sayMessage(OT_No_Disable, Skype__appOpt__msgKeyChangePrompt)
		; Prevent the next key from doing anything except calling KeyPressedEvent.
		trapKeys(True)
		; Tell the KeyPressedEvent we're expecting a Skype Command key to be typed now.
		Skype__appOpt__giSettingSkypeCommandKey = 1
		; Prevent JAWS from saying the current key value again.
		speechOff()
		scheduleFunction("speechOn", 2)
		; but then return as normal.
		; JAWS should fall silent after the above prompt.
		; The next user key will hit the KeyPressedEvent and get stored,
		; and then KeyPressedEvent will simulate a Space,
		; which will make DlgSelectFunctionToRun call this function again,
		; with iRetCurVal False but Skype__appOpt__giSettingSkypeCommandKey still 1,
		; and control will continue below.
	elif Skype__appOpt__giSettingSkypeCommandKey == 1
		; KeyPressedEvent caused this call on receiving (and setting) the new command key.
		trapKeys(False)
		; Set all relevant jkm lines to match the newly stored key.
		Skype__appOpt__resetSkypeCommandKey()
		Skype__appOpt__giSettingSkypeCommandKey = 0
	endIf  ; Skype__appOpt__giSettingSkypeCommandKey
endIf  ; !iRetCurVal
return sKey
EndFunction

void function Skype__appOpt__resetSkypeCommandKey()
var
	string sect, string fname,
	string val,
	string sKey, string sKeyList,
	int i, int n,
	string seg,
	int pos,
	int changing
sKey = Skype__opt__getUserOption(SV_SkypeCommandKey)
Skype__mkcs__setCommandKey(sKey)
endFunction

void function Skype__appOpt___addVerbosityOption(string byRef lst, string optionData)
; Adds an option to the given verbosity (JAWS 8-) / option (JAWS 9+) list.
var
	string varAndDesc
varAndDesc = formatString("jvo%1:%2",
	stringSegment(optionData, ":", 1),
	stringSegment(optionData, ":", 2))
if lst
	lst = lst +"|"
endIf
lst = lst +varAndDesc
endFunction

Script AdjustJawsOptions ()
performScript adjustJAWSVerbosity()
endScript

Script AdjustJawsVerbosity ()
; Uses Jaws's central verbosity coding.
; Modified for skype usage by Ner
; To write more toggle functions for this, see toggle functions
; They *must* be done in the same manner.
var
	string lst

lst = ""
Skype__appOpt___addVerbosityOption(lst, SV_RecognizeChat)
Skype__appOpt___addVerbosityOption(lst, SV_ChatMessageTimes)
Skype__appOpt___addVerbosityOption(lst, SV_SpeakOnlineStatus)
Skype__appOpt___addVerbosityOption(lst, SV_TypingAnnounce)
Skype__appOpt___addVerbosityOption(lst, SV_TypingSounder)
Skype__appOpt___addVerbosityOption(lst, SV_SkypeCommandKey)
if Skype__sutil__isSkypeWatchLoaded()
	Skype__appOpt___addVerbosityOption(lst, SV_SkypeWatchMessages)
endIf
OptionsTreeCore(lst)
EndScript

String Function NodeHLP (string NodeName)
	if NodeName == GetActiveConfiguration ()
return Skype__appOpt__msgNodeHlp
endIf
return nodeHlp(nodeName)
EndFunction

String Function jvoRecognizeChatHlp()
return stringSegment(SV_RecognizeChat, ":", 3)
EndFunction

String Function jvoChatMessageTimesHlp()
return stringSegment(SV_ChatMessageTimes, ":", 3)
EndFunction

String Function jvoSpeakOnlineStatusHlp()
return stringSegment(SV_SpeakOnlineStatus, ":", 3)
EndFunction

String Function jvoTypingAnnounceHlp()
return stringSegment(SV_TypingAnnounce, ":", 3)
EndFunction

String Function jvoTypingSounderHlp()
return stringSegment(SV_TypingSounder, ":", 3)
EndFunction

String Function jvoSkypeCommandKeyHlp()
return stringSegment(SV_SkypeCommandKey, ":", 3)
EndFunction

String Function jvoSkypeWatchMessagesHlp()
return stringSegment(SV_SkypeWatchMessages, ":", 3)
EndFunction

;============================== End module appOpt ==============================


;=============================== Module winfinder ==============================
; Ways to search for windows by typeCode, class substring match, and by calling test functions.
;
; Author:  Doug Lee

globals
; Variables used to communicate between Skype__winfinder__findWindow and Skype__winfinder___findWindowHelper.
; Required in order to use enumerateChildWindows().
	handle Skype__winfinder___windowFound,
	string Skype__winfinder___windowClass,
	string Skype__winfinder___windowName

int function Skype__winfinder___findWindowHelper(handle hwnd)
; Called via EnumerateChildWindows.
var
	int pos,
	string wclass,
	int found
wclass = Skype__winfinder___windowClass
if stringContains("@!", stringLeft(wclass, 1))
	pos = stringContains(wclass, "%%")
	if pos
		wclass = stringLeft(wclass, pos-1) +intToString(hwnd) +stringChopLeft(wclass, pos+1)
	endIf
endIf
if wclass == "@"
	found = stringToInt(formatStringWithEmbeddedFunctions("<" +stringChopLeft(wclass, 1) +">"))
elif wclass == "!"
	found = !stringToInt(formatStringWithEmbeddedFunctions("<" +stringChopLeft(wclass, 1) +">"))
else
	found = (!stringCompare(getWindowClass(hwnd), wclass, False) && !stringCompare(getWindowName(hwnd), Skype__winfinder___windowName, False))
endIf
if found
	Skype__winfinder___windowFound = hwnd
	return False
endIf
return True
endFunction

handle function Skype__winfinder__findWindow(handle hTop, string wclass, string wname)
; Generalized window finder. Can be used just like JAWS' FindWindow function.
; Also accepts in wclass @funcCall or !funcCall to use arbitrary function calls to test windows.
; funcCall should be a function call with %% in place of the hwnd parameter.
; NOTE: The Name option uses getWindowName, not WM_GETTEXT. They can differ.
; The JAWS built-in FindWindow function uses WM_GETTEXT according to impirical data.
var
handle hwnd,
	handle nullHandle
hwnd = findWindow(hTop, wclass, wname)
if hwnd
	return hwnd
endIf
Skype__winfinder___windowClass = wclass
Skype__winfinder___windowName = wname
Skype__winfinder___windowFound = nullHandle
EnumerateChildWindows (hTop, "Skype__winfinder___findWindowHelper")
return Skype__winfinder___windowFound
endFunction

int function Skype__winfinder___isTypeCode(handle hwnd, int typeCode, int isSubtype)
; Returns True if the given window has the given type or subtype code.
if isSubtype
	return (getWindowSubtypeCode(hwnd) == typeCode)
else
	return (getWindowTypeCode(hwnd) == typeCode)
endIf
endFunction

handle function Skype__winfinder__findWindowByTypeCode(handle hTop, int typeCode, int isSubtype)
; Finds the first window under hTop (inclusive) having the given type or subtype code.
; Author:  Doug Lee
return findWindowByType(hTop, typeCode)
endFunction

int function Skype__winfinder__getWindowRole(handle hwnd)
; Returns the given window's MSAA client object's role.
; This role, if not Role_System_Client, should represent the window's actual function.
; Author:  Doug Lee
var int childID  ; required by GetObjectFromEvent but otherwise ignored here
return getObjectFromEvent(hwnd, -4, 0, childID).accRole(0)
endFunction

int function Skype__winfinder__getWindowValue(handle hwnd)
; Returns the given window's MSAA client object's value.
; Author:  Doug Lee
var int childID  ; required by GetObjectFromEvent but otherwise ignored here
return getObjectFromEvent(hwnd, -4, 0, childID).accValue(0)
endFunction

int function Skype__winfinder___isMSAARole(handle hwnd, int iRole)
; Returns True if the given window's MSAA client object has the given role.
; Author:  Doug Lee
var int childID
return (Skype__winfinder__getWindowRole(hwnd) == iRole)
endFunction

handle function Skype__winfinder__findWindowByMSAARole(handle hTop, int iRole)
; Finds the first window under hTop (inclusive) whose MSAA client object has the given role.
; Author:  Doug Lee
return Skype__winfinder__findWindow(hTop, "@Skype__winfinder___isMSAARole(%%, " +intToString(iRole) +")", "")
endFunction

int function Skype__winfinder___isClassMatch(handle hwnd, string classPart)
; Returns True if the given window's class name contains classPart.
if stringContains(stringLower(getWindowClass(hwnd)), stringLower(classPart))
	return True
endIf
return False
endFunction

handle function Skype__winfinder__findWindowByClassSubstring(handle hTop, string classPart)
; Finds the first window under hTop (inclusive) whose class name contains classPart.
; Author:  Doug Lee
return Skype__winfinder__findWindow(hTop, "@Skype__winfinder___isClassMatch(%%, " +classPart +")", "")
endFunction

;============================= End module winfinder ============================


;================================ Module appwin ================================
; getSkypeWindow and related app window identification code.

globals
; The main Skype window
	handle Skype__appwin___mainWindow,
; A place to cache the MSAA contact list object.
	object Skype__appwin___contactListObject

int function Skype__appwin__isMainSkypeWindow(handle hwnd)
var
	string wclass
wclass = getWindowClass(hwnd)
if wclass == "TskMainForm"
|| wclass == "TZapMainForm"
	return True
endIf
return False
endFunction

handle function Skype__appwin__getMainSkypeWindow()
; Returns the top-level main Skype window if possible.
; If the main window cannot be found, also clears the MSAA contact list object cache.
var
	object nullObject,
	handle hwnd
hwnd = getTopLevelWindow(getFocus())
if Skype__appwin__isMainSkypeWindow(hwnd)
	Skype__appwin___mainWindow = hwnd
	return hwnd
endIf
hwnd = Skype__appwin___mainWindow
if Skype__appwin__isMainSkypeWindow(hwnd)
	return hwnd
endIf
hwnd = findTopLevelWindow("tSkMainForm", "")
if !hwnd
	hwnd = findTopLevelWindow("tSkMainForm.UnicodeClass", "")
endIf
if hwnd
	Skype__appwin___mainWindow = hwnd
	return hwnd
endIf
Skype__appwin___contactListObject = nullObject
return 0
endFunction

int function Skype__appwin__isChatWindow(handle hwnd)
; Returns 1 (True) if the given window is or is within a Skype Instant Messaging window.
while hwnd
	if getWindowClass(hwnd) == "TConversationForm"
		return 1
	endIf
	hwnd = getParent(hwnd)
endWhile
return 0
endFunction

int function Skype__appwin___isSkypeToolTip(handle hwnd)
; Returns True if the given window is the Skype tooltip window.
return hwnd && stringContains(stringLower(getWindowClass(hwnd)), cwsToolTip)
		;&& getParent(hwnd)
		;&& getWindowClass(getParent(hwnd)) == wcApp && getWindowName(getParent(hwnd)) == Skype__appwin__wnSkype
endFunction

int function Skype__appwin__isSkypeNotifyWindow(handle hwnd)
; Returns True if the given window is a Skype system tray notification window or call alert.
var
	string wclass
wclass = getWindowClass(hwnd)
return (wclass == "TCallNotificationWindow"
|| wclass == "TTrayAlert")
endFunction

int function Skype__appwin___isRequestedSkypeWindow(handle hwnd, string which)
; Returns True if the given window is the given Skype window.
if which == SWT_ToolTip
	return Skype__appwin___isSkypeToolTip(hwnd)
elif which == SWT_Notify
	return Skype__appwin__isSkypeNotifyWindow(hwnd)
else
	return False
endIf
endFunction

int function Skype__appwin__isSkypeWindow(handle hwnd)
; Returns True if the given window is part of Skype.
var
	string wclass
if stringRight(Skype__string__ref(getWindowOwner(hwnd)), 9) == "Skype.exe"
	return True
endIf
return False
endFunction

handle function Skype__appwin___findTopLevelSkypeWindow(string which)
; Returns the handle of the requested Skype top-level window.
; This window may not exist, such as right after Skype loads.
var
	handle hwnd
hwnd = Skype__util__getFirstTopLevelWindow()
while hwnd && !Skype__appwin___isRequestedSkypeWindow(hwnd, which)
	hwnd = getNextWindow(hwnd)
endWhile
if Skype__appwin___isRequestedSkypeWindow(hwnd, which)
	return hwnd
endIf
return 0
endFunction

handle function Skype__appwin__getActiveChatWindow(handle hTop)
var
	handle hwnd
hwnd = findWindow(hTop, "TConversationForm", "")
if !hwnd
	hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
endIf
if !hwnd || isWindowVisible(hwnd)
	; Got it or there isn't such a thing.
	return hwnd
endIf
hwnd = getNextWindow(hwnd)
while hwnd
	if isWindowVisible(hwnd) && Skype__appwin__isChatWindow(hwnd)
		return hwnd
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

handle function Skype__appwin__getSkypeWindow(string which)
; Returns the handle of the window of the indicated type.
; Note that this only works for window types that are unique.
; This function should be kept in sync with the set of SWT_* constants in skype.jsh.
var
	handle hMain, handle hwnd, handle hwnd1,
	string hList
if which == "?"
	hList = SWT_Main +"|" +SWT_Myself
		+"|" +SWT_ContactList +"|" +SWT_ConversationList +"|" +SWT_LiveConversations
		+"|" +SWT_History
	hList = hList +"|" +SWT_ChatMain +"|" +SWT_ChatInput +"|" +SWT_ChatOutput +"|" +SWT_ChatList +"|" +SWT_ChatContact +"|" +SWT_Communicator +"|" +SWT_CallQuality
	; Omitted knowingly:  SWT_ToolTip, SWT_Notify.
	return hList
endIf
hMain = Skype__appwin__getMainSkypeWindow()
if !hMain
	return 0
endIf
if which == SWT_Main
	return hMain
endIf
if which == SWT_Myself
	return Skype__winfinder__findWindowByClassSubstring(hMain, cwsMyself)
elif which == SWT_ContactList
	hwnd1 = findWindow(hMain, "TConversationList", "")
	if hwnd1
		if which == SWT_ContactList
			return hwnd1
		endIf
		return getParent(hwnd1)
	endIf
	return 0
elif which == SWT_ConversationList
	return findWindow(hMain, "TInboxList", "")
elif which == SWT_LiveConversations
	hwnd = findWindow(hMain, "TLiveConversationList", "")
	return hwnd
elif which == SWT_History
	return Skype__winfinder__findWindowByClassSubstring(hMain, "history")
elif which == SWT_Home
	return findWindow(findWindow(hMain, "THomeForm", ""), "Internet Explorer_Server", "")
elif which == SWT_ToolTip
	return Skype__appwin___findTopLevelSkypeWindow(SWT_ToolTip)
elif which == SWT_Notify
	var handle hwndn1, handle hwndn2
	hwndn1 = findTopLevelWindow("TCallNotificationWindow", "")
	hwndn2 = findTopLevelWindow("TTrayAlert", "")
	if hwndn1 && hwndn2
		if getWindowHierarchyX(hwndn1) < getWindowHierarchyX(hwndn2)
			hwnd1 = hwndn1
		else
			hwnd1 = hwndn2
		endIf
	elif hwndn1
		hwnd1 = hwndn1
	else
		hwnd1 = hwndn2
	endIf
	return hwnd1
; TZapCommunicator is in each conversation in Compact view,
; but in Default view, it's at the same level as the conversation windows.
elif which == SWT_Communicator
	hwnd1 = findWindow(hMain, "TZapCommunicator", "")
	if hwnd1
		return hwnd1
	endIf
endIf
if which == SWT_ChatPrefix || which == SWT_CallQuality
	; Conversations can be part of the Skype window tree or their own independent tree.
	hMain = getTopLevelWindow(getFocus())
	if !Skype__appwin__isChatWindow(hMain)
		hMain = Skype__appwin__getActiveChatWindow(hMain)
	endIf
	if !hMain
		return 0
	endIf
	if which == SWT_ChatMain
		return hMain
	endIf
	if which == SWT_ChatInput
		return Skype__winfinder__findWindowByClassSubstring(hMain, "TChatRichEdit")
	elif which == SWT_ChatList
		return findWindow(hMain, "TParticipantList", "")
	elif which == SWT_ChatContact
		hwnd = findWindow(hMain, "TContactProfile", "")
		if !hwnd
			hwnd = findWindow(hMain, "TConversationProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hMain, "TParticipantList", "")
		endIf
		return hwnd
	elif which == SWT_ChatOutput
		return findWindow(hMain, "TChatContentControl", "")
	elif which == SWT_CallQuality
		return findWindow(hMain, "TCallQualityControlHost", "")
	endIf
	return 0
endIf
return 0
endFunction

object function Skype__appwin__getMSAAObject(string which)
; Returns the requested MSAA object, which should be an SWT_* constant.
; SWT_MenuBar is special:  It returns the Skype menu bar object but has no corresponding purpose in Skype__appwin__getSkypeWindow().
var
	object o, int childID, int n,
	handle hwnd
if Skype__conv5__applies()
	; In Skype 5.0 Gold, these have MSAA but no window.
	if which == SWT_ContactList
		return Skype__conv5__listByName(Skype__conv5__ListContacts)
	elif which == SWT_LiveConversations
		return Skype__conv5__listByName(Skype__conv5__ListLive)
	elif which == SWT_ConversationList
		return Skype__conv5__listByName(Skype__conv5__ListConversations)
	elif which == SWT_History
		return Skype__conv5__listByName(Skype__conv5__ListHistory)
	elif which == SWT_ChatOutput
		return Skype__conv5__chatContent(Skype__appwin__getSkypeWindow(SWT_ChatOutput))
	endIf
endIf
if which != SWT_MenuBar
	hwnd = Skype__appwin__getSkypeWindow(which)
	if !hwnd
		return 0
	endIf
	o = getObjectFromEvent(hwnd, -4, 0, childID)  ; Gets the Role_System_Client (or more specific role) object for this window
	return o
endIf
hwnd = Skype__appwin__getMainSkypeWindow()
if !hwnd
	return 0
endIf
o = getObjectFromEvent(hwnd, 0, 0, childID)  ; Gets the Role_System_Window object for this window
if !o
	return 0
endIf
return o.accChild(-3)  ; application menu bar
endFunction

object function Skype__appwin__getContactListObject()
; Returns the MSAA object for the contact list, if it can be found.
var
	handle hwnd,
	object o, int childID,
	object nullObject

; Clear out the cached object if Skype is not loaded (Skype__appwin__getMainSkypeWindow does this).
if !Skype__appwin__getMainSkypeWindow()
	return nullObject
endIf

; Return the cached object if it is valid.
if Skype__appwin___contactListObject
	if Skype__appwin___contactListObject.accRole(0) == Role_System_List
	|| Skype__appwin___contactListObject.accRole(0) == Role_System_Pane
		return Skype__appwin___contactListObject
	elif Skype__appwin___contactListObject.accRole(0) == Role_System_Client
		return nullObject
	endIf
endIf

Skype__appwin___contactListObject =
Skype__conv5__listByName(Skype__conv5__ListContacts)
if Skype__appwin___contactListObject
	return Skype__appwin___contactListObject
endIf

hwnd = Skype__appwin__getSkypeWindow(SWT_ContactList)
if hwnd
	o = getObjectFromEvent(hwnd, 0-4, 0, childID)
	if !o
		Skype__appwin___contactListObject = nullObject
		return nullObject
	elif o.accRole(0) != Role_System_List && o.accRole(0) != Role_System_Pane
		Skype__appwin___contactListObject = nullObject
		return nullObject
	endIf
	Skype__appwin___contactListObject = o
	return Skype__appwin___contactListObject
endIf
Skype__appwin___contactListObject = nullObject
return nullObject
endFunction

;============================== End module appwin ==============================


;================================== Module tb ==================================
; Skype toolbar handling code.
;
; Usage:
;	Call Skype__tb__makeList(names, data) (strings) to get toolbar item names and info.
;	Call Skype__tb__clickItem(seg) to activate an item (seg from data).
;
; TODO: Depends on a number of out-of-module constants.
;
; Author:  Doug Lee

string function Skype__tb___cleanOffLeadingGraphic(string s)
; Clean off a leading graphic if there is one (e.g., "graphic 327") from the beginning of the given string, and return the result.
var
	int n, int len, int i
if s == Skype__tb___JAWSGraphic
	s = stringTrimLeadingBlanks(stringChopLeft(s, stringLength(Skype__tb___JAWSGraphic)))
	len = stringLength(s)
	i = 1
	while i <= len && stringContains("0123456789", substring(s, i, 1))
		i = i +1
	endWhile
	if i > 1
		s = stringTrimLeadingBlanks(stringChopLeft(s, i-1))
	endIf
endIf
return s
endFunction

int function Skype__tb___addGraphicButtonsWithGivenNames(string byRef names, string byRef data, int n, string givenNames)
; Adds graphic buttons to names and data.
; Assumes the JAWS or (preferably) invisible cursor is on the first button and that buttons can be reached successively via NextWord().
; Also assumes all graphics stop the cursor; use OPT_Include_Graphics=2 for best results.
; Returns the number of buttons successfully added.
var
	int i
i = 1
while i <= n
	if !getGraphicID()
		beep()
		i = i-1  ; don't count the current "icon" that really isn't one.
		Skype__log__write(formatString("Ran out of graphics, found %1 of %2",
			intToString(i), intToString(n)
		))
		return i
	endIf
	names = Skype__string__join(names, "|", stringSegment(givenNames, "|", i))
	; Window handle, x/y of natural JAWS/invisible cursor position for graphic, childID of 0.
	data = Skype__string__join(data, "|",
			intToString(getCurrentWindow()) +","
			+intToString(getCursorCol()) +"," +intToString(getCursorRow()) +",0"
			)
	nextWord()
	i = i +1
endWhile
endFunction

int function Skype__tb___addWindow(string byRef names, string byRef data, handle hwnd, string givenName)
; Add the given window, as a named toolbar item, to names and data.
; Assumes the givenName includes all type/value/state information desired; nothing is drawn from the window itself except coordinates.
; Returns 1 if successful and 0 if not.
names = Skype__string__join(names, "|", givenName)
; Window handle, x/y of center of window, childID of 0.
data = Skype__string__join(data, "|",
		intToString(hwnd) +","
		+intToString((getWindowLeft(hwnd)+getWindowRight(hwnd))/2) +","
		+intToString((getWindowTop(hwnd)+getWindowBottom(hwnd))/2) +",0"
		)
return 1
endFunction

int function Skype__tb___addIconsWithTextLabels(string byRef names, string byRef data, handle hwnd, string nameSuffix)
; Adds all icons with text labels in the given window to names and data.
; Grayed-out items are still listed but shown as disabled.
; nameSuffix, if set, is appended to each item's text.
; Returns the number of items successfully added.
var
	int oldRestriction,
	int done,
	string buf,
	int nitems, int safetyCount
saveCursor() invisibleCursor() saveCursor()
moveToWindow(hwnd)
oldRestriction = getRestriction()
if !Skype__util__setRestrictionToWindow(hwnd)
	return 0
endIf
JAWSPageUp() JAWSHome()
nextChunk() priorChunk() priorChunk()  ; Make sure we're squarely on the first one
done = 0
safetyCount = 0
while !done && getCurrentWindow() == hwnd && safetyCount < 50
	buf = Skype__tb___cleanOffLeadingGraphic(stringTrimLeadingBlanks(getChunk()))
	buf = stringTrimLeadingBlanks(stringTrimTrailingBlanks(buf))
	if stringLength(buf)
		if nameSuffix
			buf = buf +" " +nameSuffix
		endIf
		names = Skype__string__join(names, "|", buf)
		if getColorName(getColorText()) == Skype__tb___JAWSGray
			; Commented out 2007-07-10 because chat toolbar items are grey when enabled.
			;names = names +" " +Skype__tb___msgDisabled
		endIf
		; Window handle, x/y of current position on text label, childID of 0.
		; A little offset is added to the x coordinate to avoid missing the button.
		data = Skype__string__join(data, "|",
				intToString(hwnd) +","
				+intToString(getCursorCol()+5) +"," +intToString(getCursorRow()) +",0"
				)
		nitems = nitems +1
	endIf
	done = !nextChunk()
	safetyCount = safetyCount +1
endWhile
if safetyCount >= 50
	Skype__log__write("Loop ran out.")
endIf
setRestriction(oldRestriction)
return nitems
endFunction

int function Skype__tb___addMSAAIconsWithGivenNames(string byRef names, string byRef data, handle hwnd, string givenNames)
; Adds all MSAA toolbar items in the given window to names and data.
; Returns the number of items successfully added.
var
	object o, int childID,
	int i, int nitems, int nreal,
	string newname
o = getObjectFromEvent(hwnd, 0-4, 0, childID)
nitems = o.accChildCount
nreal = 0
i = 1
while i <= nitems
	newname = Skype__sutil__getToolBarItemDesc(stringSegment(givenNames, "|", i), o, i, False)
	if newname
		names = Skype__string__join(names, "|", newname)
		; Window handle, 0,0, childID.
		data = Skype__string__join(data, "|",
				intToString(hwnd) +",0,0,"
				+intToString(i)
				)
		nreal = nreal +1
	endIf
	i = i +1
endWhile
return nreal
endFunction

int function Skype__tb___addContactGroupItems(string byRef names, string byRef data, handle hwnd)
var
	int oldRestriction,
	int x, int y
saveCursor() invisibleCursor() saveCursor()
moveToWindow(hwnd)
oldRestriction = getRestriction()
setRestriction(restrictWindow)
JAWSPageUp() JAWSHome()
y = getCursorRow()
names = Skype__string__join(names, "|", Skype__tb___msgContactButtons)
x = getWindowRight(hwnd)
data = Skype__string__join(data, "|", intToString(hwnd) +"," +intToString(x-30) +"," +intToString(y) +",0")
data = Skype__string__join(data, "|", intToString(hwnd) +"," +intToString(x-15) +"," +intToString(y) +",0")
setRestriction(oldRestriction)
return True
endFunction

globals
	string Skype__tb___names,
	string Skype__tb___data

int function Skype__tb___enumHelper(handle hwnd)
; Window enumeration helper for Skype__tb__makeList.
var
	object o, int childID,
	int iRole, string sRole, int iState,
	int n,
	string name
if !isWindowVisible(hwnd)
	return True
elif getWindowClass(hwnd) == "TSkMoreButton"
	var string wname
	wname = getWindowText(hwnd, Read_Everything) +" button"
	if isWindowDisabled(hwnd)
		wname = wname +" " +Skype__tb___msgDisabled
	endIf
	Skype__tb___addWindow(Skype__tb___names, Skype__tb___data, hwnd, wname)
endIf
o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	return True
endIf
iRole = o.accRole(0)
if !iRole
	; "Give Feedback" button and Call Phones/Directory/Shop.
	Skype__tb___addIconsWithTextLabels(Skype__tb___names, Skype__tb___data, hwnd, "button")
elif iRole != Role_System_Pane
	return True
endIf
n = o.accChildCount
childID = 1
while childID <= n
	iRole = o.accRole(childID)
	if iRole != Role_System_Window
		iState = o.accState(childID)
		if !(iState & (State_System_Invisible +State_System_Offscreen))
		&& Skype__sutil__isUnobscured(o, childID)
			name = Skype__sutil__removeHTML(o.accName(childID))
			if !name
				name = Skype__MSAA__rectText(o, childID)
			endIf
			if !name
				name = getWindowName(hwnd)
				if !name
					name = getWindowClass(hwnd)
					if stringLeft(name, 1) == "t"
					&& stringContains("ABCDEFGHIJKLMNOPQRSTUVWXYZ", substring(name, 2, 1))
						name = stringChopLeft(name, 1)
						Skype__string__replace(name, "control", "", "e")
					endIf
				endIf
				name = formatString("%1:%2", name, intToString(childID))
			endIf
			if iState & State_System_Unavailable
				name = name +" " +Skype__tb___msgDisabled
			endIf
			sRole = getRoleText(iRole)
			; TODO: This is English-specific but probably can't meaningfully be localized just by making string constants.
			if sRole == "push"
				sRole = stringChopLeft(sRole, 4)
			endIf
			name = name +" " +sRole
			Skype__tb___names = Skype__string__join(Skype__tb___names, "|", name)
			; Window handle, 0,0, childID.
			Skype__tb___data = Skype__string__join(Skype__tb___data, "|",
					intToString(hwnd) +","
					+"0,0," +intToString(childID)
					)
		endIf  ; visible and on-screen items
	endIf  ; non-Role_System_Window items
	childID = childID +1
endWhile  ; childIDs in window
return True
endFunction

int function Skype__tb__makeList(string byRef names, string byRef data)
; Makes a list of buttons/controls based on the current location within Skype.
; This function is responsible for including all toolbar buttons and any special exceptions.
; Returns the number of items in the list.
var
	handle hTop
Skype__tb___names = ""
Skype__tb___data = ""
hTop = getTopLevelWindow(getFocus())
enumerateChildWindows(hTop, "Skype__tb___enumHelper")
names = Skype__tb___names
data = Skype__tb___data
Skype__tb___names = ""
Skype__tb___data = ""
return Skype__string__segmentCount(names, "|")
endFunction

int function Skype__tb__clickItem(string data)
; Click (or otherwise activate) the button/control indicated by the given data string.
; Returns True on success and False on failure.
var
	handle hwnd, int x, int y,
	object o, int childID,
	int left, int right, int top, int bottom
hwnd = stringToHandle(stringSegment(data, ",", 1))
x = stringToInt(stringSegment(data, ",", 2))
y = stringToInt(stringSegment(data, ",", 3))
childID = stringToInt(stringSegment(data, ",", 4))
if childID
	o = getObjectFromEvent(hwnd, 0-4, childID, childID)
	; o.accDoDefaultAction(childID) does nothing in this app.
	if Skype__MSAA__rect(o, childID, left, right, top, bottom)
		saveCursor() JAWSCursor() saveCursor()
		moveTo( (left+right)/2, (top+bottom)/2 )
		leftMouseButtonLock()
		delay(1)
		leftMouseButtonLock()
		return True
	endIf
else
	saveCursor() JAWSCursor() saveCursor()
	moveTo(x, y)
	leftMouseButton()
	return True
endIf
return False
endFunction

;================================ End module tb ================================


;================================= Module conv5 ================================
; Utilities for managing the Skype 5 beta 2 and Gold, and (hopefully) later conversations control.
; Also includes a function to obtain the chat output message list in these Skype versions.

globals
; The main conversations window.
	handle Skype__conv5___hwndConv,
; A flag indicating whether this window has been found to be compatible with this module.
; -1 means no, 1 means yes, 0 means don't know yet.
	int Skype__conv5___compat,
; Cached list collections for live, contacts, conversations, and history.
; Key is a constant passed to Skype__conv5__listByName(),
; value is an MSAA object.
	collection Skype__conv5___listCache,
; RealName-to-scriptName map for the cache.
; Key is an actual MSAA list's name,
; value is the constant from this module used to find the MSAA object in the cache.
	collection Skype__conv5___listNames

int function Skype__conv5__applies()
; Returns True if this appears to be a Skype version supporting this module's interface.
var variant null
var handle hwnd = Skype__conv5___mainWindow()
if !hwnd
	Skype__conv5___listCache = null
	Skype__conv5___listNames = null
	return False
endIf
if hwnd != Skype__conv5___hwndConv
	Skype__conv5___compat = 0
	Skype__conv5___listCache = null
	Skype__conv5___listNames = null
endIf
if Skype__conv5___compat == 1
	return True
elif Skype__conv5___compat == -1
	return False
endIf
var int childID
var object o = getObjectFromEvent(hwnd, -4, 0, childID)
if !o
	return False
endIf
Skype__conv5___hwndConv = hwnd
var int n = o.accChildCount
if n > 20
	; That's permanent; we know this isn't compatible.
	Skype__conv5___compat = -1
	return False
endIf
childID = 1
while childID <= n
	if o.accRole(childID) == Role_System_List
		Skype__conv5___compat = 1
		return True
	endIf
	childID = childID +1
endWhile
; Don't mark permanently incompatible,
; in case this just means no list has yet been formed.
return False
endFunction

handle function Skype__conv5___mainWindow()
; Return the main conversations control window.
var handle hwnd = Skype__appwin__getSkypeWindow(SWT_Main)
if !hwnd
	return
endIf
hwnd = findWindow(hwnd, "TConversationsControl", "")
if !hwnd
	return
endIf
return hwnd
endFunction

object function Skype__conv5___mainClient()
; Return the main conversations control MSAA client object.
var handle hwnd = Skype__conv5___mainWindow()
if !hwnd
	return
endIf
var int childID
return getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
endFunction

string function Skype__conv5__peopleOnline()
; Return the string indicating how many people are online, if it can be found.
; The algorithm used is intended to be language independent.
if !Skype__conv5__applies()
	return ""
endIf
var object o = Skype__conv5___mainClient()
if !o
	return ""
endIf
var int childID = 1
var int n = o.accChildCount
if n > 20
	; Too much here to be right for Skype 5.
	return ""
endIf
while childID <= n
	if o.accRole(childID) == Role_System_StaticText
	&& Skype__conv5___isPeopleCountString(o.accName(childID))
		return o.accName(childID)
	endIf
	childID = childID +1
endWhile
return ""
endFunction

int function Skype__conv5___isPeopleCountString(string s)
; Returns True if s appears to represent the number of people online in Skype.
; The algorithm used is intended to be language independent.
var string digs = "0123456789"
if !stringContainsChars(s, digs)
	return False
endIf
var int i = 1
var int n = stringLength(s)
var int cnt = 0
var string ch
while i <= n
	ch = substring(s, i, 1)
	if stringContains(digs, ch)
		cnt = cnt +1
		if cnt > 4
			return True
		endIf
	else
		if cnt > 0 && !stringContains(digs, substring(s, i+1, 1))
			; Two separate numbers.
			return False
		endIf
	endIf
	i = i +1
endWhile
return False
endFunction

object function Skype__conv5__listByName(string listname)
; Return the list of given name or null.
if !Skype__conv5__applies()
	return
elif collectionItemExists(Skype__conv5___listCache, listname)
	return Skype__conv5___listCache[listname]
endIf
var object o = Skype__conv5___mainClient()
if !o
	return
endIf
var int childID = 1
var int n = o.accChildCount
while childID <= n
	if o.accRole(childID) == Role_System_List
	&& Skype__conv5___nameMatches(o.accName(childID), listname)
		if !Skype__conv5___listNames
			Skype__conv5___listCache = new collection
			Skype__conv5___listNames = new collection
		endIf
		var object oList = o.accChild(childID)
		listname = o.accName(childID)
		; Warning: Case is not significant here though it was before
		; use of JAWS 12+ collections.
		if collectionItemExists(Skype__conv5___listCache, listname)
			oList = Skype__conv5___listCache[listname]
		else
			Skype__conv5___listCache[listname] = oList
			Skype__conv5___listNames[o.accName(childID)] = listname
		endIf
		return oList
	endIf
	childID = childID +1
endWhile
return
endFunction

int function Skype__conv5___nameMatches(string name, string choices)
; Returns true if name is or is in choices.
if stringCompare(name, choices, False) == 0
	return True
elif stringContains("|"+choices+"|", "|"+name+"|")
	return True
endIf
return False
endFunction

string function Skype__conv5__focusedListName()
; Return the name of the focused list, or null if there isn't one.
var object o, int childID
o = getFocusObject(childID)
if !o
	return ""
endIf
var string name = o.accName(0)
if !name
	return ""
endIf
return Skype__conv5___listNames[name]
endFunction

string function Skype__conv5__focusShowingList()
; Focus the list that is already showing on screen.
; Returns True on success and False on failure.
var object o, int childID
o = Skype__conv5___mainClient()
if !o
	return False
endIf
var int n = o.accChildCount
childID = n
while childID
	if o.accRole(childID) == Role_System_List
		o.accSelect(SelFlag_TakeFocus, childID)
		return True
	endIf
	childID = childID -1
endWhile
return False
endFunction

object function Skype__conv5__chatContent(handle hwndChatOutput)
; Get the MSAA list object of chat messages from the given chat output window handle.
if !hwndChatOutput
	return
elif !Skype__conv5__applies()
	return
endIf
var int childID
var object o = getObjectFromEvent(hwndChatOutput, ObjID_Client, 0, childID)
if !o
	return
elif Skype__conv5___nameMatches(o.accName(1), Skype__conv5__ListChatOutput)
	return o.accChild(1)
elif Skype__conv5___nameMatches(o.accName(2), Skype__conv5__ListChatOutput)
	return o.accChild(2)
elif Skype__conv5___nameMatches(o.accName(3), Skype__conv5__ListChatOutput)
	return o.accChild(3)
endIf
return
endFunction

;=============================== End module conv5 ==============================


;================================= Module live5 ================================
; Code for handling the Skype 5.1-5.2 Live Call area.
;
; Author:  Doug Lee


collection function Skype__live5__getInfo(/*?*/ handle hwnd)
; Get an info collection describing the live area whose chat handle is given:
;	hwnd, obj - The window handle and MSAA Pane object for the area.
;	participants - A 1-based collection of participants.
;	muted - True if the call is muted, False if not.
;	held - "l" for local, "r" for remote, empty or missing otherwise.
;	videoOut, videoIn - "y" if video is being sent or received, "n" if it can be but isn't.
;	callQuality - An indicator of call quality issues.
;	duration - The current call duration as hh:mm:ss.
;	peopleSelector - The Add People popup area's Pane object if present.
;	toolbar - The area's toolbar's MSAA object.
; If hwnd is null or not given, the one containing the focused window is used.
var collection oInfo oInfo = new collection
var string rightClass = "TLiveConversation"
if !hwnd
	hwnd = getFocus()
endIf
hwnd = Skype__schat__getTopOfChat(hwnd)
if !hwnd
	return oInfo
endIf
hwnd = findWindow(hwnd, rightClass, "")
if !hwnd
	return oInfo
endIf
if getWindowClass(getFirstChild(hwnd)) == rightClass
	; JAWS matched TLiveConversationPanel with TLiveConversation.
	hwnd = getFirstChild(hwnd)
endIf
if getWindowClass(getParent(hwnd)) != rightClass
	return oInfo
endIf

var object oLive = Skype__MSAA__objectFromWindow(hwnd)
if !oLive
	return oInfo
endIf

oInfo.hwnd = hwnd
oInfo.obj = oLive
var collection cParticipants cParticipants = new collection
oInfo.participants = cParticipants

var object oToolbar
var string name, int iRole
var int callHeld = False
var int n = oLive.accChildCount
var int nparts = 0
; Count backward to find the CallToolBar before absorbing arbitrary accNames as participants.
var int childID = n
while childID
	name = oLive.accName(childID)
	iRole = oLive.accRole(childID)
	if iRole == Role_System_Pane
		if oToolbar
			nparts = nparts +1
			cParticipants[intToString(nparts)] = name
			; Skype 5.5 started putting the call duration for a non-conference
			; in the participant's pane.
			if childID == 1
				var object oPart = oLive.accChild(childID)
				var int nids = oPart.accChildCount
				var string maybeDuration = oPart.accName(nids)
				if oPart.accRole(nids) == Role_System_StaticText
				&& stringContains("0123456789", stringLeft(maybeDuration, 1))
					oInfo.duration = maybeDuration
				endIf
			endIf
		elif name == "CallToolbar"
			oToolbar = oLive.accChild(childID)
			oInfo.toolbar = oToolbar
		elif name == "LiveSelectContactsWidget"
			oInfo.peopleSelector = oLive.accChild(childID)
		endIf
	elif name == Skype__schat__scHold
		callHeld = True
	elif name == Skype__schat__scSnapshot
		oInfo.videoIn = "y"
	else
		; Skip this item.
	endIf
	childID = childID -1
endWhile
if oToolbar
	Skype__live5___processToolbar(oInfo, oToolbar, callHeld)
endIf
if !collectionItemExists(oInfo, "videoIn")
	oInfo.videoIn = "n"
endIf
return oInfo
endFunction

int function Skype__live5___processToolbar(collection oInfo, object oToolbar, int callHeld)
; Absorb info from the toolbar.
; If this turns out not to be a live area containing a call,
; oInfo is emptied.
; Returns True on success and False on failure.
if !oToolbar
	return False
endIf
var int n = oToolbar.accChildCount
var int childID = 1
var string name, int iRole
var int isCall = False
while childID <= n
	name = stringTrimLeadingBlanks(oToolbar.accName(childID))
	iRole = oToolbar.accRole(childID)
	if iRole == Role_System_StaticText
	&& stringContains("0123456789", stringLeft(name, 1))
		oInfo.duration = name
	elif name == Skype__schat__scMute
		oInfo.muted = False
	elif name == Skype__schat__scUnmute
		oInfo.muted = True
	elif name == Skype__schat__scResume
		; Locally held call.
		oInfo.held = "l"
	elif name == Skype__schat__scEndCall
		isCall = True
	elif name == Skype__schat__scMyVideo
		if (oToolbar.accState(childID) & State_System_Pressed)
			oInfo.videoOut = "y"
		else
			oInfo.videoOut = "n"
		endIf
	else
		; Skip this item.
	endIf
	childID = childID +1
endWhile
if not isCall
	oInfo.removeAll()
	return True
endIf
if callHeld && !collectionItemExists(oInfo, "held")
	; Remotely held call.
	oInfo.held = "r"
endIf
return True
endFunction

;=============================== End module live5 ==============================


;============================= Module live5selector ============================
; Code for handling the Skype 5.x Live Call area's Add People contact selector.
;
; Author:  Doug Lee

int function Skype__live5selector__selected()
; Returns a status indicating whether the current item in the list is selected:
;	0:  Focus is not in a Live Call's Add People selector list.
;	1:  Focus is in one but not on a contact entry.
;	2:  Focus is in one but the current item is not selected.
;	3:  The current item is selected.
; An item is selected if its accName appears in the accValue list in the widget's Edit box.
; That list is delimited by semicolons (;).
; There are checkbox-like things to the right of each item in the selector list,
; but they semi-transparently overlay avatars, making color-based algorithms iffy at best.
var string sep = ";"

; Set up based on current focus position.
var object oSelector, int childID
oSelector = getFocusObject(childID)
if !Skype__live5selector___isSelector(oSelector)
	return 0
elif oSelector.accRole(childID) != Role_System_ListItem
	; Right list, but a subitem like a Skype button.
	return 1
elif !oSelector.accValue(childID)
	; Item like the "Other Item" one at the top:
	; Not a contact because contacts have accValue content.
	return 1
endIf
var string curItem = oSelector.accName(childID)

; Find the childID of the Edit containing the selected item list.
; This is an element, not a full-blown Accessible Object.
var object oParent = oSelector.accParent
var int idSelections = 0
var int n = oParent.accChildCount
childID = 1
while childID <= n && !idSelections
	if oParent.accRole(childID) == Role_System_Text
		; We assume just one Edit here and so don't check its accName.
		idSelections = childID
	endIf
	childID = childID +1
endWhile
if !idSelections
	; Treat like there are no selections.
	return 2
endIf

; Exact containment test.
var string sSelections = oParent.accValue(idSelections)
if !sSelections
	; Nothing is selected.
	return 2
endIf
if stringContains(sep+sSelections+sep, sep+curItem+sep)
	return 3
endIf
return 2
endFunction

int function Skype__live5selector___isSelector(object o)
; Returns True if o is the Add People selector in a live call.
if o.accRole(0) == Role_System_List
&& o.accName(0) == "LiveCheckBoxUserListPhone"
&& o.accParent.accName(0) == "LiveSelectContactsWidget"
	return True
endIf
return False
endFunction

;=========================== End module live5selector ==========================


;================================= Module conf =================================
; Code for handling conference participants.
;
; Author: Doug Lee

collection function Skype__conf__getSpeakers()
; Get a list of who is speaking in any active conference.
; On error, the collection contains .err as follows:
;	1 - Not in a chat.
;	2 - Not in a live conversation.
;	3 - Not in a conference.
return Skype__conf__getParticipants(True)
endFunction

collection function Skype__conf__getParticipants(int justSpeakers)
; Get a list of who is participating or speaking in any active conference.
; On error, the collection contains .err as follows:
;	1 - Not in a chat.
;	2 - Not in a live conversation.
;	3 - Not in a conference.
var int tc = getTickCount()
var collection speakers speakers = new collection
var handle hTop = Skype__schat__getTopOfChat(getFocus())
if !hTop
	speakers.err = 1
	return speakers
endIf
hTop = findWindow(hTop, "TLiveConversation", "")
if !hTop
	speakers.err = 2
	return speakers
endIf
if getWindowClass(getFirstChild(hTop)) == "TLiveConversation"
	hTop = getFirstChild(hTop)
endIf

var int ntries = 3
var int done = False
while ntries && !done
	speakers = new collection
	Skype__conf___fillSpeakers(speakers, hTop, justSpeakers)
	if speakers.count > 0
		done = True
	else
		ntries = ntries -1
		if !justSpeakers
			ntries = 0
		endIf
		pause()
	endIf
endWhile
speakers.tc = getTickCount() -tc
if !speakers.count
&& Skype__MSAA__objectFromWindow(hTop).accChildCount < 2
	speakers.err = 3
endIf
return speakers
endFunction

int function Skype__conf___fillSpeakers(collection speakers, handle hwnd, int justSpeakers)
; Fill the given (presumed empty) speakers collection.
; Includes all participants unless justSpeakers is true,
; which includes all people in the conference.
if !isWindowVisible(hwnd)
|| getWindowClass(hwnd) != "TLiveConversation"
	return False
endIf
var object oTop = Skype__MSAA__objectFromWindow(hwnd)
if !oTop
	beep()
	return False
endIf
var int n = oTop.accChildCount
var int i
var collection speaker
var int nspeakers = 0
for i = 1 to n
	if Skype__conf___isSpeaking(oTop, i, justSpeakers)
		nspeakers = nspeakers +1
		speaker = new collection
		speaker.name = oTop.accName(i)
		speaker.text = oTop.accChild(1).accName(0)
		speaker.chatname = Skype__schat__getChatName(hwnd, False)
		speakers[intToString(nspeakers)] = speaker
	endIf
endFor
speakers.count = nspeakers
return True
endFunction

int function Skype__conf___isSpeaking(object o, int childID, int justSpeakers)
; Returns True if the given participant appears to be speaking.

; First make sure this is actually a participant, not a toolbar.
if o.accRole(childID) != Role_System_Pane
|| o.accChild(childID).accRole(1) != Role_System_StaticText
	return False
elif !justSpeakers
	return True
endIf

; Plan: For any visible participant pane, pick a pixel to check for blueness.
; If it is more than 50 brighter Blue than Red, consider this a speaking participant.
; Currently we use a pixel 6 from left center.
var int left, int right, int top, int bottom
if !Skype__MSAA__rect(o, childID, left, right, top, bottom)
	beep()
	return False
endIf
var int x = left +6
var int y = (top +bottom) / 2
var int clr = getColorAtPoint(x, y)
var int r = (clr & 0xFF)
var int b = ((clr >> 16) & 0xFF)
if b > r+50 && b > 150
	return True
endIf
return False
endFunction

;=============================== End module conf ===============================


;=============================== Module incsearch ==============================
; Code to allow incremental search in listboxes that don't allow it,
; including JAWS pop-up lists. Also makes Home/End work where necessary.
;
; Usage: When a control needing this support is in focus,
; call Skype__incsearch__gotoNext(strKeyName) from KeyPressedEvent.
; Call Skype__incsearch__setFlags() from autoStartEvent if necessary.
;
; Author:  Doug Lee of SSB BART Group

const
; These are not defined by older JAWS versions but are by newer ones,
; so to avoid compilation problems in all cases, we define local versions here.
	Skype__incsearch___SelFlag_TakeFocus = 1,
	Skype__incsearch___SelFlag_TakeSelection = 2

globals
; The search string entered so far at any given time.
	string Skype__incsearch___search,
; Flags that can be set for this system. See Skype__incsearch__setFlags() for details.
	string Skype__incsearch___flags

void function Skype__incsearch__setFlags(string flags)
; Set flags that govern how this system works. Flags defined:
;	l: Use ListView focus instead of MSAA focus to focus an item.
;	h: Implement Home/End navigation (for lists that don't already).
; Must call from autoStartEvent if any special flags are needed.
; Call from autoFinishEvent also to clear flags if any are set and
; this module is not given a unique namespace.
let Skype__incsearch___flags = stringLower(flags)
return Skype__incsearch___flags
endFunction

void function Skype__incsearch___clear()
; Ends an existing search.
let Skype__incsearch___search = ""
endFunction

void function Skype__incsearch__gotoNext(string keyname)
; Implements the incremental search feature using MSAA.
; Also implements Home=top and End=bottom navigations.
var
	object o, int childID,
	int i, int n, int n1
var int nav = 0
if stringLength(keyname) > 1 then
	; End searches when a non-alphanumeric key (e.g., Tab or Ctrl+A) is typed.
	Skype__incsearch___clear()
	if stringCompare(keyName, "home", False) == 0 then
		let nav = 1
	elif stringCompare(keyName, "end", False) == 0 then
		let nav = 2
	endIf
	if nav && stringContains(Skype__incsearch___flags, "h") then
		PCCursor()
		let o = getObjectAtPoint(childID, getCursorCol(), getCursorRow())
		let n = o.accChildCount
		if nav == 1 then
			Skype__incsearch___focusItem(o, 1)
		else  ; 2
			Skype__incsearch___focusItem(o, n)
		endIf
	endIf
	return
endIf

; If the key typed is not the same as the last key typed,
; or there is already a multi-character search underway,
; add the new key to the search.
if stringCompare(keyname, Skype__incsearch___search, False) then
	let Skype__incsearch___search = stringChopLeft(" " +Skype__incsearch___search +keyname, 1)  ; stringChopLeft avoids arithmetic addition
endIf
scheduleFunction("Skype__incsearch___clear", 10)  ; listViews seem to allow about 1 second too.

; Search for the string entered thus far, starting at the current list item.
PCCursor()
let o = getObjectAtPoint(childID, getCursorCol(), getCursorRow())
let n = o.accChildCount
if stringLength(Skype__incsearch___search) == 1 then
	if inHJDialog() then
		; HJ dialogs already do single-key searching, so don't mess with that.
		return
	endIf
	; For a single-key search, start on the next item.
	let i = childID +1
else
	; For an already multi-key search, allow the current item to continue matching.
	let i = childID
endIf
let n1 = 0
while n1 <= n
	; Search all items once, wrapping as necessary.
	if i > n then
		let i = 1
	endIf
	if o.accName(i) == Skype__incsearch___search then
		Skype__incsearch___focusItem(o, i)
		return
	endIf
	let i = i +1
	let n1 = n1 +1
endWhile
; Beep to indicate the search failed.
beep()
endFunction

void function Skype__incsearch___focusItem(object o, int childID)
; Select/focus the indicated item.
if childID == o.accFocus then
	; Repeat this item to let the user know we're still here.
	scheduleFunction("sayObjectActiveItem", 0)
else
	; Select/focus the found item.
	if inHJDialog() then
		o.accSelect(Skype__incsearch___SelFlag_TakeSelection, childID)
	elif stringContains(Skype__incsearch___flags, "l") then
		setCurrentItem(getFocus(), childID)
	else
		o.accSelect(Skype__incsearch___SelFlag_TakeFocus+Skype__incsearch___SelFlag_TakeSelection, childID)
	endIf
			refreshWindow(getFocus())
		endIf
endFunction

;============================= End module incsearch ============================


;================================ Module saysel ================================
; Code to make the SaySelectedText script work better in some MSAA-capable list controls.
;
; Author: Doug Lee

globals
	int Skype__saysel___vpcFlag

string function getSelectedText(int includeMarkup)
; Override to make saySelectedText work.
var
	handle hwnd,
	object o, int childID, int n,
	int tc,
	string buf
if !Skype__saysel___vpcFlag
	return getSelectedText(includeMarkup)
endIf
hwnd = getFocus()
tc = getWindowTypeCode(hwnd)
if tc == WT_Edit || tc == WT_MultiLine_Edit ||tc == WT_ComboBox
|| !Skype__sutil__isMSAASupport(hwnd)
	return getSelectedText(includeMarkup)
endIf
o = getObjectFromEvent(hwnd, -4, 0, childID)
childID = 0
n = o.accChildCount
while childID < n
	childID = childID +1
	if o.accState(childID) & State_System_Selected
		buf = Skype__string__join(buf, "\r\b", Skype__string__ref(o.accName(childID)))
	endIf
endWhile
return buf
endFunction

int function isVirtualPCCursor()
if Skype__saysel___vpcFlag
	return True
endIf
return isVirtualPCCursor()
endFunction

script saySelectedText()
Skype__saysel___vpcFlag = True
performScript saySelectedText()
Skype__saysel___vpcFlag = False
endScript

;============================== End module saysel ==============================


;================================== Module nav =================================
; Code for moving among Skype control groups.

script NextDocumentWindow()
Skype__util__waitForKeysToClear(15)
Skype__nav__switchDocumentWindows(True)
endScript

script PreviousDocumentWindow()
Skype__util__waitForKeysToClear(15)
Skype__nav__switchDocumentWindows(False)
endScript

script nextControlGroup()
; Switches to the next major control group in the Skype main window or in the currently focused conversation.
Skype__util__waitForKeysToClear(15)
Skype__nav__switchControlGroups(True)
endScript

script PreviousControlGroup()
; Switches to the previous major control group in the Skype main window or in the currently focused conversation.
Skype__util__waitForKeysToClear(15)
Skype__nav__switchControlGroups(False)
endScript

void function Skype__nav__switchDocumentWindows(int isNext)
; Implements round-robin conversation window switching.
var
	int doChats
if Skype__util__isSpecialFocus()
	doChats = False
elif Skype__schat__isChatWindow(getFocus(), False)
	doChats = True
elif Skype__appwin__isMainSkypeWindow(getTopLevelWindow(getFocus()))
	; Allows Ctrl+[Shift+]Tab to jump from main window to a conversation.
	doChats = True
else
	doChats = False
endIf

if !doChats
	if isNext
		performScript nextDocumentWindow()
	else
		performScript previousDocumentWindow()
	endIf
	return
endIf

if isNext
	isNext = 1
else
	isNext = -1
endIf
Skype__schat__switchToChat(False, isNext)
endFunction

void function Skype__nav__switchControlGroups(int isNext)
; Implements round-robin switching among control groups.
if Skype__util__isSpecialFocus()
	; Do nothing.
	beep()
	return
endIf
speechOff()
; If the chat entry box is focused, move focus to its parent first.
; Otherwise, Shift+F6 (previous group) gets stuck in that box.
var handle hwnd = getFocus()
if getWindowClass(getParent(hwnd)) == "TChatEntryControl"
	setFocus(getParent(hwnd))
	pause()
endIf
var int done = False
var int safety
while !done
	hwnd = getFocus()
	if isNext
		Skype__util__setFocus(0, "n")
	else
		Skype__util__setFocus(0, "p")
	endIf
	safety = 10
	while safety && getFocus() == hwnd
		pause()
		safety = safety -1
	endWhile
	hwnd = getFocus()
	if Skype__nav___isFocusWorthy(hwnd)
		done = True
	else
		stopSpeech()
	endIf
endWhile
if getWindowClass(hwnd) == "TChatEntryControl"
	; Focus the edit part of this one.
	var handle hwnd1
	hwnd1 = findWindow(hwnd, "TChatRichEdit", "")
	if !hwnd1
		hwnd1 = findWindow(hwnd, "TChatRichEdit.UnicodeClass", "")
	endIf
	if hwnd1
		setFocus(hwnd1)
		pause()
	else
		beep()
	endIf
else
	; Speak a name when possible and make sure focus is somewhere sensible.
	var object o = Skype__MSAA__objectFromWindow(hwnd)
	if o
		speechOn()
		sayMessage(OT_Dialog_Name, o.accName(0))
		speechOff()
		var variant v = o.accFocus
		if o == v || v == 0
			var int i = 1
			var int n = o.accChildCount
			while i <= n
				if o.accState(i) & State_System_Focusable
					o.accSelect(SelFlag_TakeFocus, i)
					pause()
					i = n
				endIf
				i = i +1
			endWhile
		endIf
	endIf
endIf
speechOn()
scheduleFunction("sayObjectTypeAndText", 1)
endFunction

int function Skype__nav___isFocusWorthy(handle hwnd)
; Returns True if hwnd appears to be a window worthy of receiving focus.
if getWindowRight(hwnd) -getWindowLeft(hwnd) < 5
|| getWindowBottom(hwnd) -getWindowTop(hwnd) < 5
	return False
endIf
return True
endFunction

;================================ End module nav ===============================


;================================ Module sayall ================================
; Support for SayAll in Skype chat message lists.
;
; Author: Doug Lee

globals
; Info used by the SayAll code in this module while a SayAll is in progress.
	collection Skype__sayall___data

void function sayAllStarted()
; This function is called by the built in script function SayAll just before
; the Speak method of the SAYALL object is called for the first time.  It is
; intended that this function be overridden in a given application's scripts
; to provide a method of performing actions (such as setting JCF options)
; just before a Say All is started.  This complements the preexisting
; SayAllStoppedEvent  which is called whenever a Say All process is stopped.
; (This text was drawn from default.jss in JAWS 11-15.)
var variant null
var handle hwnd = getFocus()
if getWindowClass(hwnd) != "TChatContentControl"
	Skype__sayall___data = null
	return
endIf
Skype__sayall___data = new collection
var collection data = Skype__sayall___data
data.hwnd = hwnd
var object oMsg, int childID
oMsg = getFocusObject(childID)
if childID
	; Unexpected; oMsg should be a message object.
	Skype__sayall___data = null
	return
endIf
Skype__sayall___data.msgs = oMsg.accParent
data.index = Skype__schat__getMessageIndex(oMsg)
if !data.index
	; If we don't know where to start, we don't know where to go.
	Skype__sayall___data = null
	return
endIf
data.count = Skype__schat__getMessageCount(data.msgs)
; That count may increase as messages are read.
endFunction

int function SayAllSpeakUnit(int iSayAllMode, int iSpeakNext)
if !Skype__sayall___data
	; This is not our area, so let internal code handle this.
	return SayAllSpeakUnit(iSayAllMode, iSpeakNext)
endIf
var collection data = Skype__sayall___data
if iSpeakNext
	data.index = data.index +1
	if data.index > data.count
		; In case messages came in during reading.
		data.count = Skype__schat__getMessageCount(data.msgs)
	endIf
	if data.index > data.count
		return saSayAllEndOfDocument
	endIf
	; This should not be necessary but is; see SayAllRelocateCursor().
	typeKey("downArrow")
endIf
Skype__sayall___speakMessage()
return saSayAllUnitSpoken
endFunction

void function SayAllRelocateCursor(int iSayAllMode,int iUnitsToRetreat,int iWordsSpokenInCurrentUnit)
if !Skype__sayall___data
	; This is not our area, so let internal code handle this.
	return SayAllRelocateCursor(iSayAllMode,iUnitsToRetreat,iWordsSpokenInCurrentUnit)
endIf
var collection data = Skype__sayall___data
if iUnitsToRetreat
	data.index = data.index -iUnitsToRetreat
	; Focusing this should be enough, but even though the focus takes,
	; the next Up/Down moves relative to the previous focus.
	; Selecting does not seem to help.
	; On high-DPI machines, clicking is not safe either.
	; We therefore have to send arrows to move focus.
	; [DGL, 2014-09-14, Skype 6.18.59.106.]
	speechOff()
	var int i = iUnitsToRetreat
	while i
		typeKey("upArrow")
		i = i(-1)
	endWhile
	pause()
	speechOn()
endIf
var object msgs = data.msgs
;msgs.accSelect(SelFlag_TakeFocus, data.index)
; speak message on which cursor lands when SayAll is stopped by user.
Skype__sayall___speakMessage()
;sayInteger(iUnitsToRetreat)
;sayInteger(iWordsSpokenInCurrentUnit)
endFunction

void function SayAllStoppedEvent()
var variant null
Skype__sayall___data = null
sayAllStoppedEvent()
endFunction

void function Skype__sayall___speakMessage()
var collection data = Skype__sayall___data
; This honors timestamp user setting.
var string txt = Skype__schat__getMessageText(data.msgs, data.index, "")
sayMessage(OT_SayAll, txt)
endFunction

;============================== End module sayall ==============================


;======================== Module sutil (from skypeUtil) ========================
; Utility code for the Skype scripts.
;
; Author:  Doug Lee

object function Skype__sutil__getFocusObject(int byRef childID)
; A GetFocusObject wrapper that accounts for Skype's sometimes returning the wrong thing.
var
	object o,
	variant v
o = getFocusObject(childID)
v = o.accFocus
if !stringLength(v) && v
	o = v
	childID = 0
elif v
	childID = v +0
endIf
return o
endFunction

int function Skype__sutil__isSkypeWatchLoaded()
; Returns True if SkypeWatch is running:
;	- 1 if SkypeWatch is running via the old (Use "SkypeWatch.jsb" in default.jss) installation method.
;	- 2 if SkypeWatch is running via the new method (default.jss is loader of SkypeWatch and default_fs.jsb).
; The two are distinguished from one another by whether SkypeWatche's WindowCreatedEvent has been called since the last SkypeWatch restart.
; SkypeWatch_lastTick is updated by SkypeWatch::SkypeWatcher(), which runs once every five seconds and once two seconds after SkypeWatch first loads.
; If it hasn't been updated in 15 seconds, SkypeWatch is surely not around to do it.
; For some reason though, the delay has been seen to exceed 8 seconds. [DGL, 2005-02-08]
var
	int diff
diff = getTickCount() -SkypeWatch_lastTick
;sayInteger(diff)
if diff < 15000
	if SkypeWatch_WCECalled
		return 2
	endIf
	return 1
endIf
return 0
endFunction

globals int Skype__sutil___saidUnsupported
void function Skype__sutil__sayUnsupportedSkype(/*?*/ int maxTimes)
; Announce that this Skype version is not supported by these scripts.
; Only make the announcement once by default,
; or up to as many times as maxTimes indicates otherwise.
; A beep is produced on every call regardless of maxTimes.
if maxTimes < 1
	maxTimes = 1
endIf
if maxTimes > Skype__sutil___saidUnsupported
	Skype__sutil___saidUnsupported = Skype__sutil___saidUnsupported +1
	sayMessage(OT_Error, Skype__sutil___msgNotSupported)
endIf
beep()
endFunction

void function Skype__sutil__sayMessageToUserBuffer(string s, int format)
var
	int ot
s = s +SU_msgBuffEnd
ot = OT_User_Buffer
if format
	sayFormattedMessage(ot, s)
else
	sayMessage(ot, s)
endIf
endFunction

void function Skype__sutil__sayURMessage(string sLong, string sShort, int format)
; Similar to sayMessageToUserBuffer:  Allows JAWS 4.x/5.x/6.x to do things the same way.
; A message sent this way will flash in Braille in JAWS 6+.
var
	int wasSpeechOff,
	int ot
wasspeechOff = isSpeechOff()
if wasSpeechOff
	speechOn()
endIf
ot = OT_User_Requested_Information
if sShort
	if format
		sayFormattedMessage(ot, sLong, sShort)
	else
		sayMessage(ot, sLong, sShort)
	endIf
else
	if format
		sayFormattedMessage(ot, sLong)
	else
		sayMessage(ot, sLong)
	endIf
endIf
if wasSpeechOff
	speechOff()
endIf
endFunction

int function Skype__sutil__isMSAASupport(handle hwnd)
; Determines whether the given window provides MSAA support.
var
	object o, int childID, int iRole
o = getObjectFromEvent(hwnd, -4, 0, childID)
if !o
	return False
endIf
iRole = o.accRole(childID)
if iRole == Role_System_Client || iRole == Role_System_Window
	return False
endIf
return True
endFunction

string function Skype__sutil__getToolBarItemDesc(string givenName, object o, int childId, int includeSeps)
; Construct and return the displayable name/description for the indicated toolbar item.
; GivenName, if non-null, will be used as the name instead of whatever is found on screen or in MSAA.
; As a special case, if givenName contains a slash and the button is a check box,
; the part left of the slash will be used as the name if the box is unchecked, and the right part if checked.
; includeSeps determines whether separators will return something meaningful or a null string.
; Author:  Doug Lee
var
	int iRole, int iState,
	int left, int right, int top, int bottom,
	string name, string states,
	int altnames
iRole = o.accRole(childId)
if iRole == Role_System_Window
	return ""
elif iRole == Role_System_Separator
	if includeSeps
		if givenName
			return givenName
		else
			return SU_msgSep
		endIf
	else
		return ""
	endIf
endIf
states = ""
name = givenName
if !name
	name = o.accName(childId)
endIf
if !name
	; Example:  WordPad:  No accName but labeling graphics puts names in the object's rectangle.
	if Skype__MSAA__rect(o, childId, left, right, top, bottom)
		name = getTextInRect(left, top, right, bottom)
	endIf
endIf
if !name
	; TODO:  When separators are skipped above, item numbers here will skip also.
	name = formatString(SU_msgUnnamedItem, intToString(childId))
	states = states +" " +SU_msgUnnamed
endIf
name = name +" " +getRoleText(iRole)
iState = o.accState(childId)
if iState & State_System_Unavailable
	states = states + " " +SU_msgUnavailable
endIf
if iState & State_System_OffScreen
	states = states + " " +SU_msgOffScreen
elif iState & State_System_Invisible
	states = states + " " +SU_msgInvisible
endIf
if iRole == Role_System_CheckButton || iRole == Role_System_RadioButton
	if iRole == Role_System_CheckButton
		if stringContains(name, "/")
			; When given alternate names, we'll treat it like a toggle button, not a check box.
			altnames = True
		else
			states = " " +SU_msgCheckBox +states
		endIf
	elif iRole == Role_System_CheckButton
		states = " " +SU_msgRadioButton +states
	endIf
	if iState & State_System_Checked
		if altnames
			name = stringSegment(name, "/", 2)
		else
			states = states +" " +SU_msgChecked
		endIf
	else
		if altnames
			name = stringSegment(name, "/", 1)
		else
			states = states +" " +SU_msgNotChecked
		endIf
	endIf
elif iRole == Role_System_Separator
endIf
if states
	name = name +" -" +states
endIf
return name
endFunction

string function Skype__sutil__myGetWindowText(handle hwnd)
; Returns the text in the given window with line breaks and without graphics.
var
	int oldG,
	string buf
oldG = getJCFOption(Opt_Include_Graphics)
setJCFOption(Opt_Include_Graphics, 0)
buf = GetTextInRect (getWindowLeft(hwnd), getWindowTop(hwnd), getWindowRight(hwnd), getWindowBottom(hwnd), 0, IgnoreColor, IgnoreColor, True)
setJCFOption(Opt_Include_Graphics, oldG)
return buf
endFunction

int function Skype__sutil__versionCompare(string vcode)
; Return an integer indicating how the actual Skype version compares to the given one:
; -1 means Skype is older, 0 means they're equal, and 1 means Skype is newer.
; -2 means the real Skype version code is not known.
; vcode: A string Skype version, like 5.10.
if stringLeft(vcode, 1) != "v"
	vcode = formatString("v%1", vcode)
endIf
var string real = Skype__sutil__getSkypeVersionCode(False)
if !real
	return -2
elif stringCompare(real, vcode, True) == 0
	return 0
endIf
vcode = stringChopLeft(vcode, 1)
real = stringChopLeft(real, 1)
var int n = max(
	Skype__string__segmentCount(real, "."),
	Skype__string__segmentCount(vcode, ".")
)
var int i = 1
var string vseg, string rseg, int vint, int rint
while i <= n
	vseg = stringSegment(vcode, ".", i)
	if !stringLength(vseg)
		; This means versionCompare("5.9") returns equal for Skype 5.9.x.
		return 0
	endIf
	rseg = stringSegment(real, ".", i)
	if !stringLength(rseg)
		; versionCompare("5.9.1.234") returns -1 if Skype's version code only comes back "5.9."
		return -1
	endIf
	vint = stringToInt(vseg)
	rint = stringToInt(rseg)
	if vint > rint
		return -1
	elif vint < rint
		return 1
	; Else equal, so keep looking.
	endIf
	i = i +1
endWhile
; And if the loop ran out, all is equal.
return 0
endFunction

globals
	int Skype__sutil___vcLastTicks,
	string Skype__sutil___vcLastVer
string function Skype__sutil__getSkypeVersionCode(int force)
; Returns a code that indicates what Skype version is running on this machine.
; Codes returned will be exact (if possible) or approximate version numbers of Skype.exe.
; Approximations will always be greater than or equal to the actual version number of skype.exe and will be suffixed with "-" to indicate that they are approximations.
; Returns the null string if the version code can't be obtained or approximated.
; The returned code is prefixed with "v" because this makes the < and > operators behave reasonably enough with version codes.
; WARNING:  "Reasonably enough" means functional but still odd: "v1.3" == "v1.3.2"; == is apparently tried in JAWS before < and >.
; Note: JAWS' GetVersionInfoStrings function is unable to return the real version code from some Skype versions.
; In such a case, the Scripting.FileSystemObject is used to get it.
; Note:  In JAWS versions older than 7.10, the full version code can't be obtained except when Skype is in focus.
; This code is also used by SkypeWatch, so that's not practical in all cases.
; The plan is to get the real code when possible and cache it wisely (I hope).
var
	handle hwnd,
	int ticks, int tickDiff,
	string appPath,
	object o

; Only actually check for the version every 10 seconds, to minimize work.
; This function can be called very frequently (e.g., if used from FindCaret), so this removes the need to worry about that.
ticks = getTickCount()
tickDiff = ticks -Skype__sutil___vcLastTicks
if !force && tickDiff <= 10000 && stringLength(Skype__sutil___vcLastVer)
	return Skype__sutil___vcLastVer
endIf
Skype__sutil___vcLastTicks = ticks

; If possible, get the application path, which allows us to get the real version number.
; This requires some part of Skype to be in focus.
appPath = ""
appPath = Skype__string__ref(getWindowOwner(Skype__appwin__getSkypeWindow(SWT_Main)))
if !appPath
	appPath = getAppFilePath()
	if !stringContains(stringLower(appPath), "skype.exe")
		appPath = ""
	endIf
endIf

; If we have a path, use it.
if appPath
	; First try JAWS' way, since it seems to work in later Skypes again.
	Skype__sutil___vcLastVer = getVersionInfoString(appPath, "FileVersion")
	if stringLength(Skype__sutil___vcLastVer) > 5
		Skype__sutil___vcLastVer = "v" +Skype__sutil___vcLastVer
		return Skype__sutil___vcLastVer
	endIf
	; Get the actual version code from the fileSystemObject if it feels like cooperating.
	o = Skype__util__getFSObject()
	if o
		Skype__sutil___vcLastVer = o.getFileVersion(appPath)
		if stringLength(Skype__sutil___vcLastVer)
			Skype__sutil___vcLastVer = "v" +Skype__sutil___vcLastVer
			return Skype__sutil___vcLastVer
		endIf
	endIf
endIf

; Favor the newest Skype available when we can't tell what it really is.
; Also force an update until we get a better answer though.
Skype__sutil___vcLastVer = "v999-"
Skype__sutil___vcLastTicks = 0
return Skype__sutil___vcLastVer
endFunction

string function Skype__sutil__getSkypeUserName()
; Returns the Skype ID of the current Skype user.
var
	handle hwnd,
	int pos,
	string buf
hwnd = Skype__appwin__getSkypeWindow(SWT_Main)
if !hwnd
	return ""
endIf
buf = getWindowName(hwnd)
pos = stringContainsFromRight(buf, " - ")
if !buf
	return ""
endIf
buf = stringChopLeft(buf, pos+2)
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(buf))
endFunction

string function Skype__sutil__gotoSkypeScriptPage(string page, int deferred)
; "Runs" or returns the run command for the given page.
; Returns the null string if deferred is False and the command to run if deferred is True.
; page: The page to launch.
; deferred: False to run, True to return the command to be run as a string.
var
	string strURLString,
	string strBrailleDriver, string strBrailleDisplay, string strBraillePort,
	string strSkypeProgramPath,
	string strSkypeProgramDate

if deferred
	var string cmd
	cmd = "Skype__sutil__gotoSkypeScriptPage(\"" +page +"\", 0)"
	return cmd
endIf

strSkypeProgramPath = getAppFilePath()
strSkypeProgramDate = IntToString(GetFileDate (strSkypeProgramPath))
BrailleGetDeviceInfo(strBrailleDriver, strBrailleDisplay, strBraillePort)

strURLString = CNSTSkypeScriptsURL +page
+"?JFWVersion="+IntToString(Skype__util__getJFWVersion())
+"&ScriptVersion="+Script_Version
+"&ScriptRevision="+Script_Revision
+"&BrailleDisplay="+strBrailleDisplay
+"&SkypeProgramDate="+strSkypeProgramDate
+"&SkypeProgramVersion=" +stringChopLeft(Skype__sutil__getSkypeVersionCode(True), 1)
+"&WindowsOSVersion="+IntToString(GetWindowsOS())
+"&SkypeWatch="+intToString(Skype__sutil__isSkypeWatchLoaded())
+"&SkypeUserName="+Skype__sutil__getSkypeUserName()
strURLString = stringReplaceSubstrings(strURLString, " ", "%20")

if UserBufferIsActive () UserBufferDeactivate() EndIf
SayFormattedMessage (OT_Message, msgSkypeScriptsHomePage_L, msgSkypeScriptsHomePage_S)
Run("\"" +strURLString +"\"")
endFunction

int function Skype__sutil___splitContactInfo(string buffer, string byRef name, string byRef id, string byRef stat)
; Returns by reference the contact name, Skype ID, and status from the given buffer.
; The buffer should have come directly from MSAA's AccHelp property.
; Buffer format: <contactName>[ (<SkypeID>)] - <statusInfo>
; Returns True on success and False on failure.
; Author:  Doug Lee
var
	int pos

if stringIsBlank(buffer)
	; Happens on contact list entries for multichats.
	beep()
	return False
endIf

; Start by removing and collecting the status text (Online, Offline and last-seen time, etc.).
pos = stringContainsFromRight(buffer, "/>")  ; should always be found
if !pos
	return False
endIf
stat = stringChopLeft(buffer, pos+3)
buffer = stringLeft(buffer, stringContainsFromRight(buffer, "<")-2)
; Current buffer format: <contactName>[ (<SkypeID>)].
buffer = stringSegment(buffer, "\13", 1)

; Find out if there's a second name, and set name and id accordingly.
; If there's not, all we have is the Skype ID, which can't contain parentheses.
; If the display name differs from the Skype ID, we'll have both here.
pos = stringContainsFromRight(buffer, "(")  ; only found if second name is present
if pos
	name = stringLeft(buffer, pos-2)
	id = substring(buffer, pos+1, stringLength(buffer)-pos-1)
else
	name = buffer
	id = buffer
endIf
return True
endFunction

int function Skype__sutil___isPhoneNumber(string s)
; Returns True if the given string appears to be a phone number as might be found in Skype.
var
	int i, int n, string ch
s = stringTrimLeadingBlanks(stringTrimTrailingBlanks(s))
n = stringLength(s)
if n > 50
	; 50 is a pretty liberal limit for the length of a phone number...
	return False
endIf
i = 1
if stringLeft(s, 1) == "+"
	i = 2
endIf
while i <= n
	ch = substring(s, i, 1)
	if !stringIsBlank(ch) && !stringContains(",-.0123456789", ch)
		return False
	endIf
	i = i +1
endWhile
return True
endFunction

string function Skype__sutil__removeHTML(string s)
; Remove HTML tags and (some) character names from s.
var
	string lhs, string rhs,
	int pos,
	string result
pos = stringContains(s, "<")
while pos
	result = result +" " +stringLeft(s, pos-1)
	s = stringChopLeft(s, pos)
	pos = stringContains(s, ">")
	if pos
		s = stringChopLeft(s, pos)
	else
		s = ""
	endIf
	pos = stringContains(s, "<")
endWhile
result = stringChopLeft(result +" " +s, 1)
Skype__string__replace(result, "&nbsp;", " ", "e")
Skype__string__replace(result, "&lt;", "<", "e")
Skype__string__replace(result, "&gt;", ">", "e")
Skype__string__replace(result, "&lquot;", "\"", "e")
Skype__string__replace(result, "&rquot;", "\"", "e")
return result
endFunction

collection function Skype__sutil__getSkypeContactInfoFromMSAA(object o, int childID)
; Get as much information from MSAA about the indicated contact as possible.
; Currently this includes name, ID, status, ExtStatus, SpeedDial, and MoodText.
var
	string buf, string name, string ID, string stat, string extstat, string mood,
	int pos,
	collection cContact
name = ""
id = ""
stat = ""
extstat = ""

buf = Skype__sutil__removeHTML(o.accHelp(childID))
if stringLength(buf) && Skype__sutil___splitContactInfo(buf, name, id, stat)
	; Separate "Last seen: ..." from "Offline" when found.
	pos = stringContainsFromRight(stat, "(")
	if pos
		extstat = stringChopLeft(stat, pos)
		extstat = stringChopRight(extstat, 1)  ; chops off ')'
		stat = stringTrimTrailingBlanks(stringLeft(stat, pos-1))
	endIf
else
	; Use accValue when there's no accHelp.
	; AccHelp for contacts vanished around Skype 4.0 beta 3.
	buf = o.accValue(childID)
	if stringLength(buf) && stringRight(buf, 1) == ")"
		pos = stringContainsFromRight(buf, "(")
		if pos
			name = stringTrimTrailingBlanks(stringLeft(buf, pos-1))
			id = stringChopRight(stringChopLeft(buf, pos), 1)
		endIf
	endIf
endIf

cContact = new collection
cContact.Name = name
cContact.ID = id
cContact.Status = stat
if extstat
	cContact.ExtStat = extstat
endIf
; Include extra stuff from accDescription if found (Speed dial for example).
; Known format as of Skype 2.0.0.78: <status>[, <moodText>][ Speed dial <speedDialNumber>]
; TODO:  See if "Speed dial" is localized in accDescription.
buf = stringTrimTrailingBlanks(o.accDescription(childID))
if buf
	if buf == stat || !stringLength(stat)   ; match up to the end of status text
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(stat)))
	else
		; accDescription should always begin with status text.
		beep()
	endIf
	mood = ""
	if buf == ", "
		; There is mood text.
		buf = stringChopLeft(buf, 2)
		pos = stringContainsFromRight(buf, SU_MSAASpeedDial)
		if pos
			mood = stringLeft(buf, pos-1)
			buf = stringChopLeft(buf, pos)
		else
			mood = buf +""
			buf = ""
		endIf
	endIf
	if mood
		cContact.MoodText = mood
	endIf
	if buf == SU_MSAASpeedDial
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(SU_MSAASpeedDial)))
		cContact.SpeedDial = intToString(stringToInt(buf))
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(intToString(stringToInt(buf)))))
	endIf
	if !stringIsBlank(buf)
		cContact.Other = buf
	endIf
endIf  ; non-empty buf from accDescription
return cContact
endFunction

int function Skype__sutil___refreshSkypeWindow(handle hwnd)
if False && getWindowClass(getFirstChild(hwnd)) == "TScrollBar"
	;speechOff()
	setFocus(getFirstChild(hwnd))
	pause()
	setFocus(hwnd)
	pause()
	speechOn()
	return True
endIf
;sayInteger(sendMessage(hwnd, 14, 0, 0))
return False
endFunction

collection function Skype__sutil__getCurrentContactInfo()
; Returns a collection consisting of all retrievable fields about the current contact, with field names.  Both MSAA and screen text are used.
; When a field appears on screen but its name can't be determined, it is named "Field" plus its ordinal number in the field list (e.g., "Field5").
; Currently recognized fields are Name, ID, Status, ExtStatus, SpeedDial, MoodText, Location, Time, TimeZone, and Phone1-3.
var
	handle hContactList,
	object o, int childID,
	int tries,
	int left, int right, int top, int bottom,
	int oldRestriction,
	int i, int ibase, int phoneCount, int done,
	string buf, int pos,
	string mood, string stat, string id,
	collection cContact

; First get the MSAA information.
o = Skype__sutil__getFocusObject(childID)
if !o
	o = Skype__appwin__getMSAAObject(SWT_ContactList)
endIf
if !o
	return ""
endIf
childID = o.accFocus
if !childID
	return ""
endIf
cContact = Skype__sutil__getSkypeContactInfoFromMSAA(o, childID)

; Now restrict the invisible cursor to the contact's rectangle as indicated by its MSAA coordinates.
; Obscured windows are skipped because that means the information from the screen might not be valid anyway.
; This whole step is also skipped if focus is not in the contact list, because even other contact lists with the same MSAA structure do not share the same on-screen format.
; Example:  the participant list in a chat window (Skype 3-).
hContactList = Skype__appwin__getSkypeWindow(SWT_ContactList)
if !hContactList
	; TODO: Exits here for Skype 5 because there is no contact list window.
	return cContact
endIf
saveCursor() invisibleCursor() saveCursor()
oldRestriction = getRestriction()
tries = 2
done = False
while !done
	if !tries || getFocus() != hContactList || isWindowObscured(hContactList) || !Skype__MSAA__rect(o, childID, left, right, top, bottom)
		;beep()
		;sayString("Rect " +getTextInRect(left, top, right, bottom, 0, IgnoreColor, IgnoreColor, True))
		setRestriction(oldRestriction)
		return cContact
	endIf
	moveTo((left+right)/2, (top+bottom)/2)
	setRestrictionToRect(left, top, right, bottom)
	; Make sure we're on the first text chunk in the rectangle.
	JAWSPageUp() JAWSHome() nextChunk() priorChunk() priorChunk()
	if stringContains(getLine(), cContact.Name)
		done = True  ; causes the loop to exit
	else
		if tries == 2
			Skype__sutil___refreshSkypeWindow(hContactList)
		endIf
		pause()
		;sayString("retry")
	endIf
	tries = tries -1
endWhile

; Collect the data one text chunk at a time from the screen.
; The first line is skipped in all cases because it is just the contact name, which we already got from MSAA.
; For SkypeOut contacts, the phone number appears on the second line, but we already have that from MSAA also.
; Field names can be filled in retroactively by this code as context is discovered.
if nextLine()
	JAWSHome()
	stat = cContact.Status
	id = cContact.ID
	if stringContains(stringLower(stat), SU_MSAASkypeOut) && getChunk() == ID
		if !nextChunk()
			setRestriction(oldRestriction)
			return cContact
		endIf
	endIf
	done = False
	ibase = collectionItemCount(cContact)
	i = ibase
	mood = cContact.MoodText
	while !done && i < 50
		buf = getChunk()
		if stringCompare(buf, SU_scMenu, True) && stringCompare(buf, mood, True)
		pos = stringContainsFromRight(buf, "("+SU_scGMT)
		if pos
			; Skype 3.x structure: <time> (<tz>)[, <location>]
			ccontact.Time = stringTrimTrailingBlanks(stringLeft(buf, pos-1))
			buf = stringChopLeft(buf, pos)  ; also chops off '('
			pos = stringContains(buf, ")")
			cContact.TimeZone = stringLeft(buf, pos-1)
			buf = stringChopLeft(buf, pos)
			if buf == ","
				cContact.Location = stringTrimLeadingBlanks(stringChopLeft(buf, 1))
			endIf
		else  ; not a time/location line
			if Skype__sutil___isPhoneNumber(buf)
				phoneCount = phoneCount +1
				; This arranges for a single phone number to be called "Phone" but multiple phone numbers to be called "Phone1," "Phone2," etc.
				if phoneCount == 1
					cContact.Phone = buf
				elif phoneCount == 2
					cContact["Phone"+intToString(phoneCount)] = buf
					cContact.Phone1 = cContact.Phone
					collectionItemRemove(cContact, "Phone")
				else
					cContact["Phone"+intToString(phoneCount)] = buf
				endIf
			else
				; We don't know the name of this field, so we'll assign a generic one (and hope a better one is found above soon).
				cContact["Field"+intToString(i/2+1)] = buf
			endIf
			i = i +2
		endIf
		endIf  ; chunk is "Menu"
		done = !nextChunk()
	endWhile
endIf
setRestriction(oldRestriction)
return cContact
endFunction

globals string Skype__sutil___lastStat
string function Skype__sutil__getMyOnlineStatus()
; Returns this user's online/offline/etc. status as a string.
; In Skype 3.2+, the return string is already localized, as it comes directly from MSAA.
; Author:  Doug Lee
var
	string sVer,
	object o, int childID, int n,
	int found,
	string stat
o = Skype__appwin__getMSAAObject(SWT_Myself)
if o
	n = o.accChildCount
	childID = 1
	found = False
	while childID <= n
		if o.accRole(childID) == Role_System_ButtonDropDown
			stat = o.accName(childID)
			if !stat
				stat = SU_msgStatUnknown
			endIf
			Skype__sutil___lastStat = stringLower(stat)
			found = True
		endIf
		childID = childID +1
	endWhile
endIf
return Skype__sutil___lastStat
endFunction

string function Skype__sutil__speakableSkypeID(string id)
; Speakable version of the given Skype ID:
;	- Conference IDs say "conference."
;	- US phone numbers are formatted as +1 nnn-nnn-nnnn.
if id == "#"
	id = msgConference
elif id == "+" && stringLength(id) == 12
	id = formatString("%1 %2-%3-%4",
		stringLeft(id, 2),
		substring(id, 3, 3),
		substring(id, 6, 3),
		stringRight(id, 4)
	)
endIf
return id
endFunction

int function Skype__sutil__play(string sound)
; Play the indicated sound file.
var
	string spath
;spath = findJAWSSoundFile(sound)
spath = getJAWSSettingsDirectory() +"\\sounds\\" +sound
if !fileExists(spath)
	spath = getJAWSSettingsDirectory() +"\\" +sound
endIf
if !fileExists(spath)
	spath = findJAWSSoundFile(sound)
endIf
if spath
	playSound(spath)
endIf
endFunction

string function Skype__sutil__getWindowClassPath(handle hwnd)
var
	string path,
	string sep
sep = "/"
while hwnd
	path = getWindowClass(hwnd) +"/" +path
	hwnd = getParent(hwnd)
endWhile
return stringChopRight(path, 1)
endFunction

int function Skype__sutil__isUnobscured(object o, int childID)
; Returns True if the indicated MSAA element appears to be unobscured by something else.
; Uses GetObjectAtPoint, so requires .AccHitTest to work.
var
	int left, int right, int top, int bottom,
	object o1, int childID1
if !Skype__MSAA__rect(o, childID, left, right, top, bottom)
	; If it doesn't have screen space, it's sort of obscured by default.
	return False
endIf
o1 = getObjectAtPoint(childID1, (left+right)/2, (top+bottom)/2)
if childID1 == childID && o.accName(0) == o1.accName(0)
	return True
endIf
return False
endFunction

int function Skype__sutil__isAncestorOf(handle hwndChild, handle hwndAncestor)
; Returns True if and only if hwndChild descends at some level from hwndAncestor.
; True is also returned if the two are equal.
while hwndChild
	if hwndChild == hwndAncestor
		return True
	endIf
	hwndChild = getParent(hwndChild)
endWhile
return False
endFunction

int function Skype__sutil__getMSAAItemDim(object o, int childID, string which)
; Return left, right, top, bottom, width, or height for the given MSAA item.
var
	int left, int top, int width, int height
o.accLocation(intRef(left), intRef(top), intRef(width), intRef(height), childId)
if !left && !top && !width && !height
	return 0
endIf
left = left +1
top = top +1
if which == "l" return left
elif which == "t" return top
elif which == "w" return width
elif which == "h" return height
elif which == "r" return left+width
elif which == "b" return top+height
else return 0
endIf
endFunction

int function Skype__sutil__findNthColorTransition(handle hwnd, int delta, int n)
; Finds the nth color transition in delta's direction from the current active cursor's location.
; Returns True on success and False on failure.
; The active cursor is moved to the found spot only if a spot is found.
; The search is limited by the edge of hwnd in delta's direction if hwnd is not null.
; Author:  Doug Lee
var
	int x, int y,
	int limit, int tooFar,
	int clr, int clr0
if hwnd
	if delta > 0
		limit = getWindowRight(hwnd)
	else
		limit = getWindowLeft(hwnd)
	endIf
else
	if delta > 0
		limit = screenGetWidth() +1
	else
		limit = 0
	endIf
endIf
x = getCursorCol() +(delta *3)  ; arbitrary safety margin to avoid edge of text characters
y = getCursorRow()
clr0 = getColorAtPoint(x, y)
tooFar = False
while n && !tooFar
	clr = getColorAtPoint(x, y)
	if clr != clr0
		n = n -1
		clr0 = clr
	endIf
	x = x +delta
	if delta > 1
		tooFar = (x > limit)
	else
		tooFar = (x < limit)
	endIf
endWhile
if tooFar
	return False
endIf
moveTo(x, y)
return True
endFunction


;--------------------------------- Module xlist --------------------------------
; XList, the arbitrary list (array and hash emulation) module.
;
; Terminology:
;	XList - An arbitrary list of items (an array, basically).
;	XLHash - An XList in which odd-numbered items are keys and even-numbered items are their values.
;
; Implementation Notes:
; XLists always begin and end with a delimiter; a null list is just a single delimiter character.
; This means it is NOT possible to use boolean (if sl) or length (if stringLength(sl)) to check for emptiness; use isEmpty or itemCount.
; Objects can not be stored in an XList directly, and attempting this will cause an error message to speak.
; Translation Notes:  The only translatable strings appear in SayMessage calls.
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

const
; Indexes used to request the character tables supported by this module.
	Skype__sutil__xlist___CharANSI = 1,
; The corresponding character tables.
; The double quote (") and backslash characters must be escaped.
	Skype__sutil__xlist___CharANSICharList = "\1\2\3\4\5\6\7\8\9\10\11\12\13\14\15\16\17\18\19\20\21\22\23\24\25\26\27\28\29\30\31 !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\127\128\129\130\131\132\133\134\135\136\137\138\139\140\141\142\143\144\145\146\147\148\149\150\151\152\153\154\155\156\157\158\159\160\161\162\163\164\165\166\167\168\169\170\171\172\173\174\175\176\177\178\179\180\181\182\183\184\185\186\187\188\189\190\191\192\193\194\195\196\197\198\199\200\201\202\203\204\205\206\207\208\209\210\211\212\213\214\215\216\217\218\219\220\221\222\223\224\225\226\227\228\229\230\231\232\233\234\235\236\237\238\239\240\241\242\243\244\245\246\247\248\249\250\251\252\253\254\255"

;------------------------------- End module xlist ------------------------------

;====================== End module sutil (from skypeUtil) ======================


;================================== Module log =================================
; Allow logging to a text file.
;
; Interface:  Skype__log__open, Skype__log__isOpen, Skype__log__close, Skype__log__write, Skype__log__setUserID.
; Basic usage: Skype__log__open() from AutoStartEvent, Skype__log__write("...") to log something.
; Usage:
;	- Call Skype__log__open() from AutoStartEvent.
;	- Logging will occur by default only if the log file already exists (good for debugging during development).
;	- Call as Skype__log__open("", True) if you want logging to occur even if the log file doesn't already exist.
;	- Pass a file name instead of "" for a custom file name (rarely needed).
;	- Call Skype__log__write("Oops I saw an error") to log something.
;	- Use Skype__log__write("Oops this error needs to speak too", True) to speak what is being logged.
;	- Call Skype__log__close() to disable logging (rarely needed).
;
; Author:  Doug Lee of SSB BART Group

; Constants for Skype__log__write and related functions.
; Not localized because log text is not localized.

; Used as a user ID when none is found.
const Skype__log___msgNoName = "<no name>"
; What "name" is used for the caller string when JAWS is too old for getScriptCallStack().
; %1 is the JAWS version.
const Skype__log___msgJAWSTooOld = "(caller unknown; JAWS version (%1) too old)"
; Name used for the caller when it can't be determined for an unknown reason,
; probably a getScriptCallStack return value format surprise.
const Skype__log___msgNoCaller = "(caller unknown)"

globals
; FileSystemObject and log stream object obtained from it.
	object Skype__log___fso,
	object Skype__log___logstream,
; Name of file.
	string Skype__log___fileName,
	int Skype__log___canCreate,
	string Skype__log___userID,
	string Skype__log___machineID

int function Skype__log___setup()
; Internal function called to set things up when necessary before any log text is generated.
if !stringIsBlank(Skype__log___userID) then
	return True
endIf
let Skype__log___fso = createObjectEx("Scripting.FileSystemObject", False)
if !Skype__log___fso then
	return False
endIf
let Skype__log___machineID = ""
var object oNet let oNet = createObjectEx("WScript.Network", False)
if oNet then
	let Skype__log___machineID = oNet.ComputerName
endIf
if stringIsBlank(Skype__log___machineID) then
	let Skype__log___machineID = Skype__log___msgNoName
endIf
let Skype__log___userID = getJAWSUserName()
if stringIsBlank(Skype__log___userID) && oNet then
	let Skype__log___userID = oNet.UserName
endIf
if stringIsBlank(Skype__log___userID) then
	let Skype__log___userID = Skype__log___msgNoName
endIf
return True
endFunction

int function Skype__log__setUserID(string id)
; Optional means of setting a custom user ID.
; If not called, the value of getJAWSUserName() is used.
let Skype__log___userID = id
return True
endFunction

int function Skype__log__close()
; Close the log file.
Skype__log___close()
let Skype__log___fileName = ""
return True
endFunction

int function Skype__log___setFileName(string name)
; Set the log file name.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Helper for Skype__log__open().
var string basename
let basename = name
if stringRight(basename,4) == ".log" then
	let basename = stringChopRight(basename, 4)
endIf
if !basename then
	let basename = getActiveConfiguration()
endIf
if !basename then
	let basename = getAppFileName()
	var int pos let pos = stringContainsFromRight(basename, ".")
	if pos > 0 then
		let basename = stringLeft(basename, pos-1)
	endIf
endIf
let Skype__log___fileName = getJAWSSettingsDirectory() +"\\" +basename
if stringRight(Skype__log___fileName, 4) != ".log" then
	let Skype__log___fileName = Skype__log___fileName +".log"
endIf
return True
endFunction

int function Skype__log__open(/*?*/ string name, int canCreate)
; Open and optionally create the log file.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Call from autoStartEvent, or logging will not occur.
; Returns True if the file is opened and False if not.
if !Skype__log___setup() then
	return False
endIf
let Skype__log___canCreate = canCreate
Skype__log__close()
if !Skype__log___setFileName(name) || !Skype__log___open() then
	let Skype__log___fileName = ""
	return False
endIf
return True
endFunction

int function Skype__log___open()
; Raw open/create.  Sets Skype__log___logstream.
; Returns True on success and False on failure.
if Skype__log___logstream then
	return True
endIf
if fileExists(Skype__log___fileName) then
	let Skype__log___logstream = Skype__log___fso.openTextFile(Skype__log___fileName, 8)  ; Append
elif Skype__log___canCreate then
	let Skype__log___logstream = Skype__log___fso.CreateTextFile(Skype__log___fileName)
endIf
if !Skype__log___logstream then
	return False
endIf
scheduleFunction("Skype__log___close", 30)
return True
endFunction

int function Skype__log___close()
; Raw cloase; unsets Skype__log___logstream.
var variant null
if !Skype__log___logstream then
	return True
endIf
Skype__log___logstream.Close()
let Skype__log___logstream = null
return True
endFunction

int function Skype__log___isOpen()
; Returns True if the log file is actually open and False if not.
if Skype__log___logstream then
	return True
endIf
return False
endFunction

int function Skype__log__isOpen()
; Returns True if the log file is open and False if not.
; "Open" here means can be written to but may not necessarily mean the file is physically opened at the moment.
if Skype__log___fileName then
	return True
endIf
return False
endFunction

string function Skype__log___getStamp()
; Returns a formatted date/time stamp for the log file.
var string tm let tm = sysGetTime("HH:mm:ss")
var string dt let dt = sysGetDate("ddd MMM dd yyyy")
return formatString("%1 %2", tm, dt)
endFunction

int function Skype__log__write(string s, /*?*/ int alsoSpeak)
; Write a string to the log file, and also optionally speak it.
if alsoSpeak then
	sayMessage(OT_Error, s)
endIf
if !Skype__log__isOpen() then
	return False
endIf
if !Skype__log___open() then
	return False
endIf
var string buf let buf = formatString(
	"*** %1 %2@%3 (%4)\n%5:\n",
	Skype__log___getStamp(), Skype__log___userID, Skype__log___machineID, getAppFileName(),
	Skype__log___getCaller()
) +s +"\n\n"
; Standardize any CR/LF characters in buf before writing.
let buf = stringReplaceChars(buf, "\r\n", "\n")
let buf = stringReplaceSubstrings(buf, "\n", "\r\n")
Skype__log___logstream.write(buf)
return True
endFunction

string function Skype__log___getCaller()
; Get and return the caller of this logging system.
; Helper for Skype__log__write().
if getJFWVersion() < 1000000 then
	var string jver let jver = stringSegment(Skype__util__getJAWSInfo(), "|", 1)
	return formatString(Skype__log___msgJAWSTooOld, jver)
endIf
var string stack let stack = Skype__util__callStackInfo(2, "s", 2)
if !stringIsBlank(stack) then
	return stack
endIf
return formatString("%1\nStack begin:\n%2\nStack end.",
	Skype__log___msgNoCaller,
	Skype__util__vcast(getScriptCallStack())
)
endFunction

string function Skype__log___splitOff(string byRef s, string sep, /*?*/ int fromRight)
; Split what's before sep off and return it.
; If fromRight is True, split everything before the last sep off and return that.
; Returns by reference s with both that and sep removed.
var string s1
var int pos
if fromRight then
	let pos = stringContainsFromRight(s, sep)
else
	let pos = stringContains(s, sep)
endIf
if pos then
	let s1 = stringLeft(s, pos-1)
	let s = stringChopLeft(s, pos+stringLength(sep)-1)
else
	let s1 = s
	let s = ""
endIf
return s1
endFunction

;================================ End module log ===============================


;================================= Module mkcs =================================
; This module implements a KeymapChangedEvent-based multi-key command sequence system.
;
; Usage:
; Include this module in the main jss.
; Create a message constant for the MKCS help text (see the mkcsHelp script for format info).
; Load the help text from AutoStartEvent with a call to Skype__mkcs__setHelpText().
; Set the command key (and again if it changes) by calling Skype__mkcs__setCommandKey().
; Write key mappings in the jkm file using "&" between layers on each line.
; For each command level, make the following additional mappings (the examples are for the top level with [ as the command key):
;	[&Escape=mkcsExit
;	[&Tab*=mkcsHelp("next")
;	[&Shift+Tab*=mkcsHelp("prev")
;	[&JAWSKey+Tab*=mkcsHelp("cur")
;	[&Insert+Tab*=mkcsHelp("cur")
;	[&Enter=mkcsHelp("exec")
; NOTE: The first line above is mandatory, as it is also used by this code
; to find the original command key when changing it to a new one.
; See Skype__mkcs__setCommandKey().
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

include "hjglobal.jsh"

globals
; The command key that starts all sequences recognized here.
	string Skype__mkcs___commandKey

int function Skype__mkcs__setCommandKey(string sKey, /*?*/ string fname, string sect)
; Sets the key used as a prefix for all command sequences in this application.
; Call from AutoStartEvent or appropriate initialization code.
; Also call if the key is ever changed.
; Updates the jkm file if necessary.
; If fname is not passed, defaults to <activeConfiguration>.jkm.
; If sect is not passed, defaults to "Common Keys."
; Returns 0 if things are not left in sync, 1 if they are, and 2 if the jkm changed.
; Things should not be left out of sync unless there is an error.
if !sKey
	; Can't set to null key, doing nothing is inconsistent,
	; so we set to a string that is not a valid key.
	sKey = "<NoKey>"
endIf
Skype__mkcs___commandKey = sKey
if !fname
	fname = getActiveConfiguration() +".jkm"
endIf
if !stringContains(fname, "\\")
	fname = getJAWSSettingsDirectory() +"\\" +fname
endIf
if !sect
	sect = "Common Keys"
endIf

var string jkm = Skype__util__stringFromFile(fname)
if !jkm
	; Chances are this means we're on the secure desktop and can't read files.
	return 0
endIf
var string sep = "\n"
var int pos = stringContains(stringLower(jkm), "&escape=mkcsexit")
if !pos
	sayMessage(OT_Error, Skype__mkcs___msgNotFound1)
	return 0
endIf
var string sKey0 = stringLeft(jkm, pos-1)
pos = stringContainsFromRight(sKey0, sep)
if !pos
	sayMessage(OT_Error, Skype__mkcs___msgNotFound2)
	return 0
endIf
sKey0 = stringChopLeft(sKey0, pos)
if stringCompare(sKey0, sKey, False) == 0
	return 1
endIf

sayUsingVoice(VCTX_Message, Skype__mkcs___msgUpdating, OT_Message)
var int n = stringSegmentCount(jkm, sep)
var string jkm1 = ""
var int i = 1
var int changed = False
var int inSect = False
var string line
while i <= n
	line = stringSegment(jkm, sep, i)
	line = stringTrimLeadingBlanks(stringTrimTrailingBlanks(line))
	if line == "[" +sect +"]"
		inSect = True
	elif stringLeft(line, 1) == "[" && stringRight(line, 1) == "]"
		inSect = False
	elif inSect
		if line == sKey0+"&"
			line = sKey +stringChopLeft(line, stringLength(sKey0))
			changed = True
		endIf
	endIf
	if jkm1
		jkm1 = jkm1 +"\r" +sep
	endIf
	jkm1 = jkm1 +line
	i = i +1
endWhile
jkm1 = jkm1 +"\r" +sep
if changed
	Skype__util__stringToFile(fname, jkm1)
	sayUsingVoice(VCTX_Message, Skype__mkcs___msgDone, OT_Message)
	scheduleFunction("reloadAllConfigs", 2)
	return 2
else
	sayMessage(OT_Error, Skype__mkcs___msgUpdateError)
	return 0
endIf
endFunction

string function Skype__mkcs__getCommandKey()
; Returns the key used as a prefix for all command sequences in this application.
; This will be the value passed to the latest call to Skype__mkcs__setCommandKey().
return Skype__mkcs___commandKey
endFunction

string function Skype__mkcs___getActiveKeyboardLayout()
; Returns the name of the keyboard layout in effect or null if this can't be determined.
; Tested in JAWS 9.0.2169.
var string fname, string layout
fname = getJAWSSettingsDirectory() +"\\default.jcf"
layout = iniReadString("Options", "KeyboardType", "", fname)
if !layout
	fname = getSharedSettingsDirectory() +"\\default.jcf"
	layout = iniReadString("Options", "KeyboardType", "", fname)
endIf
return layout
endFunction

string function Skype__mkcs___getActiveKeymapFile()
; Returns the full path to the active .jkm file or nul if this can't be determined.
var
	string fname
fname = getActiveConfiguration()
if !fname
	return ""
endIf
fname = fname +".jkm"
fname = findJAWSSettingsFile(fname)
if !fname
	return ""
endIf
return fname
endFunction

int function Skype__mkcs___updateKeyList(string byRef list, string prefix, string jkm, string layout)
; Adds any key sequences to the given list that begin with the given prefix.
; Key sequences in the given keyboard layout section of the indicated jkm file are scanned for matches.
; Returns the number (0 or more) of entries added to the list.
var
	string keys,
	string sKey,
	int pos,
	int n, int i,
	int prefixLen,
	int additions
keys = iniReadSectionKeys(layout +" Keys", jkm)
n = stringSegmentCount(keys, "|")
if !keys || !n
	return 0
endIf
prefix = stringStripAllBlanks(prefix)
prefixLen = stringLength(prefix)
additions = 0
i = 1
while i < n
	sKey = stringStripAllBlanks(stringSegment(keys, "|", i))
	; Case-insensitive prefix match that doesn'tsKey be a substring of the prefix.
	if stringLeft(sKey, prefixLen) == prefix && stringLength(sKey) > prefixLen
		; +1 cuts off the "&" after the prefix.
		sKey = stringChopLeft(sKey, prefixLen+1)
		; Only consider this key level, not subsequent ones.
		; Also remove the trailing "*" that means "stay at this level" if it is there.
		pos = stringContains(sKey, "&")
		if pos
			sKey = stringLeft(sKey, pos-1)
		elif stringRight(sKey, 1) == "*"
			sKey = stringChopRight(sKey, 1)
		endIf
		; Avoid adding duplicates or keys that are part of this help system.
		if !stringContains(stringLower(sKey), "tab")
		&& !stringContains(stringLower(sKey), "enter")
		&& !stringContains(stringLower(sKey), "esc")
		&& !stringSegmentIndex(list, "|", sKey, True)
			list = list +"|" +sKey
			additions = additions +1
		endIf
	endIf
	i = i +1
endWhile
return additions
endFunction

string function Skype__mkcs___getKeyList(string prefix)
; Returns a list of the key sequences starting with the given prefix.
; Key sequences in Common Keys and the active keyboard layout section of the active jkm file are scanned for matches.
; Returned sequences omit the prefix.
var
	string layout,
	string jkm,
	string list
layout = Skype__mkcs___getActiveKeyboardLayout()
if !layout
	return ""
endIf
jkm = Skype__mkcs___getActiveKeymapFile()
if !jkm
	return ""
endIf
list = ""
Skype__mkcs___updateKeyList(list, prefix, jkm, "Common")
Skype__mkcs___updateKeyList(list, prefix, jkm, layout)
list = stringChopLeft(list, 1)  ; remove extra leading separator.
return list
endFunction

const
; Values for iKeyStatus in keymapChangedEvent
	Skype__mkcs___KeySequenceComplete = 0,
	Skype__mkcs___KeySequenceRepeating = 1,
	Skype__mkcs___KeySequencePending = 2,
	Skype__mkcs___KeySequenceCanceled = 3,
	Skype__mkcs___KeySequenceStart = 4,
	Skype__mkcs___KeySequenceRestarted = 5
globals
	int Skype__mkcs___active,
	string Skype__mkcs___helpText,
	int Skype__mkcs___lastKeyCode,
	string Skype__mkcs___sequence,
; The next three are used for the Enter/exec functionality.
	string Skype__mkcs___lastSeq,
	string Skype__mkcs___execScript,
	int Skype__mkcs___lastIndex,
	int Skype__mkcs___lastKeyStatus,
	int Skype__mkcs___helpIndex

int function Skype__mkcs___sayKeyPrompt(string seq)
; Speak the prompt for the given (or default) key sequence if there is a prompt for it.
; Returns True if the speaking of the prompt was handled and False if not.
var
	int idx,
	string prompt,
	int isScheduledCall
isScheduledCall = False
if !seq
	seq = Skype__mkcs___sequence
	isScheduledCall = True
endIf
; Cut off the command key from the sequence before looking up prompts.
idx = stringContains(seq, "&")
if idx
	seq = stringChopLeft(seq, idx)
else
	seq = ""
endIf
; Find the key's entry in the help text.
idx = stringContains(stringLower(Skype__mkcs___helpText), "\10"+stringLower(seq)+"=")
if idx
	; Chop off everything through seq=
	prompt = stringChopLeft(Skype__mkcs___helpText, idx+stringLength(seq)+1)
	; Chop off everything past the key's entry.
	prompt = stringSegment(prompt, "\10", 1)
	; Chop off the actual help text (left of the prompt).
	prompt = stringSegment(prompt, "|", 2)
	if prompt
		if !isScheduledCall
			stopSpeech()
		endIf
		sayMessage(OT_Help, prompt)
	endIf
	return True
endIf
return False
endFunction

void function Skype__mkcs__setHelpText(string helpText)
; Set the help text to be used by this system.
; Call from AutoStartEvent in the application script file.
Skype__mkcs___helpText = "\10" +helpText +"\10"
endFunction

script mkcsHelp(string which)
; Implements round-robin keyboard help for the current level of a multikey sequence.
; Which: next to move forward, prev to move backward, cur to stay put.
; These should correspond to Tab, Shift+Tab, and JAWSKey+Tab/Insert+Tab, respectively.
; Exec, which corresponds to Enter, executes the last entry reported.
; For all this to work, appropriate help must have been registered with Skype__mkcs__setHelpText().
; Required help format:
;	- One line per help entry.
;	- Line format: <sequence>=<help>[|<prompt>]
;	- <sequence> is the sequence to which the help applies, minus the command prefix and its "&".
;	- <help> is the help text to speak for this sequence.
;	- <prompt> (optional) is the prompt to speak when the user actually enters this sequence (not when getting help).
;	- The first line should provide the prompt for the actual command key (e.g., "=|Command").
var
	string prefix,
	int len,
	int pos,
	int i, int n,
	string seg,
	string helps,
	string seq

; Get the prefix, which is the current position in key sequences without the leading command key or its "&".
; Also get len, its length.
; Prefix will end with an "&" if it is not empty.
; It's empty for top-level keys though (keys that immediately follow the command key).
if which == "exec"
	; The sequence should have just now completed,
	; so the sequence is in Skype__mkcs___lastSeq, not Skype__mkcs___sequence.
	seq = Skype__mkcs___lastSeq
else
	seq = Skype__mkcs___sequence
endIf
prefix = seq
pos = stringContains(prefix, "&")
if pos
	; Get rid of the leading command character and its "&"
	; Then put an "&" at the end.
	prefix = stringChopLeft(prefix, pos) +"&"
else
	; We're at the top level (right after entry of the command key).
	prefix = ""
endIf
len = stringLength(prefix)

; Loop through help text lines, collecting those that apply to the current level.
; JAWS seems to use \10, not \13, to separate lines in a Messages text string.
n = stringSegmentCount(Skype__mkcs___helpText, "\10")
helps = ""
i = 2  ; Skype__mkcs__setHelpText prepends a delimiter, so the first segment is null.
while i <= n
	seg = stringSegment(Skype__mkcs___helpText, "\10", i)
	seg = stringTrimLeadingBlanks(stringTrimTrailingBlanks(seg))
	; The first check here keeps out the initial Command prompt line.
	; The second part lets through current entries if we're not at the top level.
	; The third lets through top-level entries if that's where we are.
	if stringLeft(seg, 1) != "=" && ((len && stringLeft(seg, stringLength(prefix)) == prefix) || !len)
		; Chop all but the current level's keystroke off this entry if we're not at the top level.
		if len
			seg = stringChopLeft(seg, len)
		endIf
		; Look for entries that are just this level, not deeper ones.
		; This logic allows "&" to occur in help or prompt text.
		if !stringContains(seg, "&") || stringContains(seg, "&") > stringContains(seg, "=")
			; Remove the prompt part if it's there.
			pos = stringContainsFromRight(seg, "|")
			if pos > stringContains(seg, "=")
				seg = stringLeft(seg, pos-1)
			endIf
			helps = helps +"|" +seg
		endIf
	endIf
	i = i +1
endWhile
helps = stringChopLeft(helps, 1)  ; extra leading delimiter

; Figure out which entry to speak, and do it.
if which == "next" || which == "prev"
	if which == "next"
		Skype__mkcs___helpIndex = Skype__mkcs___helpIndex +1
	elif which == "prev"
		Skype__mkcs___helpIndex = Skype__mkcs___helpIndex -1
	endIf
	if Skype__mkcs___helpIndex <= 0
		Skype__mkcs___helpIndex = stringSegmentCount(helps, "|")
	elif Skype__mkcs___helpIndex > stringSegmentCount(helps, "|")
		Skype__mkcs___helpIndex = 1
	endIf
endIf  ; next or prev
var int idx = Skype__mkcs___helpIndex
if which == "exec"
	idx = Skype__mkcs___lastIndex
endIf
seg = stringSegment(helps, "|", idx)
if which == "exec"
	var string sKey = stringSegment(seg, "=", 1)
	;sayMessage(OT_User_Requested_Information, sKey)
	; TODO: Disabled as this does not work in a layer: The key is typed
	; but the script is not run. Simulating seq+"&"+sKey actually fails.
	; [DGL, 2014-08-24, JAWS 15.0]
	;simulateKey(sKey)
	if !Skype__mkcs___simulateKey(seq, sKey)
		beep()
	endIf
else
	; This says the help text, pauses, then says the keystroke.
	sayMessage(OT_User_Requested_Information, formatString("%2, %1",
		stringSegment(seg, "=", 1),
		stringSegment(seg, "=", 2)
	))
endIf
endScript

int function Skype__mkcs___simulateKey(string seq, string sKey)
; Simulate the final key in a multikey sequence by calling its script.
sKey = formatString("%1&%2", seq, sKey)
var string mapping = getScriptAssignedTo(sKey)
if !mapping
	mapping = Skype__mkcs___getScriptName(sKey)
endIf
if !mapping
	return False
endIf
; Don't call directly because this may cause recursion problems.
; Don't use scheduleFunction directly in case "()" appears in the jkm.
; Schedule a formatStringWithEmbeddedFunctions to handle all this.
Skype__mkcs___execScript = mapping
scheduleFunction("Skype__mkcs___exec", 0)
return True
endFunction

string function Skype__mkcs___getScriptName(string sKey)
; Get the script name attached to the given key or sequence.
; Works on sequences where getScriptAssignedTo() fails.
var string layouts = Skype__mkcs___getActiveKeyboardLayout()
if layouts
	layouts = layouts +" Keys|"
endIf
layouts = layouts +"Common Keys"
var string files = ""
var string cfg = getActiveConfiguration()
if cfg
	files = files +formatString("|%1|%2",
		getJAWSSettingsDirectory() +"\\" +cfg +".jkm",
		getSharedSettingsDirectory() +"\\" +cfg +".jkm"
	)
endIf
files = files +formatString("|%1|%2",
	getJAWSSettingsDirectory() +"\\default.jkm",
	getSharedSettingsDirectory() +"\\default.jkm"
)
files = stringChopLeft(files, 1)
var int nfiles = stringSegmentCount(files, "|")
var int nlayouts = stringSegmentCount(layouts, "|")
var int i, int j, string fpath, string layout
var string mapping = ""
for i = 1 to nfiles
	fpath = stringSegment(files, "|", i)
	for j = 1 to nlayouts
		layout = stringSegment(layouts, "|", j)
		mapping = iniReadString(layout, sKey, "", fpath)
		if mapping
			return mapping
		endIf
		mapping = iniReadString(layout, sKey+"*", "", fpath)
		if mapping
			return mapping
		endIf
	endFor
endFor
return ""
endFunction

void function Skype__mkcs___exec()
; Execute a script called via Enter.
formatStringWithEmbeddedFunctions("<$"+Skype__mkcs___execScript+">")
Skype__mkcs___execScript = ""
endFunction

globals string Skype__mkcs___soundPath
int function Skype__mkcs___playExitSound()
; Play the "exit sequence" sound file.
if !Skype__mkcs___soundPath
	Skype__mkcs___soundPath = Skype__mkcs___findSound("mkcsExit.wav")
	if !Skype__mkcs___soundPath
		Skype__mkcs___soundPath = Skype__mkcs___findSound("CarLockClose4.wav")
	endIf
endIf
if Skype__mkcs___soundPath
	return playSound(Skype__mkcs___soundPath)
endIf
return False
endFunction

string function Skype__mkcs___findSound(string sound)
; Find and return the full path to the given sound file,
; or return null if it can't be found.
var string spath, string base
;spath = findJAWSSoundFile(sound)
base = getJAWSSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
base = getSharedSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
return ""
endFunction

void function keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
if Skype__mkcs__keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
	return
endIf
if getJFWVersion() >= 1100000
	; Chaining in older JAWS versions causes "unknown function call to KeyMapChangedEvent" messages for some reason.
	; JAWS 11.0 starts including a KeymapChangedEvent in default.jss.
	keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
endIf
endFunction

void function Skype__mkcs__keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
; Helper for KeyMapChangedEvent.
; Returns True if the default KeyMapChangedEvent should be blocked and False if not.

;sayString(formatString("Last %1 %2, current %4 %5",
;	Skype__mkcs___sequence, intToString(Skype__mkcs___lastKeyStatus), intToString(Skype__mkcs___lastKeyCode),
;sKeyName, intToString(iKeyStatus), intToString(iKeyCode)
;))
var
	string seq
; Filter out unrelated sequences and make any wanted sounds indicating changes into andout of our sequence.
if iKeyStatus == Skype__mkcs___KeySequenceStart || iKeyStatus == Skype__mkcs___KeySequenceRestarted
	if !isSameKey(sKeyName, Skype__mkcs___commandKey)
		; Not our sequence.
		Skype__mkcs___active = False
		return False
	endIf
	; Start or restart of our sequence.
	Skype__mkcs___active = True
elif !Skype__mkcs___active
	; Something other than activation going on, but it's not our sequence.
	return False
elif iKeyStatus == Skype__mkcs___KeySequenceCanceled
|| (iKeyStatus == Skype__mkcs___KeySequenceComplete && (Skype__mkcs___lastKeyStatus != Skype__mkcs___KeySequenceComplete && Skype__mkcs___lastKeyStatus != Skype__mkcs___KeySequenceCanceled))
	; End of our sequence.
	Skype__mkcs___playExitSound()
	Skype__mkcs___active = False
endIf

; Establish the sequence of keys that got us here.
if iKeyStatus == Skype__mkcs___KeySequenceStart || iKeyStatus == Skype__mkcs___KeySequenceRestarted
	Skype__mkcs___sequence = ""
	Skype__mkcs___helpIndex = 0
endIf
seq = ""
if Skype__mkcs___sequence
	seq = Skype__mkcs___sequence +"&"
endIf
seq = seq +sKeyName

; Announce the new key if there's an announcement to make for it.
if iKeyStatus == Skype__mkcs___KeySequenceCanceled
	if iKeyCode == Skype__mkcs___lastKeyCode
	&& (Skype__mkcs___lastKeyStatus == Skype__mkcs___KeySequenceStart || Skype__mkcs___lastKeyStatus == Skype__mkcs___KeySequenceRestarted)
		; Pass a repeat of the root key through and speak its name.
		sayMessage(OT_Keyboard, sKeyName)
		if !gbKeyboardHelp
			typeKey(sKeyName)
		endIf
	else
		; Avoid saying the invalid key's name and stop any previous speech.
		; Then announce the sequence cancellation.
		stopSpeech()
		;sayMessage(OT_Help, "Canceled", "")
	endIf
elif iKeyStatus != Skype__mkcs___KeySequenceComplete || Skype__mkcs___lastKeyStatus != Skype__mkcs___KeySequenceComplete
	; Say the key sequence's prompt if there is one, or the key name if not..
	; The above condition keeps this block from applying to non-layered script keys.
	if !Skype__mkcs___sayKeyPrompt(seq)
		stopSpeech()
		; TODO: Disabled because we don't always want default labels.
		;sayMessage(OT_Keyboard, sKeyName)
	endIf
endIf

; Set the sequence to remember for next time.
; Also set other globals to remember parameters to this call.
if iKeyStatus == Skype__mkcs___KeySequenceComplete || iKeyStatus == Skype__mkcs___KeySequenceCanceled
	if iKeyStatus == Skype__mkcs___KeySequenceComplete
		Skype__mkcs___lastSeq = Skype__mkcs___sequence
		Skype__mkcs___lastIndex = Skype__mkcs___helpIndex
	endIf
	Skype__mkcs___sequence = ""
	Skype__mkcs___helpIndex = 0
elif iKeyStatus == Skype__mkcs___KeySequenceRepeating
	; Do nothing; the new key stayed at this level.
	;scheduleFunction("Skype__mkcs___sayKeyPrompt", 0)
else
	; Pending:  Include the new key.
	; Start/Restarted:  Start with only the new key (already done above).
	Skype__mkcs___sequence = seq
	Skype__mkcs___helpIndex = 0
endIf
Skype__mkcs___lastKeyCode = iKeyCode
Skype__mkcs___lastKeyStatus = iKeyStatus
; This avoids an extra top-level prompt on the first non-Command scripted key after a sequence cancellation.
if iKeyStatus == Skype__mkcs___KeySequenceCanceled
	Skype__mkcs___lastKeyStatus = Skype__mkcs___KeySequenceComplete
endIf

; Chain this event.
;sayString("map " +sKeyName +" " +stringSegment("complete|repeating|pending|canceled|start|restarted", "|", iKeyStatus+1))
return True
endFunction

variant function unknown(string sScript, int isScript,
	variant byRef v1, variant byRef v2, variant byRef v3, variant byRef v4, variant byRef v5,
	variant byRef v6, variant byRef v7, variant byRef v8, variant byRef v9, variant byRef v10)
if sScript == "keymapChangedEvent"
	return
endIf
return unknown(sScript, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endFunction

script mkcsExit()
;sayMessage(OT_Help, "Exit", "", False)
endScript

;=============================== End module mkcs ===============================


;=============================== Module winclick ===============================
; Support for sending mouse clicks of various sorts directly to a window.
; This permits clicking on obscured windows
; or clicking without checking for what is in the foreground.
;
; Author:  Doug Lee, with credit to Jim Bauer for the idea and some initial research.
;
; TODO:
;	- No support for mouse wheel events.
;	- Alt+clicks not supported.
;	- Non-client clicks not supported, such as for scrollbar elements.


int function Skype__winclick__send(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

; Returns the sendMessage() result.
return Skype__winclick___doClick(True, hwnd, which, x, y)
endFunction

void function Skype__winclick__post(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

Skype__winclick___doClick(False, hwnd, which, x, y)
endFunction

int function Skype__winclick___doClick(int useSend, handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() or postMessage call.
; See Skype__winclick__sendClick() or Skype__winclick__postClick() for details on usage.
; Returns the sendMessage() result if sendMessage() is used, or nothing if postMessage() is used.
; Fine print: This function sends two messages, ; one for buttonDown and one for buttonUp.
; The sendMessage result comes from the first of these calls when sendMessage is used.
; Only one message is sent for a hover or leave.
let which = stringLower(which)
var int msgDown, int msgUp
var int wParam let wParam = 0
if stringContains(which, "hover") then
	let msgDown = 0x02A1  ; WM_MOUSEHOVER
	let msgUp = 0  ; skip that one.
elif stringContains(which, "leave") then
	let msgDown = 0x02A3  ; WM_MOUSELEAVE
	let msgUp = 0  ; skip that one.
elif stringContains(which, "left") then
	if stringContains(which, "dbl") then
		let msgDown = 0x0203  ; WM_LBUTTONDBLCLK
	else
		let msgDown = 0x0201  ; WM_LBUTTONDOWN
	endIf
	let msgUp = 0x0202  ; WM_LBUTTONUP
	let wParam = wParam | 1  ; MK_LBUTTON
elif stringContains(which, "right") then
	if stringContains(which, "dbl") then
		let msgDown = 0x0206  ; WM_RBUTTONDBLCLK
	else
		let msgDown = 0x0204  ; WM_RBUTTONDOWN
	endIf
	let msgUp = 0x0205  ; WM_RBUTTONUP
	let wParam = wParam | 2  ; MK_RBUTTON
elif stringContains(which, "middle") then
	if stringContains(which, "dbl") then
		let msgDown = 0x0209  ; WM_MBUTTONDBLCLK
	else
		let msgDown = 0x0207  ; WM_MBUTTONDOWN
	endIf
	let msgUp = 0x0208  ; WM_MBUTTONUP
	let wParam = wParam | 0x10  ; MK_MBUTTON
else
	beep()
	return False
endIf
if stringContains(which, "shift") then
	let wParam = wParam | 4  ; MK_SHIFT
endIf
if stringContains(which, "ctrl") || stringContains(which, "control") then
	let wParam = wParam | 8  ; MK_CONTROL
endIf
var int lParam let lParam = 0
if x || y then
	if !msgUp then
		saveCursor()  JAWSCursor()  saveCursor()
		moveTo(x, y)
	endIf
	var int left, int right, int top, int bottom
	; -4 is ObjID_Client
	if Skype__MSAA__windowRect(hwnd, -4, left, right, top, bottom) then
		let x = x -left
		let y = y -top
	else
		let x = x -getWindowLeft(hwnd)
		let y = y -getWindowTop(hwnd)
	endIf
	let lParam = makeLong(x, y)
endIf
var int result
if useSend then
	let result = sendMessage(hwnd, msgDown, wParam, lParam)
	if msgUp then
		sendMessage(hwnd, msgUp, 0, lParam)
	endIf
else
	postMessage(hwnd, msgDown, wParam, lParam)
	if msgUp then
		postMessage(hwnd, msgUp, 0, lParam)
	endIf
endIf
if !msgUp then
	pause()
endIf
return result
endFunction

;============================= End module winclick =============================


;================================= Module MSAA =================================
; Miscellaneous functions for MSAA access
;
; Author:  Doug Lee

string function Skype__MSAA__windowName(handle hwnd, /*?*/ int childID, int objID)
; Get the MSAA name for the given window.
return Skype__MSAA__windowProp(hwnd, "name", childID, objID)
endFunction

int function Skype__MSAA__windowRect(handle hwnd, int objID, int byRef left, int byRef right, int byRef top, int byRef bottom)
; Return by reference the rectangle for the given object ID in the given window.
; Returns True on success and False on failure.
let left = 0
let right = 0
let top = 0
let bottom = 0
var object o, int childID
let o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o then
	return False
endIf
if !Skype__MSAA__rect(o, 0, left, right, top, bottom) then
	return False
endIf
return True
endFunction

variant function Skype__MSAA__windowProp(handle hwnd, string whichProp, /*?*/ int childID, int objID)
; Returns any one of several possible MSAA properties for the given window.
; Supported properties:
;	Name, Role, State, Value.
;	Description, KeyboardShortcut, Help, ChildCount.
var
	variant result,
	object o
if whichProp == "childCount" || whichProp == "role" || whichProp == "state" then
	let result = Skype__util__vcast(0)
else
	let result = Skype__util__vcast("")
endIf
if !hwnd then
	return result
endIf
let o = Skype__MSAA__objectFromWindow(hwnd, objID)
if !o then
	return result
endIf
let childID = childID +0
if whichProp == "name" then
	return o.accName(childID)
elif whichProp == "state" then
	return o.accState(childID)
elif whichProp == "role" then
	return o.accRole(childID)
elif whichProp == "value" then
	return o.accValue(childID)
elif whichProp == "description" then
	return o.accDescription(childID)
elif whichProp == "help" then
	return o.accHelp(childID)
elif whichProp == "keyboardShortcut" then
	return o.accKeyboardShortcut(childID)
elif whichProp == "childCount" then
	if childID then
		return o.accChild(childID).accChildCount
	else
		return o.accChildCount
	endIf
endIf
return result
endFunction

int function Skype__MSAA__click(object o, int childID)
; Click the center of the indicated MSAA element.
; This function does not check directly for invisible, offscreen, obscured, etc.,
; but it avoids clicking if the element at the click location seems to be the wrong one.
; Returns True if clicked and False if not.
var
	int left, int right, int top, int bottom
var string name let name = o.accName(childID)
var string role let role = o.accRole(childID)
if !Skype__MSAA__rect(o, childID, left, right, top, bottom) then
	return False
endIf
saveCursor()  JAWSCursor()  saveCursor()
var int x let x = (left+right) / 2
var int y let y = (top+bottom) / 2
moveTo(x, y)
if Skype__util__abs(x-getCursorCol()) > 2 || Skype__util__abs(y-getCursorRow()) > 2 then
	; That's to protect against off-screen coordinates mostly.
	return False
endIf
var object o1, int childID1
let o1 = getObjectAtPoint(childID1, x, y)
if childID1 != childID
|| o.accRole(childID1) != role
|| stringCompare(o1.accName(childID1), name, True) != 0 then
	return False
endIf
	leftMouseButton()
	return True
endFunction

int function Skype__MSAA__childByName(object o, string name, /*?*/ string flags, int startID, int endID)
; Returns the child ID whose object's name matches the given name, or 0 if no match is found.
; Flags:
;	c = case sensitive, otherwise not.
;	p = prefix match (returned name may be longer), otherwise full match.
;	s = substring (name contains given text).
;	x = exact, do not strip leading and trailing blanks.
; p and s are mutually exclusive.
; startID: The starting childID to check.
; endID:  The last childID to check.
var
	int childID,
	string curname
if o.accRole(0) == 9 then  ; 9 == Role_System_Window
	let o = o.accChild(-4)  ; client object
endIf
if !startID then
	let startID = 1
endIf
let childID = o.accChildCount
if !endID || endID > childID then
	let endID = childID
endIf
let childID = startID
var int stripEnds let stripEnds = (stringContains(flags, "x") == 0)
if stripEnds then
	let name = stringTrimLeadingBlanks(stringTrimTrailingBlanks(name))
endIf
var int ignoreCase let ignoreCase = (stringContains(flags, "c") == 0)
if ignoreCase then
	let name = stringLower(name)
endIf
var int len let len = stringLength(name)
var int substringMatch let substringMatch = (stringContains(flags, "s") > 0)
var int prefixMatch let prefixMatch = (stringContains(flags, "p") >= 0)
var int curlen
while childID <= endID
	let curname = o.accName(childID)
	if stripEnds then
		let curname = stringTrimLeadingBlanks(stringTrimTrailingBlanks(curname))
	endIf
	let curlen = stringLength(curname)
	if curlen == 0
	&& !(o.accState(childID) & (State_System_OffScreen | State_System_Invisible)) then
		; Use on-screen text instead.
		let curName = Skype__MSAA__rectText(o, childID)
		if stripEnds then
			let curname = stringTrimLeadingBlanks(stringTrimTrailingBlanks(curname))
		endIf
		let curlen = stringLength(curname)
	endIf
	if ignoreCase then
		let curname = stringLower(curname)
	endIf
	if substringMatch then
		if stringContains(curname, name) then
			return childID
		endIf
	elif prefixMatch
		if curlen >= len
			curname = stringLeft(curname, len)
			if name == curname
				return childID
			endIf
		endIf
	elif stringCompare(name, curname, True)
		return childID
	endIf
	let childID = childID +1
endWhile
return 0
endFunction

string function Skype__MSAA__rectText(object o, int childID, /*?*/ int attributes, int textColor, int backgroundColor, int bIncludeLineBreaks, int bGetBrailleText)
var int left, int right, int top, int bottom
if !Skype__MSAA__rect(o, childID, left, right, top, bottom) then
	return ""
endIf
; This handles optional parameters as best we can do portably.
if bGetBrailleText then
	return getTextInRect(left, top, right, bottom, attributes, textColor, backgroundColor, bIncludeLineBreaks, bGetBrailleText)
elif bIncludeLineBreaks then
	return getTextInRect(left, top, right, bottom, attributes, textColor, backgroundColor, bIncludeLineBreaks)
elif stringLength(Skype__util__vcast(backgroundColor)) then
	return getTextInRect(left, top, right, bottom, attributes, textColor, backgroundColor)
elif stringLength(Skype__util__vcast(textColor)) then
	return getTextInRect(left, top, right, bottom, attributes, textColor)
elif stringLength(Skype__util__vcast(attributes)) then
	return getTextInRect(left, top, right, bottom, attributes)
else
	return getTextInRect(left, top, right, bottom)
endIf
endFunction

int function Skype__MSAA__rect(object o, int childID, int byRef left, int byRef right, int byRef top, int byRef bottom)
; Get the bounding rectangle for the given MSAA object and childID.
; The rectangle is returned 1-based, as JAWS likes its rectangles.
; MSAA provides 0-based left/top/width/height.
if !o then
	let left = 0
	let right = 0
	let top = 0
	let bottom = 0
	return False
endIf
if !childID then
	let childID = 0
endIf
o.accLocation(intRef(left), intRef(top), intRef(right), intRef(bottom), childID)
if !left && !right && !top && !bottom then
	return False
endIf
; but we want 1-based left/right/top/bottom.
let left = left +1
let top = top +1
let right = left +right
let bottom = top +bottom
return True
endFunction

globals
	handle Skype__MSAA___lastHwnd,
	int Skype__MSAA___lastObjID,
	object Skype__MSAA___lastObj

object function Skype__MSAA__objectFromWindow(handle hwnd, /*?*/ int objID)
; Return the requested (default ObjID_Client) MSAA object from the given window.
if !stringLength(Skype__util__vcast(objID)) then
	let objID = -4  ; ObjID_Client
endIf
if hwnd == Skype__MSAA___lastHwnd && objID == Skype__MSAA___lastObjID && Skype__MSAA___lastObj then
	return Skype__MSAA___lastObj
endIf
var object o, int childID
let o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o then
	return o
endIf
let Skype__MSAA___lastHwnd = hwnd
let Skype__MSAA___lastObjID = objID
let Skype__MSAA___lastObj = o
return o
endFunction

;=============================== End module MSAA ===============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

int function Skype__util__waitForKeysToClear(/*?*/ int nTenths)
; Wait up to ntenths of a second for shift-type keys to be released.
; This was initially written for NT-style Windows versions, in which TypeKey (and {keyName} constructs)
; have been seen to fail when such keys are considered down already.
; If nTenths is 0 or not passed, this function will wait indefinitely for the keys to be released.
; The keys waited for are alt, control, shift, and insert.
; Returns True on success and False on failure.
; Fails if GetKeyState fails to get a key state, or if nTenths expires.
; Written 01/02/03 by DGL.
var
	string keyList,
	int i,
	int isDown, int isOn,
	string keyName,
	int count
let keyList = "alt|control|shift|insert"
let count = 0
let i = 1
let keyName = stringSegment(keyList, "|", i)
if !nTenths then
	let nTenths = 0
endIf
while keyName && count <= nTenths
	if !getKeyState(keyName, isDown, isOn) then
		; Fail if GetKeyState fails
		return False
	endIf
	if isDown then
		; try again in a tenth of a second unless we're out of time
		if nTenths then
			let count = count +1
		endIf
		if count <= nTenths then
			delay(1)
		endIf
	else
		; This key is not down; go to the next one if there is another.
		let i = i +1
		let keyName = stringSegment(keyList, "|", i)
		; No extra delay or count increment when jumping to another key name.
	endIf
endWhile
; If all keys are up, isDown will be False.
return !isDown
EndFunction

int function Skype__util__abs(int n)
; Return the absolute value of n.
if n < 0 then  return 0-n
else return n
endIf
endFunction

int function Skype__util__isSpecialFocus()
; Returns True if there is a special condition for which normal focus handling should not apply.
; Currently, the conditions that are considered special are menusActive, userBufferIsActive, inHJDialog, and switchingTasks (Alt+Tab is active).
return menusActive() ||userBufferIsActive() ||inHJDialog() || Skype__util__switchingTasks()
endFunction

int function Skype__util__switchingTasks()
; Returns True if the user is currently Alt+Tabbing between tasks.

; This approach is adapted from JAWS 9.0.2169's default.jss FocusChangedEventEx function.
if getWindowClass(getFocus()) == "#32771"
&& getObjectTypeCode()== WT_LISTBOXITEM then
	return True
endIf
return False
endFunction

int function Skype__util__setRestrictionToWindow(handle hwnd)
; Restrict the active cursor to the given window's borders.
var int left, int right, int top, int bottom
if !getWindowRect(hwnd, left, right, top, bottom) then
	return False
elif !left && !right && !top && !bottom then
	return False
endIf
return setRestrictionToRect(left, top, right, bottom)
endFunction

string function Skype__util__stringFromFile(string fname, /*?*/ string fmt)
; Returns the file contents as a string.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt,
	string s
let iFmt = -2  ; TristateUseDefault
if fmt == "a" then
	let iFmt = 0  ; TristateFalse
elif fmt == "u" then
	let iFmt = -1  ; TristateTrue
endIf
let fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso then
let fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso then
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
let f = fso.openTextFile(fname, 1, 0, iFmt)  ; 1 is ForReading, 0 means don't create
if !f then
	;sayString("No code file")
	return ""
endIf
let s = f.readAll()
f.close()
if !stringLength(s) then
	;sayString("No code in the file")
endIf
return s
endFunction

int function Skype__util__stringToFile(string fname, string s, /*?*/ string fmt)
; Writes the string to the indicated file, creating or replacing the file.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt
let iFmt = -2  ; TristateUseDefault
if fmt == "a" then
	let iFmt = 0  ; TristateFalse
elif fmt == "u" then
	let iFmt = -1  ; TristateTrue
endIf
let fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso then
let fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso then
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
let f = fso.openTextFile(fname, 2, 1, iFmt)  ; 2 is ForWriting, 1 means create
if !f then
	;sayString("Can't create file")
	return ""
endIf
f.write(s)
f.close()
return True
endFunction

handle function Skype__util__getTaskBarWindow()
return findWindow(Skype__util__getDesktopWindow(), "MSTaskSwWClass", "")
endFunction

handle function Skype__util__getFirstTopLevelWindow()
; Returns the handle of the first window at the top level of the window tree.
var
	handle hwnd

let hwnd = Skype__util__getDesktopWindow()
if hwnd then
	return getFirstChild(hwnd)
endIf

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
let hwnd = getFocus()
if !hwnd then
	let hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf

; Find the top level of the window tree without relying on getAppMainWindow.
; Note that this can actually reach the parent of the top level if we are starting on a ComboLBox window!
while getParent(hwnd)
	let hwnd = getParent(hwnd)
endWhile

; make sure this is the level we want, not the desktop itself.
; At least in Windows 95/98, the desktop has window type 14 (desktop), class "#32769", and handle 128.
; The handle is 65566 rather than 128 in Windows NT.  [DGL, 10/07/01]
; We use the first of the above properties because it seems the most likely to remain unique for a window with no parents.
if !getNextWindow(hwnd) && !getPriorWindow(hwnd) then
	let hwnd = getFirstChild(hwnd)
else
	let hwnd = getFirstWindow(hwnd)
endIf
return hwnd
endFunction

handle function Skype__util__getDesktopWindow()
; Analogous to the Windows call of the same name.
var
	handle hwnd,
	string hlist, int hcount, int i

; Switching to the invisible cursor keeps GetWindowTypeCode from returning 3 (WT_Edit) for all windows when the virtual cursor is active.
saveCursor() invisibleCursor()

; First try handles known to be the Desktop window in various versions of Windows.
; 128=95/98, 65566=NT, 65556=XP, 65546=2k.
; Thanks to Victor Tsaran for helping collect these handles. [DGL, 04/28/03]
; 65552 for Windows 8.1.
let hlist = "128|65566|65556|65552|65546|65536|65526|65516|65506"
let hcount = 8
let i = 1
while i <= hcount
	let hwnd = stringToHandle(stringSegment(hlist, "|", i))
	if getWindowTypeCode(hwnd) == WT_Desktop then
		return hwnd
	endIf
	let i = i +1
endWhile

; The hard way...

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
let hwnd = getFocus()
if !hwnd then
	let hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf
let hwnd = getFirstWindow(getTopLevelWindow(hwnd))
while hwnd
	if getWindowTypeCode(getParent(hwnd)) == WT_Desktop then
		return getParent(hwnd)
	endIf
	let hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

string function Skype__util__callStackInfo(int nlevels, string flags, /*?*/ int skip)
; Produce a representation of the current call stack formatted according to the given parameters.
; The following shortenings are applied to each script/function call in the return value:
;	- The shared folder path is replaced with "shared."
;	- The user folder path is removed.
;	- The word "function" is removed.
;	- The word "script" is replaced with "$."
; Requires JAWS 10 or later because the JAWS 10+ getScriptCallStack() is required.
; nlevels: Number of levels from starting position up the stack to print.  1 means just current function.  0 or less means everything.
; flags: f for forward order (default is reverse), s for short (one line and no parameters).
; skip: How many calls to skip over besides this function itself, default 0.
; Returns the requested call stack information or null if the JAWS version is too old.
var
	string userFolder, string sharedFolder,
	int short, int forward,
	int stacksize
if Skype__util__getJFWVersion() < 1000000 then
	return ""
endIf
let flags = stringLower(flags)
let short = stringContains(flags, "s")
let forward = stringContains(flags, "f")
let userFolder = getJAWSSettingsDirectory()
let sharedFolder = getSharedSettingsDirectory()
var collection stack = Skype__util__callStack()
var int i let i = 2  ; skip this function's entry.
let i = i +skip
let stacksize = stack.count
if nlevels > 0 then
	let nlevels = min(nlevels, stacksize-i+1)
else
	let nlevels = stacksize-i+1
endIf
if nlevels <= 0 then
	return ""
endIf
var int delta let delta = 1
if forward then
	let delta = -1
	let i = i +nlevels -1
endIf
var string buf let buf = ""
while nlevels
	if buf then
		if forward then
			let buf = buf +" -> "
		else
			let buf = buf +" <- "
		endIf
	endIf
	var collection  entry = stack[intToString(i)]
	var string fname = entry.file
	let fname = stringReplaceSubstrings(fname, sharedFolder, "")
	let fname = stringReplaceSubstrings(fname, userFolder, "user")
	if stringRight(fname, 4) == ".jsb" then
		let fname = stringChopRight(fname, 4)
	endIf
	let buf = buf +fname +" "
	if entry.type == "script" then
		let buf = buf +"$"
	endIf
	if short then
		let buf = buf +stringSegment(entry.call, "(", 1)
	else
		let buf = buf +Skype__string__ref(entry.call)
	endIf
	let i = i +delta
	let nlevels = nlevels -1
endWhile
return buf
endFunction

globals object Skype__util___csRE
collection function Skype__util__callStack()
; Return a collection of collections representing the script call stack.
; Requires JAWS 11 update 2 or later and the VBScript.RegExp object.
; Index the result by stringified 1-based integers.
; The most recent call is item 1.
; Structure of an item:
;	file: The file from which the script/function came, either "builtin" or a full path.
;	type: The type ("script" or "function", lower case).
;	call: The full call with parameters.
if !Skype__util___csRE then
	let Skype__util___csRE = Skype__util__createObject("VBScript.RegExp")
	if !Skype__util___csRE then
		return
	endIf
	let Skype__util___csRE.Global = -1
	let Skype__util___csRE.IgnoreCase = 0
	let Skype__util___csRE.Pattern = "\10([^\10\t]*)\t(script |function )"
endIf
var string sStack let sStack = "\10" +getScriptCallStack()
var object matches let matches = Skype__util___csRE.execute(sStack)
var object match
var int n let n = matches.count
var int i let i = 1  ; Skip this function itself.
var string tmp, int pos1, int pos2
var collection stack let stack = new collection
var collection entry
stack.count = 0
while i < n
	entry = new collection
	stack.count = stack.count +1
	stack[intToString(stack.count)] = entry
	let match = matches(i)
	; Skip the \10 here.
	; Also switch from 0-base to 1-base.
	let pos1 = match.firstIndex +2
	if i == n-1 then
		let pos2 = stringLength(sStack) +1
	else
		let pos2 = matches(i+1).firstIndex +1
	endIf
	let tmp = substring(sStack, pos1, pos2-pos1)
	entry.file = Skype__string__segmentNext(tmp, "\t")
	entry.type = Skype__string__segmentNext(tmp, " ")
	entry.call = tmp
	let i = i +1
endWhile
return stack
endFunction

globals object Skype__util___fso
object function Skype__util__getFSObject()
; Cached access to a filesystem object.
if !Skype__util___fso then
	let Skype__util___fso = Skype__util__createObject("Scripting.FilesystemObject")
endIf
return Skype__util___fso
endFunction

globals
	int Skype__util___JAWSVer
int function Skype__util__getJFWVersion()
; GetJFWVersion with caching to make it faster.
; Also includes the build number even when GetJFWVersion() does not.
var
	string js
if !Skype__util___JAWSVer then
	let Skype__util___JAWSVer = getJFWVersion()
	let js = intToString(Skype__util___JAWSVer)
	if stringRight(js, 3) == "000" then
		let js = Skype__util__getJAWSInfo()
		let js = stringSegment(js, "|", 1)
		let js = stringSegment(js, ".", 3)
		let Skype__util___JAWSVer = Skype__util___JAWSVer +stringToInt(js)
	endIf
endIf
return Skype__util___JAWSVer
endFunction

string function Skype__util__getJAWSInfo()
; Returns a list of fields, separated by "|", as follows:
; 1.  The JAWS version number from the JFWUI2 window, in the form [M]M.mm.bbb; example, 5.00.584.
; 2.  The JAWS serial number from the JFWUI2 window.
; 3.  The JAWS serial number from GetJFWSerialNumber().
; The build number returned by this function is accurate, whereas it is "000" in GetJFWVersion in older JAWS versions.
var
	handle hwnd0, handle hwnd,
	int pos,
	string sn, string ver

; Find the JAWS dialog, which exists even when invisible or when JAWS is on the system tray.
let hwnd0 = getFirstChild(findTopLevelWindow("JFWUI2", "JAWS"))
if !hwnd0 then
	return ""
endIf

; Get the serial number from the status bar's window name.
let hwnd = getNextWindow(hwnd0)
let sn = stringLower(getWindowName(hwnd))
let pos = stringContains(sn, "ber")
if pos then
	let sn = intToString(stringToInt(substring(sn, pos+4, 10)))
else
	let sn = ""
endIf

; Find the static containing the version number and get it.
let hwnd = getFirstChild(hwnd0)
let ver = ""
while hwnd && !ver
	let ver = stringTrimTrailingBlanks(getWindowName(hwnd))
	let pos = stringContainsFromRight(ver, " ")
	if pos then
		let ver = stringChopLeft(ver, pos)
	else
		let ver = ""
	endIf
	let hwnd = getNextWindow(hwnd)
endWhile

; The above version retrieval sometimes fails, so try the normal way.
if !ver then
	ver = intToString(getJFWVersion())
	var int verlen = stringLength(ver)
	ver = formatString("%1.%2.%3",
		stringChopRight(ver, 5),
		substring(ver, verlen-4, 1),
		stringRight(ver, 4)
	)
endIf

if !sn && !ver then
	return ""
else
	return ver +"|" +sn +"|" +getJFWSerialNumber()
endIf
endFunction

object function Skype__util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o let o = createObjectEx(objname, False)
if !o then
	let o = createObject(objname)
endIf
return o
endFunction

int function Skype__util__setFocus(handle hwnd, /*?*/ string method, handle hwndDest)
; Enhanced SetFocus function allowing any of a few methods to set focus.
; Example usages:
;	Skype__util__setFocus(hwnd):  Same as setFocus(hwnd).
;	Skype__util__setFocus(hwnd, "w"):  Use accSelect on the MSAA Window object.
;	Skype__util__setFocus(hwnd, "d"):  Use WM_NEXTDLGCTL against hwnd's top-level window.
;	Skype__util__setFocus(hwnd, "d", hwndOther):  Use WM_NEXTDLGCTL against hwndOther.
;	Skype__util__setFocus(0, "n"):  Use WM_NEXTDLGCTL to move one tab stop forward in the current screen.
; Method:  One of the following:
;	null or not passed:  Default, use the JAWS SetFocus function.
;	w:  Use MSAA accSelect(SELFLAG_TAKEFOCUS, ObjID_Window).
;	c:  Use MSAA accSelect(SELFLAG_TAKEFOCUS, ObjID_Client).
;	d:  Use the WM_NEXTDLGCTL Windows API message.
;	n:  Use WM_NEXTDLGCTL to go to the next tab stop.
;	p:  Use WM_NEXTDLGCTL to go to the previous tab stop.
; hwndDest: Optional and only used with methods d, n, and p above.
; Specifies which window to send the WM_NEXTDLGCTL message to.
; If not passed, the top-level window above hwnd is used.
; For n and p, hwnd may be 0, in which case hwndDest defaults to the current top-level window.
; By default, focus is not changed if focus is already on hwnd.
; Instead, SayObjectTypeAndText() is called to announce as if focus was changed.
; Method can also contain an exclamation (!) to force the focus change even if focus is already on hwnd.
; Returns True if a focus change was attempted and False if not.
var int force let force = False
if stringContains(method, "!") then
	let method = stringReplaceSubstrings(method, "!", "")
	let force = True
endIf
if !force && hwnd && getFocus() == hwnd then
	sayObjectTypeAndText()
	return False
endIf
if !method then
	return setFocus(hwnd)
elif method == "w" || method == "c" then
	var int objID let objID = 0  ; ObjID_Window
	if method == "c" then
		let objID = -4  ; ObjID_Client
	endIf
	var int childID
	var object o let o = getObjectFromEvent(hwnd, objID, 0, childID)
	if !o then
		return False
	endIf
	; 1 is SelFlag_TakeFocus
	o.accSelect(1, 0)
	return True
elif method == "d" || method == "n" || method == "p" then
	if !hwndDest then
		if hwnd then
			let hwndDest = getTopLevelWindow(hwnd)
		else
			let hwndDest = getTopLevelWindow(getFocus())
		endIf
	endIf
	; 40 is WM_NEXTDLGCTL.
	var int msgno let msgno = 40
	var int wParam, int lParam
	if method == "n" then
		let wParam = 0  ; next tab stop
		let lParam = 0
	elif method == "p" then
		let wParam = 1  ; non-zero: previous tab stop.
		let lParam = 0
	else  ; "d"
		let wParam = Skype__util__vcast(hwnd)
		; 1 (True) means set focus to the hwnd given in wParam,
		let lParam = 1
	endIf
	sendMessage(hwndDest, msgno, wParam, lParam)
	return True
endIf
return False
endFunction

variant function Skype__util__vcast(variant v)
; Cast any type to variant.
return v
endFunction

int function Skype__util__collectionIsEmpty(collection c)
; Return True if the given collection is empty and False if not.
var string k
forEach k in c
	; If this line is reached, there is at least one collection member.
	return False
endForEach
; And if this line is reached, there are none.
return True
endFunction

;=============================== End module util ===============================


;================================ Module string ================================
; String manipulation functions.
;
; Author:  Doug Lee

string function Skype__string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
let pos = 0
if sep then
	let pos = stringContains(s, sep)
endIf
if !pos || !sep then
	let s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	let s = ""
	return s1
endIf
let s1 = stringLeft(s, pos-1)
let s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

int function Skype__string__segmentCount(string s, string sep)
; Returns the number of segments in the given string.
; If sep is null, the result is 0.
; As of 03/15/03, the null string is defined to have zero segments (previously one).
; Any non-null string containing no separators has one segment.
; WARNING: Unlike the built-in (as of JAWS 7.0) stringSegmentCount function,
; this function counts the null string after a trailing delimiter;
; thus, Skype__string__segmentCount("a|b|", "|") is 3, where JAWS would say 2.
; Also, this function considers a multi-character sep as a multi-character delimiter,
; whereas stringSegmentCount allows each character of sep to be a delimiter.
; This function requires JAWS 7.0 or later.
var
	int len,
	int seplen,
	int l,
	int n
let seplen = stringLength(sep)
if seplen == 1 then
	let n = stringSegmentCount(s, sep)
	if stringRight(s, 1) == sep then
		; JAWS does not count the null string after a trailing delimiter, but I do.
		let n = n +1
	endIf
	return n
endIf
let len = stringLength(s)
if !len then
	return 0
endIf
if !seplen then
	return 0
endIf

; A little optimization made possible by JAWS 6.0's StringReplaceSubstrings function...
; Count segments by removing all separators and comparing string lengths.
let l = stringLength(Skype__string__ref(stringReplaceSubstrings(s, sep, "")))
let l = len -l
return (l / seplen) +1
endFunction

int function Skype__string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s) then
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0") then
	let whatToDo = "l"
	elif stringToInt(whatToDo) then
	let whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

let whatToDo = stringLower(whatToDo)
let trimEnds = stringContains(whatToDo, "t")
let considerCase = stringContains(whatToDo, "c")
let fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e" then
	if trimEnds then
		let sa = stringStripAllBlanks(s)
	else
		let sa = Skype__string__ref(s)
	endIf
	let s = ""
	let l1 = stringLength(sa)
	let pos = 1
	while pos < l1
		let s = Skype__string__cat(s, Skype__string__cat(substring(sa, pos, 1), r2))
		let pos = pos +1
	endWhile
	let s = Skype__string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1) then
	if fromWhere == "l" || fromWhere == "a" then
		if trimEnds then
			let sb = stringTrimLeadingBlanks(s)
		else
			let sb = s +""
		endIf
		let sa = stringLeft(sb, 1)
		let sb = stringChopLeft(sb, 1)
		if trimEnds then
			let sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds then
			let sb = stringTrimTrailingBlanks(s)
		else
			let sb = s +""
		endIf
		let sa = stringChopRight(sb, 1)
		if trimEnds then
			let sa = stringTrimTrailingBlanks(sa)
		endIf
		let sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb) then
		return 0
	endIf
	let s = Skype__string__cat(sa, Skype__string__cat(r2, sb))
	return 1
endIf

let changeCount = 0
while !done
	let l1 = stringLength(r1)
	let changed = false
	if fromWhere == "l" then
		if !stringCompare(stringLeft(s, l1), r1, considerCase) then
			if trimEnds then
				let s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				let s = r2 + s
			else
				let s = r2 + stringChopLeft(s, l1)
			endIf
			let changed = true
		endIf
	elif fromWhere == "r" then
		if !stringCompare(stringRight(s, l1), r1, considerCase) then
			if trimEnds then
				let s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				let s = s + r2
			else
				let s = stringChopRight(s, l1) + r2
			endIf
			let changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase then
			let pos = StringContains(s, r1)
		else
			let pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos then
			if trimEnds then
				let sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				let sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				let s = " " +sa +r2 +sb
				let s = stringChopLeft(s, 1)
			else
				let s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				let s = stringChopLeft(s, 1)
			endIf
			let changed = True
		endIf
	endIf
	if fromWhere == "e" then
		let done = !changed
	else
		let done = True
	endIf
	if changed then
		let changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function Skype__string__join(string s1, string sep, string s2)
; Returns the two given strings joined with the given separator.
; If s1 or s2 is empty, the separator is not included.
if !s1 then
	return s2
elif !s2 then
	return s1
else
	return s1 +sep +s2
endIf
endFunction

string function Skype__string__quantify(string s)
; A bit like FormatString for singular/plural issues:
; In s, any sequence of the form {<n>:<s1>/<s2>} becomes <s1> if <n> is 1 and <s2> otherwise.
; Also, any sequence of the form {<n>:<s0>/<s1>/<s2>} becomes <s0> if n is 0, <s1> if <n> is 1, and <s2> otherwise.
; Works well in conjunction with FormatString().
; Example: Skype__string__quantify(formatString("{%1:No calls/%1 call/%1 calls} {%1:has/have} come in today.", intToString(nchats))):
; nchats=0: "No calls have come in today."
; nchats=1: "1 call has come in today."
; nchats=2: "2 calls have come in today."
; Substrings can be null; example: Skype__string__quantify(formatString("%1 call{%1:/s}", intToString(nchats)))
; (0 calls, 1 call, 2 calls, etc.).
var
	int pos1, int pos2,
	string tbuf,
	int n, string lhs, string rhs,
	int zeroSeg,
	string result,
	int safetyCount
let safetyCount = 500
let result = ""
while safetyCount
	let pos1 = stringContains(s, "{")
	if !pos1 then
		return Skype__string__cat(result, s)
	endIf
	let pos2 = stringContains(stringChopLeft(s, pos1), "}")
	if !pos2 then
		return Skype__string__cat(result, s)
	endIf
	let pos2 = pos2 +pos1
	let tbuf = substring(s, pos1+1, pos2-pos1-1) ; string between but not including braces
	let lhs = stringLeft(s, pos1-1)  ; up to but not including {
	let rhs = stringChopLeft(s, pos2)  ; after but not including }
	if stringContains(tbuf, ":") && stringContains(tbuf, "/") then
		let n = stringToInt(stringSegment(tbuf, ":", 1))
		let tbuf = stringChopLeft(tbuf, stringContains(tbuf, ":"))
		; The next line returns 0 if there's only one slash and 1 if more than one slash.
		let zeroSeg = (stringContains(tbuf, "/") != stringContainsFromRight(tbuf, "/"))
		if n == 0 && zeroSeg then
			let n = 1
		elif n == 1 then
			let n = 1 +zeroSeg
		else
			let n = 2 +zeroSeg
		endIf
		let tbuf = stringSegment(tbuf, "/", n)
	else
		let tbuf = "{" +tbuf +"}"  ; original material not already in lhs/rhs
	endIf
	let result = Skype__string__cat(result, Skype__string__cat(lhs, tbuf))
	let s = rhs +""
	let safetyCount = safetyCount -1
endWhile
; TODO: Report.
return result
endFunction

string function Skype__string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function Skype__string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;============================== End module string ==============================

Script Saving ()
SaveCursor ()
if !Skype__schat__isChatWindow(getFocus(), False)
sayMessage(OT_Error, msgNotInChatWindow)
elif FindString (GetFocus (), "Salva", S_BOTTOM, S_UNRESTRICTED) then
LeftMouseButton ()
SayString ("Premuto")
else
SayString ("Non trovo il pulsante salva")
EndIf
RestoreCursor ()
EndScript

Script Canceling ()
SaveCursor ()
if !Skype__schat__isChatWindow(getFocus(), False)
sayMessage(OT_Error, msgNotInChatWindow)
elif FindString (GetFocus (), "annulla", S_BOTTOM, S_UNRESTRICTED) then
LeftMouseButton ()
SayString ("Annullato!")
Else
SayString ("Non trovo il pulsante annulla")
EndIf
RestoreCursor ()
EndScript

Script Joining ()
SaveCursor ()
SpeechOff ()
;Skype___switchToControl(SWT_LiveConversations)
{alt+2}
pause ()
{shift+tab}
pause ()
JawsHome ()
pause ()
{Shift+F10}
pause ()
pause ()
{t}
pause ()
SpeechOn ()
SayString ("contatto unito alla conferenza")
RestoreCursor ()
EndScript

Script RemainingTime ()
var
Handle x,
Int a,
int b
Let x=GetFocus ()
SaveCursor ()
if !Skype__schat__isChatWindow(getFocus(), False)
sayMessage(OT_Error, msgNotInChatWindow)
elif (FindString (x, "rimanenti", S_BOTTOM, S_UNRESTRICTED)) then
SayLine ()
Elif (FindString (x, " stato", S_BOTTOM, S_UNRESTRICTED))
&&
FindString (x, "file inviato", S_BOTTOM, S_UNRESTRICTED) then
FindString (x, " stato", S_BOTTOM, S_UNRESTRICTED)
Let a=GetLineTop ()
FindString (x, "file inviato", S_BOTTOM, S_UNRESTRICTED)
Let b=GetLineTop ()
If (a>b) then
SayString ("Il file  stato ricevuto")
else
SayString ("file inviato")
EndIf
Elif (FindString (x, " stato", S_BOTTOM, S_UNRESTRICTED)) then
SayString ("Il file  stato ricevuto")
Elif (FindString (x, "file inviato", S_BOTTOM, S_UNRESTRICTED)) then
SayLine ()
 ElIf (FindString (x, "completato", S_BOTTOM, S_UNRESTRICTED)) then
SayString ("File completati")
Else
SayString ("non trovo il tempo rimanente")
EndIf
RestoreCursor ()
EndScript

Script ShowingFile ()
SaveCursor ()
if !Skype__schat__isChatWindow(getFocus(), False)
sayMessage(OT_Error, msgNotInChatWindow)
Elif (FindString (GetFocus (), " stato", S_BOTTOM, S_UNRESTRICTED)) then
JAWSEnd ()
PriorWord ()
LeftMouseButton ()
ElIf (FindString (GetFocus (), "completato", S_BOTTOM, S_UNRESTRICTED)) then
JAWSEnd ()
PriorWord ()
LeftMouseButton ()
EndIf
RestoreCursor ()
EndScript